Package: NCA
Type: Package
Title: Necessary Condition Analysis
Version: 5.0.1
Date: 2026-04-29
Authors@R: c(person(given = "Jan",
                    family = "Dul",
                    role = "aut",
                    email = "jdul@rsm.nl"),
             person(given = "Govert",
                    family = "Buijs",
                    role = "cre",
                    email = "buijs@rsm.nl"))
Description: Performs a Necessary Condition Analysis (NCA). (Dul, J. 2016. Necessary Condition Analysis (NCA). ''Logic and Methodology of 'Necessary but not Sufficient' causality." Organizational Research Methods 19(1), 10-52) <doi:10.1177/1094428115584005>.
  NCA identifies necessary (but not sufficient) conditions in datasets, where x causes (e.g. precedes) y. Instead of drawing a regression line ''through the middle of the data'' in an xy-plot, NCA draws the ceiling line. The ceiling line y = f(x) separates the area with observations from the area without observations.
  (Nearly) all observations are below the ceiling line: y <= f(x). The empty zone is in the upper left hand corner of the xy-plot (with the convention that the x-axis is ''horizontal'' and the y-axis is ''vertical'' and that values increase ''upwards'' and ''to the right''). The ceiling line is a (piecewise) linear non-decreasing line: a linear step function or a straight line. It indicates which level of x (e.g., an effort, a characteristic) is necessary but not sufficient for a (desired or undesired) level of y (e.g., good performance or disease). A quick start guide for using this package can be found here: <https://repub.eur.nl/pub/78323/> or <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2624981>.
URL:
        https://www.eur.nl/en/erim/erim/research-initiatives/necessary-condition-analysis
License: GPL (>= 3)
Depends: R (>= 3.5.0)
Imports: gplots, quantreg, KernSmooth, lpSolve, ggplot2, doParallel,
        foreach, iterators, plotly, truncnorm, DBI, RSQLite
NeedsCompilation: no
Repository: CRAN
Suggests: testthat
Packaged: 2026-04-28 16:56:59 UTC; govert
Author: Jan Dul [aut],
  Govert Buijs [cre]
Maintainer: Govert Buijs <buijs@rsm.nl>
Date/Publication: 2026-04-28 17:30:02 UTC
Built: R 4.5.2; ; 2026-05-01 18:57:04 UTC; unix

Package: NCA
Type: Package
Title: Necessary Condition Analysis
Version: 5.0.2
Date: 2026-06-16
Authors@R: c(person(given = "Jan",
                    family = "Dul",
                    role = "aut",
                    email = "jdul@rsm.nl"),
             person(given = "Govert",
                    family = "Buijs",
                    role = "cre",
                    email = "buijs@rsm.nl"))
Description: Performs a Necessary Condition Analysis (NCA). (Dul, J. 2016. Necessary Condition Analysis (NCA). ''Logic and Methodology of 'Necessary but not Sufficient' causality." Organizational Research Methods 19(1), 10-52) <doi:10.1177/1094428115584005>.
  NCA identifies necessary (but not sufficient) conditions in datasets, where x causes (e.g. precedes) y. Instead of drawing a regression line ''through the middle of the data'' in an xy-plot, NCA draws the ceiling line. The ceiling line y = f(x) separates the area with observations from the area without observations.
  (Nearly) all observations are below the ceiling line: y <= f(x). The empty zone is in the upper left hand corner of the xy-plot (with the convention that the x-axis is ''horizontal'' and the y-axis is ''vertical'' and that values increase ''upwards'' and ''to the right''). The ceiling line is a (piecewise) linear non-decreasing line: a linear step function or a straight line. It indicates which level of x (e.g., an effort, a characteristic) is necessary but not sufficient for a (desired or undesired) level of y (e.g., good performance or disease). A quick start guide for using this package can be found here: <https://repub.eur.nl/pub/78323/> or <https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2624981>.
URL:
        https://www.eur.nl/en/erim/erim/research-initiatives/necessary-condition-analysis
License: GPL (>= 3)
Depends: R (>= 3.5.0)
Imports: gplots, quantreg, KernSmooth, lpSolve, ggplot2, doParallel,
        foreach, iterators, plotly, truncnorm, DBI, RSQLite
NeedsCompilation: no
Repository: CRAN
Suggests: testthat
Packaged: 2026-06-17 14:11:07 UTC; govert
Author: Jan Dul [aut],
  Govert Buijs [cre]
Maintainer: Govert Buijs <buijs@rsm.nl>
Date/Publication: 2026-06-17 14:50:11 UTC

export(nca, nca_analysis, nca_output, nca_outliers,
       nca_random, nca_power, nca_powerplot, nca_extract, nca_util_normalize,
       nca_difference,
       ceilings, line.colors, line.types, line.width, point.type, point.color)
S3method(plot, nca_result)
S3method(print, nca_result)
S3method(summary, nca_result)
S3method(print, nca_outliers)
S3method(print, nca_difference)
import(gplots)
import(ggplot2)
import(grid)
importFrom("plotly", "plot_ly", "layout", "add_trace", "add_lines")
importFrom("parallel", "detectCores")
importFrom("doParallel", "registerDoParallel", "stopImplicitCluster")
importFrom("foreach", "foreach", "%dopar%", "registerDoSEQ")
importFrom("iterators", "iter")
importFrom("lpSolve", "lp")
importFrom("KernSmooth", "dpik")
importFrom("utils", "packageDescription", "combn")
importFrom("graphics", "abline", "legend", "par", "plot", "strwidth", "title", "clip", "lines", "points", "text")
importFrom("grDevices", "dev.new", "dev.off", "pdf")
importFrom("stats", "coef", "complete.cases", "ecdf", "lm", "quantile", "aggregate")
importFrom("utils", "tail", "head", "capture.output")
importFrom("quantreg", "rq")
importFrom("stats", "rnorm", "runif", "var", "qnorm", "median", "setNames")
importFrom("grDevices", "col2rgb")
importFrom("truncnorm", "rtruncnorm")
importFrom("DBI", "dbConnect", "dbDisconnect", "dbWriteTable", "dbExecute", "dbReadTable")
importFrom("RSQLite", "SQLite")

p_display_bottleneck <-
  function (bottlenecks, title = "Bottleneck", pdf = FALSE, path = NULL) {
    if (pdf) {
      # Put all tables in 1 file
      p_new_pdf("bottlenecks", colnames(bottlenecks[[1]])[1], path, paper = "A4r")

      for (method in names(bottlenecks)) {
        p_display_table_pdf(bottlenecks[[method]], p_pretty_name(method), title)
      }

      # Close the file
      dev.off()
      cat("")
    } else {
      for (method in names(bottlenecks)) {
        p_display_table_screen(bottlenecks[[method]], p_pretty_name(method), title)
      }
    }
  }

p_display_table_pdf <-
  function (bn, method, title) {
    names <- colnames(bn)
    bn.x <- attr(bn, "bn.x")
    bn.y <- attr(bn, "bn.y")
    bn.y.id <- attr(bn, "bn.y.id")
    rows <- nrow(bn)
    x.length <- ncol(bn) - 1
    size <- attr(bn, "size")
    cutoff <- attr(bn, "cutoff")

    # TODO Weird bug : can't replace colnames in bn
    tmp <- matrix(nrow = rows, ncol = x.length)
    for (i in 1:x.length) {
      if (bn.x == 'percentile') {
        suppressWarnings(cases <- round(size * as.numeric(bn[, i + 1]) / 100, digits = 0))
        cases <- ifelse(bn[, i + 1] == "NN", 0, cases)
        tmp[, i] <- paste0(bn[, i + 1], ' (', as.character(cases), ')')
      } else {
        tmp[, i] <- bn[, i + 1]
      }
    }
    if (length(tmp) == 0) {
      return()
    }

    # Set y precision
    if (bn.y.id %in% c(1, 2)) {
      digits <- ifelse((100 / (rows - 1)) %% 1 == 0, 0, 1)
    } else {
      digits <- p_get_digits(bn[, 1])
    }
    col.names <- as.character(1:x.length)
    row.names <- sapply(bn[, 1], p_pretty_number, "", digits)

    if (!is.null(title) && title != "") {
      title <- paste(title, method, ":", names[1])
    }
    legend <- paste0(bn.y, " / ", bn.x, "\n")
    for (i in seq(x.length)) {
      legend <- paste0(legend, i, " ", names[i + 1], "\n")
    }

    start <- 1
    while (start < (rows)) {
      end <- min(start + 30, rows)

      tmp.part <- matrix(tmp[start:end,], ncol = x.length)
      colnames(tmp.part) <- col.names
      rownames(tmp.part) <- row.names[start:end]

      # A new window for each part
      textplot(tmp.part, cex = 1, halign = "left", valign = "top", mar = c(0, 0, 3, 0))
      title(title, cex.main = 1, sub = legend)

      start <- end
    }
  }

p_display_table_screen <-
  function (bn, method, title) {
    names <- colnames(bn)
    bn.x <- attr(bn, "bn.x")
    bn.y <- attr(bn, "bn.y")
    bn.y.id <- attr(bn, "bn.y.id")
    rows <- nrow(bn)
    x.length <- ncol(bn) - 1
    size <- attr(bn, "size")
    cutoff <- attr(bn, "cutoff")

    # TODO Weird bug : can't replace colnames in bn
    tmp <- matrix(nrow = rows, ncol = x.length)
    for (i in 1:x.length) {
      values <- bn[, i + 1]
      if (bn.x == 'percentile') {
        cases <- attr(values, "cases")
        tmp[, i] <- paste0(values, ' (', as.character(cases), ')')
      } else {
        tmp[, i] <- values
      }
    }
    if (length(tmp) == 0) {
      return()
    }

    # Set y precision
    if (bn.y.id %in% c(1, 2)) {
      digits <- ifelse((100 / (rows - 1)) %% 1 == 0, 0, 1)
    } else {
      digits <- p_get_digits(bn[, 1])
    }
    colnames(tmp) <- as.character(1:x.length)
    rownames(tmp) <- sapply(bn[, 1], p_pretty_number, "", digits, TRUE)

    # Display header
    fmt <- sprintf(" %%-%ds", max(nchar(names)))
    cat("\n", strrep('-', dash.count), "\n", sep = "")
    message(title, " ", method, " (cutoff = ", cutoff, ")")
    message("Y", sprintf(fmt, names[1]), " (", bn.y, ")")
    for (i in seq(x.length)) {
      message(i, sprintf(fmt, names[i + 1]), " (", bn.x, ")")
    }
    cat(strrep('-', dash.count), "\n", sep = "")

    # Display table, insert the 'Y' for the first column
    cat("Y")
    print(tmp, quote = FALSE)
    message()
  }

p_display_table_screen_tab <-
  function (bn, method, title) {
    names <- colnames(bn)
    bn.x <- attr(bn, "bn.x")
    bn.y <- attr(bn, "bn.y")
    bn.y.id <- attr(bn, "bn.y.id")
    rows <- nrow(bn)
    x.length <- ncol(bn) - 1
    size <- attr(bn, "size")
    cutoff <- attr(bn, "cutoff")

    # TODO Weird bug : can't replace colnames in bn
    tmp <- matrix(nrow = rows, ncol = x.length)
    for (i in 1:x.length) {
      if (bn.x == 'percentile') {
        suppressWarnings(
          cases <- round(size * as.numeric(bn[, i + 1]) / 100, digits = 0)
        )
        tmp[, i] <- paste0(bn[, i + 1], ' (', as.character(cases), ')')
      } else {
        tmp[, i] <- bn[, i + 1]
      }
    }
    if (length(tmp) == 0) {
      return()
    }

    # Set y precision
    if (bn.y.id %in% c(1, 2)) {
      digits <- ifelse((100 / (rows - 1)) %% 1 == 0, 0, 1)
    } else {
      digits <- p_get_digits(bn[, 1])
    }
    colnames(tmp) <- as.character(1:x.length)
    rownames(tmp) <- sapply(bn[, 1], p_pretty_number, "", digits, TRUE)

    # Display header
    tab <- 8
    tabs <- ceiling((3 + max(nchar(names))) / tab)
    offset <- floor(log10(x.length)) + 2
    fmt <- sprintf("%%%ds ", offset - 1)
    cat("\n----------------------------------------")
    cat("----------------------------------------\n")
    message(title, " ", method, " (cutoff = ", cutoff, ")")
    needed <- tabs - floor((offset + nchar(names[1])) / tab)
    message(sprintf(fmt, "Y"), names[1], strrep("\t", needed), "(", bn.y, ")")
    for (i in seq(x.length)) {
      needed <- tabs - floor((offset + nchar(names[i + 1])) / tab)
      message(sprintf(fmt, i), names[i + 1], strrep("\t", needed), "(", bn.x, ")")
    }
    cat("----------------------------------------")
    cat("----------------------------------------\n")

    # Display table
    tabs <- ceiling(max(nchar(tmp)) / tab)
    first <- ceiling(max(nchar(colnames(tmp))) / tab)
    output <- c("Y", strrep("\t", first))
    output <- c(output, paste0(colnames(tmp), strrep("\t", tabs)), "\n")
    for (idx.row in seq_len(nrow(tmp))) {
      needed <- first - floor(nchar(rownames(tmp)[idx.row]) / tab)
      output <- c(output, rownames(tmp)[idx.row], strrep("\t", needed))
      for (idx.col in seq_len(ncol(tmp))) {
        needed <- tabs - floor(nchar(tmp[idx.row, idx.col]) / tab)
        output <- c(output, tmp[idx.row, idx.col], strrep("\t", needed))
      }
      output <- c(output, "\n")
    }
    message(paste(output, collapse = ''))

    message()
  }

p_add_extracts <-
  function (model) {
    for (idx in seq_along(p_GLOBAL_NAMES)) {
      name <- p_GLOBAL_NAMES[idx]
      model[[name]] <- lapply(model$summaries, function (s) { return(s$global[idx,]) })
    }

    for (idx in seq_along(p_PARAM_NAMES)) {
      name <- p_PARAM_NAMES[idx]
      if (name == " ") {
        next
      }
      model[[name]] <- lapply(model$summaries, function (s) { return(as.list(s$params[idx,])) })
    }

    return(model)
  }

nca_extract <-
  function (model, x = NULL, ceiling = NULL, param = 'Effect size') {
    # Validate X
    if (is.null(x)) {
      x <- names(model$summaries)[1]
      cat(paste0("'x' is not defined, using '", x, "'\n\n"))
    }
    names <- names(model$summaries)
    if (!(x %in% names)) {
      stop(x, " is not a valid independent variable name\n",
           "  Options: '", paste(names, collapse = "' '"), "'")
    }

    # Validate ceiling
    if (is.null(ceiling)) {
      ceiling <- colnames(model$summaries[[1]]$params)[1]
      cat(paste0("'ceiling' is not defined, using '", ceiling, "'\n\n"))
    }
    names <- colnames(model$summaries[[1]]$params)
    if (!(ceiling %in% names)) {
      stop(ceiling, " is not a valid ceiling name\n",
           "  Options: '", paste(names, collapse = "' '"), "'")
    }

    # Validate param
    names <- setdiff(c(p_GLOBAL_NAMES, p_PARAM_NAMES), " ")
    if (!(param %in% names)) {
      stop(param, " is not a valid param name\n",
           "  Options: '", paste(names, collapse = "' '"), "'")
    }

    if (param %in% p_GLOBAL_NAMES) {
      global <- model$summaries[[x]]$global
      row.idx <- match(param, rownames(global))
      return(global[row.idx])
    }
    else {
      params <- model$summaries[[x]]$params
      row.idx <- match(param, rownames(params))
      col.idx <- match(ceiling, colnames(params))
      return(params[[row.idx, col.idx]])
    }
  }

p_GLOBAL_NAMES  <- c("Number of observations", "Scope",
                     "Xmin", "Xmax", "Ymin", "Ymax")
p_PARAM_NAMES  <- c(
  "Ceiling zone", "Effect size", "# above",
  "Slope", "Intercept",
  "p-value", "p-accuracy",
  " ",
  "Complexity", "Fit", "Ceiling accuracy", "Accuracy nof", "Noise", "Noise nof",
  "Exceptions", "Exceptions nof", "Support", "Support nof", "Spread", "Sharpness",
  "  ",
  "Abs. ineff.", "Rel. ineff.",
  "Condition ineff.", "Outcome ineff.")

p_ceilings_step  <- c("ce_vrs", "ce_fdh")
p_ceilings_line  <- c("ols", "cols", "qr",
                      "cr_vrs", "cr_fdh", "c_lp")
ceilings         <- c(p_ceilings_step, p_ceilings_line)
p_ceiling_custom <- "custom"
p_no_bottleneck  <- c("ols")
p_no_peer_line   <- c("cols", "qr", "c_lp", "custom")

# Keep in sync with line.colors.Rd and line.type.Rd
line.colors     <- list(ols="green",        c_lp="blue",          cols="darkgreen",
                        qr="lightpink",     ce_vrs="orchid4",     cr_vrs="violet",
                        ce_fdh="red",       cr_fdh="orange",      custom="purple")
line.types      <- list(ols=1,              c_lp=2,               cols=3,
                        qr=4,               ce_vrs=5,             cr_vrs=1,
                        ce_fdh=6,           cr_fdh=1,             custom=7)

line.width      <- .75
point.type      <- 21
point.color     <- "blue"
point.size      <- 2.0

dash.count      <- 75

# Used to compare floats
epsilon         <- 1e-10
delta           <- 1E6

p_parallel_force <- "NCA_PARALLEL_FORCE"
p_parallel_start <- "NCA_PARALLEL_START"
p_skip_purity    <- "NCA_SKIP_PURITY"

p_ENLARGEMENT_CONSTANT <- 1e-10

p_ceiling <-
function (loop.data, slope, intercept) {
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y

  if (is.na(slope) || is.na(intercept)) {
    return( NaN )
  }

  # Upper left and lower right
  if ( flip.x == flip.y && slope < 0) {
    return( NaN )
  }
  # Lower left and upper right
  if ( flip.x != flip.y && slope > 0) {
    return( NaN )
  }

  # Get the intersections of the line and theo[1] / theo[2]
  theo <- loop.data$scope.theo
  p.l <- c(theo[1], slope * theo[1] + intercept)
  p.r <- c(theo[2], slope * theo[2] + intercept)

  # Line completly outside scope
  if (p.l[2] > theo[4] && p.r[2] > theo[4]) {
    return( ifelse(!flip.y, NaN, loop.data$scope.area) )
  } else if (p.l[2] < theo[3] && p.r[2] < theo[3]) {
    return( ifelse(!flip.y, loop.data$scope.area, NaN) )
  }

  if (slope > 0) {
    # Line doesn't go through the left boundary but lower boundry
    if (p.l[2] < theo[3]) {
      p.l <- c((theo[3] - intercept) / slope, theo[3])
    }

    # Line doesn't go through the right boundary but upper boundry
    if (p.r[2] > theo[4]) {
      p.r <- c((theo[4] - intercept) / slope, theo[4])
    }

    area_above <- 0.5 * (p.r[1] - p.l[1]) * (p.r[2] - p.l[2])
    area_above <- area_above + (theo[2] - theo[1]) * (theo[4] - p.r[2])
    area_above <- area_above + (p.l[1] - theo[1]) * (theo[4] - theo[3])
    area_above <- area_above - (p.l[1] - theo[1]) * (theo[4] - p.r[2])
  } else if (slope < 0) {
    # Nope, line goes through upper boundary
    if (p.l[2] > theo[4]) {
      p.l <- c((theo[4] - intercept) / slope, theo[4])
    }

    # Line doesn't go through lower boundary but right boundary
    if (p.r[2] < theo[3]) {
      p.r <- c((theo[3] - intercept) / slope, theo[3])
    }

    area_above <- 0.5 * (p.r[1] - p.l[1]) * (p.l[2] - p.r[2])
    area_above <- area_above + (theo[2] - theo[1]) * (theo[4] - p.l[2])
    area_above <- area_above + (theo[2] - p.r[1]) * (theo[4] - theo[3])
    area_above <- area_above - (theo[2] - p.r[1]) * (theo[4] - p.l[2])
  } else if (slope == 0) {
    area_above <- (theo[2] - theo[1]) * (theo[4] - intercept)
  }

  return( ifelse(!flip.y, area_above, loop.data$scope.area - area_above) )
}

p_ce_ceiling <-
function (loop.data, peers, method) {
  emp <- loop.data$scope.emp
  theo <- loop.data$scope.theo
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y

  # Get the part of the ceiling between theo and emp scope
  ceiling <- p_scope_ceiling(peers, theo, emp, flip.x, flip.y)

  # Single peer, different cases for FHD+VRS and conf lines
  if (nrow(unique(peers)) <= 1) {
    # if there is only one peer, the ceiling zone is zero or 'scope ceiling'
    if (method %in% c("fdh", "vrs")) {
      return ( ceiling )
    } else {
      stop("method not allowed")
    }
  }

  for (i in 1:(nrow(peers)-1)) {
    if (method == "fdh") {
      x.length <- peers[i+1,1] - emp[1 + flip.x]
      y.length <- peers[i+1,2] - peers[i,2]
      ceiling <- ceiling + abs(x.length * y.length)
    } else if (method == "vrs") {
      part.a <- (peers[i+1,2] - peers[i,2]) * (peers[i+1,1] - peers[1,1])
      part.b <- 0.5 * (peers[i+1,2] - peers[i,2]) * (peers[i+1,1] - peers[i,1])
      ceiling <- ceiling + abs(part.a) - abs(part.b)
    } else {
      stop("method not allowed")
    }
  }

  return( ceiling )
}

p_scope_ceiling <-
function (peers, theo, emp, flip.x, flip.y) {
  if (identical(emp, theo)) {
    return(0)
  }

  x1 <- emp[1] - theo[1]
  x2 <- theo[2] - emp[2]
  y1 <- emp[3] - theo[3]
  y2 <- theo[4] - emp[4]

  left    <- x1 * (theo[4] - theo[3])
  right   <- x2 * (theo[4] - theo[3])
  lower   <- (theo[2] - theo[1]) * y1
  upper   <- (theo[2] - theo[1]) * y2

  if (!flip.x && !flip.y) {
    # Upper left corner
    cross <- x1 * y2
    return (left + upper - cross)
  } else if (flip.x && !flip.y) {
    # Upper right corner
    cross <- x2 * y2
    return (right + upper - cross)
  } else if (flip.x && flip.y) {
    # Lower right corner
    cross <- x2 * y1
    return (right + lower - cross)
  } else if (!flip.x && flip.y) {
    # Lower left corner
    cross <- x1 * y1
    return (left + lower - cross)
  }
}

# TODO Is this the correct place?
p_nca_wrapper <-
function (ceiling, loop.data, bn.data, effect_aggregation) {
  analysis <- do.call(paste0("p_nca_", ceiling), list(loop.data, bn.data))

  loop.data$flip.x  <- !loop.data$flip.x
  if (2 %in% effect_aggregation) {
    tmp_analysis <- do.call(paste0("p_nca_", ceiling), list(loop.data, bn.data))
    analysis$effect <- analysis$effect + tmp_analysis$effect
  }

  loop.data$flip.y  <- !loop.data$flip.y
  if (4 %in% effect_aggregation) {
    tmp_analysis <- do.call(paste0("p_nca_", ceiling), list(loop.data, bn.data))
    analysis$effect <- analysis$effect + tmp_analysis$effect
  }

  loop.data$flip.x  <- !loop.data$flip.x
  if (3 %in% effect_aggregation) {
    tmp_analysis <- do.call(paste0("p_nca_", ceiling), list(loop.data, bn.data))
    analysis$effect <- analysis$effect + tmp_analysis$effect
  }

  return ( analysis )
}

test_that("str_length is number of characters", {
  if (!exists("p_ceiling")) {
    # Needed for testing manually
    source("../../R/p_ceiling.R")
  }

  loop.data <- list(scope.theo=c(1, 6, 2, 6), scope.area=20)

  test1 <- TRUE
  test2 <- TRUE
  test3 <- TRUE
  test4 <- TRUE

  # Upper left upper corner
  if (test1) {
    loop.data$flip.x <- FALSE
    loop.data$flip.y <- FALSE

    expect_equal(p_ceiling(loop.data, 1,    8), NaN)
    expect_equal(p_ceiling(loop.data, 0,    8), NaN)
    expect_equal(p_ceiling(loop.data, -2/3, 8), NaN)
    expect_equal(p_ceiling(loop.data, -4/3, 8), NaN)

    expect_equal(p_ceiling(loop.data, 2/3,  4), 4/3)
    expect_equal(p_ceiling(loop.data, 0,    3), 15)
    expect_equal(p_ceiling(loop.data, 0,    5), 5)
    expect_equal(p_ceiling(loop.data, -2/3, 4), NaN)

    expect_equal(p_ceiling(loop.data, 4/3,  0), 8)
    expect_equal(p_ceiling(loop.data, 2/3,  0), 17)
    expect_equal(p_ceiling(loop.data, 0,    0), 20)
    expect_equal(p_ceiling(loop.data, -1,   0), NaN)
  }

  # Upper right corner
  if (test2) {
    loop.data$flip.x <- TRUE
    loop.data$flip.y <- FALSE

    expect_equal(p_ceiling(loop.data, 1,    8), NaN)
    expect_equal(p_ceiling(loop.data, 0,    8), NaN)
    expect_equal(p_ceiling(loop.data, -2/3, 8), 3)
    expect_equal(p_ceiling(loop.data, -4/3, 8), 12)

    expect_equal(p_ceiling(loop.data, 2/3,  4), NaN)
    expect_equal(p_ceiling(loop.data, 0,    3), 15)
    expect_equal(p_ceiling(loop.data, 0,    5), 5)
    expect_equal(p_ceiling(loop.data, -2/3, 4), 56/3)

    expect_equal(p_ceiling(loop.data, 4/3,  0), NaN)
    expect_equal(p_ceiling(loop.data, 2/3,  0), NaN)
    expect_equal(p_ceiling(loop.data, 0,    0), 20)
    expect_equal(p_ceiling(loop.data, -1,   0), 20)
  }

  # Lower right corner
  if (test3) {
    loop.data$flip.x <- TRUE
    loop.data$flip.y <- TRUE

    expect_equal(p_ceiling(loop.data, 1,    8), 20)
    expect_equal(p_ceiling(loop.data, 0,    8), 20)
    expect_equal(p_ceiling(loop.data, -2/3, 8), NaN)
    expect_equal(p_ceiling(loop.data, -4/3, 8), NaN)

    expect_equal(p_ceiling(loop.data, 2/3,  4), 56/3)
    expect_equal(p_ceiling(loop.data, 0,    3), 5)
    expect_equal(p_ceiling(loop.data, 0,    5), 15)
    expect_equal(p_ceiling(loop.data, -2/3, 4), NaN)

    expect_equal(p_ceiling(loop.data, 4/3,  0), 12)
    expect_equal(p_ceiling(loop.data, 2/3,  0), 3)
    expect_equal(p_ceiling(loop.data, 0,    0), NaN)
    expect_equal(p_ceiling(loop.data, -1,   0), NaN)
  }

  # Lower left corner
  if (test4) {
    loop.data$flip.x <- FALSE
    loop.data$flip.y <- TRUE

    expect_equal(p_ceiling(loop.data, 1,    8), NaN)
    expect_equal(p_ceiling(loop.data, 0,    8), 20)
    expect_equal(p_ceiling(loop.data, -2/3, 8), 17)
    expect_equal(p_ceiling(loop.data, -4/3, 8), 8)

    expect_equal(p_ceiling(loop.data, 2/3,  4), NaN)
    expect_equal(p_ceiling(loop.data, 0,    3), 5)
    expect_equal(p_ceiling(loop.data, 0,    5), 15)
    expect_equal(p_ceiling(loop.data, -2/3, 4), 4/3)

    expect_equal(p_ceiling(loop.data, 4/3,  0), NaN)
    expect_equal(p_ceiling(loop.data, 2/3,  0), NaN)
    expect_equal(p_ceiling(loop.data, 0,    0), NaN)
    expect_equal(p_ceiling(loop.data, -1,   0), NaN)
  }
})

p_nca_ce_vrs <-
  function (loop.data, bn.data) {
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y

    # Find the points on the ceiling ("PEERS")
    peers <- p_peers(loop.data, vrs = TRUE)

    line <- p_vrs_line(loop.data, peers, flip.x, flip.y)
    line_matrix <- p_vrs_line(loop.data, peers, flip.x, flip.y)
    ceiling <- p_ce_ceiling(loop.data, peers, "vrs")
    effect <- ceiling / loop.data$scope.area
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs_ce(loop.data, peers)
    bottleneck <- p_bottleneck_ce(loop.data, bn.data, peers, "vrs")
    complexity <- nrow(unique(peers))

    metrics <- p_purity_empty
    # Extra metrics for certain lines
    skip <- !is.null(loop.data[[p_skip_purity]]) || Sys.getenv(p_skip_purity) == TRUE
    skip <- skip || loop.data$flip.x || loop.data$flip.y
    if (!skip) {
      df <- data.frame(loop.data$x, loop.data$y)
      condition <- colnames(loop.data$x)
      outcome <- colnames(loop.data$y)
      colnames(df) <- c(condition, outcome)
      metrics <- p_purity_metrics(df, condition, outcome, "ce_vrs")
    }

    return(list(line = line, line_matrix = line_matrix, peers = peers,
                slope = NA, intercept = NA,
                ceiling = ceiling, effect = effect,
                above = 0, accuracy = 100, accuracy_nof = nrow(loop.data$x),
                fit = fit, ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_vrs_line <-
  function (loop.data, peers, flip.x, flip.y) {
    x.points <- c(loop.data$scope.emp[1 + flip.x],
                  peers[, 1],
                  loop.data$scope.theo[2 - flip.x])
    y.points <- c(loop.data$scope.theo[3 + flip.y],
                  peers[, 2],
                  loop.data$scope.emp[4 - flip.y])
    return(array(c(x.points, y.points), dim = c(length(x.points), 2)))
  }

p_nca_qr <-
  function (loop.data, bn.data) {
    x <- loop.data$x
    y <- loop.data$y
    flip.y <- loop.data$flip.y
    # Make sure 0 <= qr.tau <= 1
    qr.tau <- loop.data$qr.tau
    if (qr.tau < 0 || qr.tau > 1) {
      qr.tau <- 0.95
    }

    if (!flip.y) {
      # Upper corners
      qr <- rq(y ~ x, tau = qr.tau)
    } else if (flip.y) {
      # Lower corner
      qr <- rq(y ~ x, tau = 1 - qr.tau)
    }

    intercept <- unname(coef(qr)["(Intercept)"])
    slope <- unname(coef(qr)["x"])
    line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
    ceiling <- p_ceiling(loop.data, slope, intercept)
    effect <- ceiling / loop.data$scope.area
    above <- p_above(loop.data, slope, intercept)
    accuracy <- p_accuracy(loop.data, above)
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1
    metrics <- p_purity_empty

    return(list(line = qr, line_matrix = line_matrix, peers = NULL,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_corners <-
  function (df, condition, outcome) {
    peers <- p_get_peers(df, condition, outcome)

    # corner
    df$corner <- df$caseNo %in% rownames(peers)

    # start
    df$start <- FALSE
    for (peer.x in peers[, 1]) {
      same.x <- df[condition] == peer.x
      lowest.y <- df[outcome] == min(df[df[condition] == peer.x, outcome])
      df$start <- df$start | same.x & lowest.y
    }

    # peer
    highest.peer.idx <- as.integer(tail(rownames(peers), n = 1))
    df$peer <- data.frame(peer = rep(highest.peer.idx, nrow(df)))
    if (nrow(peers) >= 2) {
      for (row.idx in nrow(peers):2) {
        peer.x <- peers[row.idx, 1]
        peer.idx <- as.integer(rownames(peers)[row.idx - 1])
        df$peer[df[condition] < peer.x] <- peer.idx
      }
    }

    # size
    df$size <- NA
    counts <- table(df$peer)
    caseNos <- as.integer(names(counts))
    for (i in 1:nrow(counts)) {
      value <- counts[i]
      caseNo <- caseNos[i]
      df$size[df$caseNo == caseNo] <- value
    }

    return(df)
  }

p_get_peers <-
  function (df, condition, outcome, vrs = FALSE) {
    # TODO Flipping
    x <- as.matrix(df[condition])
    y <- as.matrix(df[outcome])
    loop.data <- list(x = x, y = y, flip.x = FALSE, flip.y = FALSE)
    return(p_peers(loop.data, vrs))
  }

p_fdh_nooks <-
  function (peers) {
    nooks <- matrix(nrow = 0, ncol = 2)
    for (i in 1:nrow(peers)) {
      peer <- peers[i,]
      # Add previous point if between peers
      if (i > 1) {
        nook.x <- peers[i, 1]
        nook.y <- peers[i - 1, 2]
        nooks <- rbind(nooks, c(nook.x, nook.y))
      }
      # Add the peer
      nooks <- rbind(nooks, peer)
    }

    rownames(nooks) <- NULL
    return(nooks)
  }

p_expand_fdh_nooks <-
  function (fdh_nooks) {
    if (nrow(fdh_nooks) == 0) {
      return(NULL)
    }

    fdh_nooks_expanded <- cbind(fdh_nooks, 1, 1)
    curr.col <- curr.row <- 1
    for (row.idx in 2:nrow(fdh_nooks_expanded)) {
      prev <- fdh_nooks_expanded[row.idx - 1,]
      curr <- fdh_nooks_expanded[row.idx,]
      if (prev[1] != curr[1]) {
        curr.row <- curr.row + 1
      }
      if (prev[2] != curr[2]) {
        curr.col <- curr.col + 1
      }
      fdh_nooks_expanded[row.idx, 3] <- curr.row
      fdh_nooks_expanded[row.idx, 4] <- curr.col
    }

    colnames(fdh_nooks_expanded) <- c("x", "y", "col", "row")
    return(fdh_nooks_expanded)
  }

p_crRank <-
  function (x) {
    # x: numericals to be ranked in col/row
    d <- diff(sort(x))
    d <- ifelse(d == 0, 0, 1)
    cd <- cumsum(d)   # my rank: cumulative sum
    r <- c(1, 1 + cd) # compensate for the one value lost due to differing
  }

p_addColRow2xyPoints <-
  function (xy) {
    if (nrow(xy) == 0) {
      return(NULL)
    }

    # possibly changes order of rows
    xy <- xy[order(xy$x, xy$y),]
    xy$col <- p_crRank(xy$x) # add column data
    xy$row <- p_crRank(xy$y) # add row data

    return(xy) # Warning: the order of rows possibly changed
  }

p_closingBr <-
  function (nooks, scope) {
    ceil <- xceil <- nooks
    peers_scope <- c(min(nooks[, 1]), max(nooks[, 1]), min(nooks[, 2]), max(nooks[, 2]))

    # projections
    if (peers_scope[3] > scope[3]) {
      # nooks does not meet the bottom (south side) of scope
      xceil <- rbind(c(peers_scope[1], scope[3]), xceil)
      if (peers_scope[1] > scope[1]) {
        ceil <- rbind(c(peers_scope[1], scope[3]), ceil)
      }
    }
    if (peers_scope[2] < scope[2]) {
      # nooks does not meet the east side of bb
      xceil <- rbind(xceil, c(scope[2], peers_scope[4]))
      if (peers_scope[4] < scope[4]) {
        ceil <- rbind(ceil, c(scope[2], peers_scope[4]))
      }
    }

    # add sw and ne corners of br (if not already there):
    if (length(xceil[xceil[, 1] == scope[1] & xceil[, 2] == scope[3],]) == 0) {
      # no sw so add
      xceil <- rbind(c(scope[1], scope[3]), xceil)
    }
    if (length(xceil[xceil[, 1] == scope[2] & xceil[, 2] == scope[4],]) == 0) {
      # no ne so add
      xceil <- rbind(xceil, c(scope[2], scope[4]))
    }

    return(list(ceil = ceil, xceil = xceil))
  }

p_creduceCeil <-
  function (nooks, scope) {
    south <- nooks[nooks[, 1] == scope[3],]
    west <- nooks[nooks[, 1] == scope[1],]
    if (length(south) == (2 * ncol(nooks)) || length(west) == (2 * ncol(nooks))) {
      # delete south-west (1st_pnt)
      nooks <- matrix(nooks[-1,], ncol = ncol(nooks))
    }

    north <- nooks[nooks[, 1] == scope[4],]
    east <- nooks[nooks[, 1] == scope[2],]
    if (length(north) == (2 * ncol(nooks)) || length(east) == (2 * ncol(nooks))) {
      # delete north-east
      nooks <- matrix(nooks[1:(nrow(nooks) - 1),], ncol = ncol(nooks))
    }

    # TODO I'm not sure this is correct: shouldn't this be the 'peers' scope?
    if (nrow(nooks) == 1) {
      # TODO Why not always?
      if ((nooks[1, 1] == scope[1]) & (nooks[1, 2] == scope[4])) {
        nooks <- NULL
      }
    }
    return(nooks)
  }

p_makeStp <-
  function (peers, scope) {
    if (nrow(peers) == 0) {
      xceil <- matrix(c(scope[1], scope[1], scope[2],
                        scope[3], scope[4], scope[4]), ncol = 2)
      ceilings <- list(ceil = NULL, xceil = xceil)
      return(ceilings)
    }

    nooks <- p_fdh_nooks(peers)
    ceilings <- p_closingBr(nooks, scope)
    ceil <- p_expand_fdh_nooks(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_expand_fdh_nooks(ceilings$xceil)

    return(list(ceil = ceil, xceil = xceil))
  }

p_makeVrs <-
  function (vrs_peers, scope) {
    ceilings <- p_closingBr(vrs_peers, scope)
    ceil <- p_expand_fdh_nooks(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_expand_fdh_nooks(ceilings$xceil)

    return(list(ceil = ceil, xceil = xceil))
  }

p_doLinCeil <-
  function (xyP, scope, ilk = "RG") {
    # determine line and bounding rectangle from xy points in xyP

    # we now have the raw values for a and b
    ab <- p_abCeil4Data(xyP, ilk = ilk)
    iLB <- p_intersectLineBox(ab, scope) # non-inflated
    # intersections: south-west (swi) and north-east (nei) intersections
    swi <- unname(iLB$swi) # on boundary of BB as both ab and BR derive from the same data
    nei <- unname(iLB$nei) # on boundary of BB as both ab and BR derive from the same data
    xyP <- unique(data.frame(x = c(swi[1], nei[1]), y = c(swi[2], nei[2])))
    ceilings <- p_closingBr(xyP, scope)
    # Note. For received linear cases the extended ceiling gets traced out by 4 points.

    # add col/row information on the points in the (x)ceiling:
    ceil <- p_addColRow2xyPoints(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_addColRow2xyPoints(ceilings$xceil)

    deltay <- scope[4] - scope[3]
    a1 <- (ab["a"] - scope[3] + ab["b"] * scope[1]) / deltay
    b1 <- (scope[2] - scope[1]) * ab["b"] / deltay
    nrm <- c(a = a1, b = b1)

    return(ceilings = list(ceil = ceil, xceil = xceil, ab = list(raw = ab, nrm = nrm)))
  }

###############################################################################
# HERE START
###############################################################################
p_make2SidedWeb <-
  function (ceilings, scope) {
    # compute on meshes formed by the extended ceiling's break-points
    # Meshes (maze) form a partition of the bounding rectangle based on the breakpoints
    # in the extended ceiling.

    # Note: enlceil is a dataframe, not an matrix (mixed types)
    enlceil <- cbind(data.frame(ceilings$xceil), strain = "regular")

    # if enlargement then add points to xceil to enlarge it and obtain enlceil
    if (p_ENLARGEMENT_CONSTANT > 0) {
      dx <- scope[2] - scope[1]
      dy <- scope[4] - scope[3]
      extra.row <- data.frame(x = scope[1] - p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[3] - p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = 0, row = 0, strain = "enlarged")
      enlceil <- rbind(extra.row, enlceil)

      tmp <- tail(enlceil, n = 1)
      extra.row <- data.frame(x = scope[2] + p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[4] + p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = tmp[3] + 1, row = tmp[4] + 1, strain = "enlarged")
      enlceil <- rbind(enlceil, extra.row)
      enlceil[, 3:4] <- enlceil[, 3:4] + 1
      row.names(enlceil) <- NULL
    }

    ceilings$enlceil <- enlceil

    # xy (and col-row) raster with info
    ras <- p_twoSidedRaster(enlceil)
    rp <- ras$pnts

    # TODO Rename to scope.enl
    # bounding rectangle (enlarged) coordinates from xyCR (= xceil):
    br <- c(min(rp['x']), max(rp['x']), min(rp['y']), max(rp['y']))
    # TODO Rename to scopes
    BRs <- list(obr = scope, br = br)

    # Out of the raster points we build the mesh (= partition by mazes = rectangles)
    # Meshes: raster points (x, y) are considered as south-east points in a mesh
    # below procedure may give at first some incomplete/degenerate/non-existing
    # nw points/meshes which subsequently get removed
    colnames(rp) <- c('col', 'se_x', 'strainx', 'cardDirEW',
                      'row', 'se_y', 'strainy', 'cardDirNS', 'se_pos')

    # Now compute the meshes with current points in rp considered as se-points.
    r1 <- rp
    r1 <- cbind(r1, rp[, 'row'] - 1, rp[, 'col'] + 1)
    colnames(r1) <- c('col.y', 'nw_x', 'strainx', 'cardDirEW',
                      'row.y', 'nw_y', 'strainy', 'cardDirNS',
                      'nw_pos', 'row', 'col')

    # Remove "invalid" meshes - having corners or row-col with NA-values:
    rp <- rp[complete.cases(rp[, c("row", "col")]),]
    r1 <- r1[complete.cases(r1[, c("row", "col", "row.y", "col.y", "nw_x")]),]
    rp$insol <- NA
    rp$inver <- NA

    ribbon_db <- dbConnect(SQLite(), ":memory:")

    dbWriteTable(ribbon_db, "rp", rp)
    dbWriteTable(ribbon_db, "r1", r1)
    dbWriteTable(ribbon_db, "colrangebyrow", data.frame(ras$colrangebyrow))
    dbWriteTable(ribbon_db, "rowrangebycol", data.frame(ras$rowrangebycol))

    dbExecute(ribbon_db, CREATE_MAZE,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))

    # Deviance areas to north-west
    dbExecute(ribbon_db, UPDATE_MAZE_1)

    # Deviance areas to south-east
    dbExecute(ribbon_db, UPDATE_MAZE_2)

    # Do lookup stuff (lm = lookup_maze)
    dbExecute(ribbon_db, UPDATE_MAZE_3)

    dbExecute(ribbon_db, UPDATE_MAZE_4,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))

    dbExecute(ribbon_db, UPDATE_MAZE_5)

    dbExecute(ribbon_db, UPDATE_RP)

    maze <- dbReadTable(ribbon_db, "maze")
    ras$pnts <- dbReadTable(ribbon_db, "rp")
    dbDisconnect(ribbon_db)

    # effect size:
    se_maze <- maze[maze['s_row'] == min(maze['s_row']) & maze['e_col'] == max(maze['e_col']),]
    nces <- 1 - se_maze$se_insol

    dossier <- list(message = "", portrayal = "")
    if (nrow(maze) == 1) {
      if (maze$stance != "bisect") { # ceiling fully in bb boundary
        if (maze$stance == "nw") {
          TM1 <- "\n Single mesh: Empty space goes to south-east of bounding rectangle! \n"
          TM2 <- "There is no feasible space: in-veracity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)veracity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_feasible_space"
        } else { # maze$stance == "se"
          TM1 <- "\n Single mesh: Feasible point at the north-west of bounding rectangle! \n"
          TM2 <- "There is no empty space and in-solidity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)solidity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_empty_space"
        }
      }
    }

    return(list(ceilings = ceilings, raster = ras, BRs = BRs, enlarged = p_ENLARGEMENT_CONSTANT > 0,
                maze = maze, nces = nces, dossier = dossier))
  }

p_twoSidedRaster <-
  function (enlceil) {
    # raster from xy-column-rows (later the meshes)

    col.max <- max(enlceil['col'])
    row.max <- max(enlceil['row'])

    rangeby <- NULL
    for (row_idx in seq_len(nrow(enlceil))) {
      row <- enlceil[row_idx,]
      rows <- enlceil[enlceil['row'] == row$row,]
      cols <- enlceil[enlceil['col'] == row$col,]
      rangeby <- rbind(rangeby, c(row = row$row, col = row$col,
                                  x.low = min(rows['x']), x.high = max(rows['x']),
                                  y.low = min(cols['y']), y.high = max(cols['y']),
                                  col.low = min(rows['col']), col.high = max(rows['col']),
                                  row.low = min(cols['row']), row.high = max(cols['row'])))
    }
    rownames(rangeby) <- NULL
    colrangebyrow <- rangeby[!duplicated(rangeby[, 'row']),
                             c('row', 'col.low', 'col.high', 'x.low', 'x.high')]
    rowrangebycol <- rangeby[!duplicated(rangeby[, 'col']),
                             c('col', 'row.low', 'row.high', 'y.low', 'y.high')]

    tmp <- enlceil
    tmp$cardDirEW <- ifelse(tmp$col == 1, "west", ifelse(tmp$col == col.max, "east", "inner"))
    tmp$cardDirNS <- ifelse(tmp$row == 1, "south", ifelse(tmp$row == row.max, "north", "inner"))
    distinct_col <- tmp[!duplicated(tmp$col), c('col', 'x', 'strain', 'cardDirEW')]
    colnames(distinct_col) <- c('col', 'x', 'strainx', 'cardDirEW')
    distinct_row <- tmp[!duplicated(tmp$row), c('row', 'y', 'strain', 'cardDirNS')]
    colnames(distinct_row) <- c('row', 'y', 'strainy', 'cardDirNS')

    pnts <- merge(distinct_col, distinct_row, by = NULL)
    pnts <- pnts[order(pnts$col, pnts$row),]
    pnts$hpos <- ""
    for (row_idx in seq_len(nrow(pnts))) {
      row <- pnts[row_idx,]
      tmp <- colrangebyrow[colrangebyrow[, "row"] == row$row]
      pnts[row_idx, 'hpos'] <- ifelse(row$col >= tmp[2] & row$col <= tmp[3], 'on',
                                      ifelse(row$col < tmp[2], 'nw', 'se'))
    }
    rownames(pnts) <- NULL

    return(list(pnts = pnts,
                colrangebyrow = colrangebyrow,
                rowrangebycol = rowrangebycol,
                Mcol = col.max,
                Mrow = row.max))
  }

p_sv4Points <-
  function (xyP, wb) {
    # compute sv (raw + normalized (net)) for points

    mz <- wb$maze
    br <- wb$BRs$br # (possibly) inflated bounding rectangle
    rp <- wb$raster$pnts
    # no mixing with x and y in xyP
    colnames(rp) <- c('col', 'rx', 'strainx', 'cardDirEW', 'row', 'ry',
                      'strainy', 'cardDirNS', 'hpos', 'insol', 'inver')

    # We cannot rely on xyP having caseNo as a variable identifying a row.
    xyP <- cbind(pointID = paste0("point_", xyP$caseNo), xyP)
    xyOri <- xyP # keep (almost) original input data in xyO

    xyP <- p_sv4PointsSql(xyP, mz, rp, br)

    # names <- c('pointID', 'x', 'y', 'ezon', 'feas', 'idRE', 'idRW', 'insol', 'inver')
    names <- c('caseNo', 'pointID', 'x', 'y', 'insol', 'inver')
    xyQ <- xyP[, names]

    # add normalized (in-)solidity / (in-)veracity
    nces <- wb$nces

    xyQ$insol_net <- p_nciRaw2Norm(xyQ$insol, nces, solidity = TRUE)
    xyQ$inver_net <- p_nciRaw2Norm(xyQ$inver, nces, solidity = FALSE)
    xyQ$sol_net <- 1 - xyQ$insol_net
    xyQ$ver_net <- 1 - xyQ$inver_net

    xyR <- merge(x = xyOri, y = xyQ, by = "pointID", all.x = TRUE)
    xyR <- xyR[order(xyR$pointID),]

    res <- list(compacts = xyQ, specifics = xyR)

    return(res)
  }

p_sv4PointsSql <-
  function (xyP, mz, rp, br) {
    # TODO Check what's needed
    xyP$ezon <- FALSE
    xyP$feas <- FALSE
    xyP$ver_num <- NA
    xyP$ver_den <- NA
    xyP$sol_num <- NA
    xyP$sol_den <- NA
    xyP$ver_devi <- NA
    xyP$sol_devi <- NA
    xyP$insol <- NA
    xyP$inver <- NA
    xyP$nw <- FALSE
    xyP$se <- FALSE

    sv_db <- dbConnect(SQLite(), ":memory:")
    dbWriteTable(sv_db, "xyP", xyP)
    dbWriteTable(sv_db, "rp", rp)
    dbWriteTable(sv_db, "mz", mz)

    dbExecute(sv_db, CREATE_STANCE)
    dbExecute(sv_db, CREATE_E1)
    dbExecute(sv_db, UPDATE_XYP_1)
    dbExecute(sv_db, UPDATE_XYP_2,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))
    dbExecute(sv_db, UPDATE_XYP_3, ,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))
    dbExecute(sv_db, UPDATE_XYP_4)

    xyP <- dbReadTable(sv_db, "xyP")

    dbDisconnect(sv_db)

    return (xyP)
  }

###############################################################################
# HERE END
###############################################################################

p_make2SidedWebOrg <-
  function (ceilings, scope) {
    # compute on meshes formed by the extended ceiling's break-points
    # Meshes (maze) form a partition of the bounding rectangle based on the breakpoints
    # in the extended ceiling.

    # Note: enlceil is a dataframe, not an matrix (mixed types)
    enlceil <- cbind(data.frame(ceilings$xceil), strain = "regular")

    # if enlargement then add points to xceil to enlarge it and obtain enlceil
    enlarged <- FALSE
    if (p_ENLARGEMENT_CONSTANT > 0) {
      enlarged <- TRUE
      dx <- scope[2] - scope[1]
      dy <- scope[4] - scope[3]
      extra.row <- data.frame(x = scope[1] - p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[3] - p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = 0, row = 0, strain = "enlarged")
      enlceil <- rbind(extra.row, enlceil)


      tmp <- tail(enlceil, n = 1)
      extra.row <- data.frame(x = scope[2] + p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[4] + p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = tmp[3] + 1, row = tmp[4] + 1, strain = "enlarged")
      enlceil <- rbind(enlceil, extra.row)
      enlceil[, 3:4] <- enlceil[, 3:4] + 1
      row.names(enlceil) <- NULL
    }

    ceilings$enlceil <- enlceil
    # xy (and col-row) raster with info
    ras <- p_twoSidedRasterOrg(enlceil)
    rp <- ras$pnts

    # TODO Rename to scope.enl
    # bounding rectangle (enlarged) coordinates from xyCR (= xceil):
    br <- c(min(rp['x']), max(rp['x']), min(rp['y']), max(rp['y']))

    # TODO Rename to scopes
    BRs <- list(obr = scope, br = br)

    # Out of the raster points we build the mesh (= partition by mazes = rectangles)
    # Meshes: raster points (x, y) are considered as south-east points in a mesh
    # below procedure may give at first some incomplete/degenerate/non-existing
    # nw points/meshes which subsequently get removed
    colnames(rp) <- c('col', 'se_x', 'strainx', 'cardDirEW',
                      'row', 'se_y', 'strainy', 'cardDirNS', 'se_pos')

    # Now compute the meshes with current points in rp considered as se-points.
    r1 <- rp
    r1 <- cbind(r1, rp[, 'row'] - 1, rp[, 'col'] + 1)
    colnames(r1) <- c('col.y', 'nw_x', 'strainx', 'cardDirEW',
                      'row.y', 'nw_y', 'strainy', 'cardDirNS',
                      'nw_pos', 'row', 'col')

    mesh <- merge(rp, r1, by = c("row", "col"))
    # Remove "invalid" meshes - having corners or row-col with NA-values:
    mesh <- mesh[complete.cases(mesh[, c("row.y", "col.y", "col", "row", "nw_x")]),]
    mesh <- mesh[order(mesh$col),]

    # fixing up remaining, valid, meshes, such as
    # using only the 4 ordinal directions (e, w, s, n) for the mesh:
    maze <- cbind(mesh['col'], mesh['se_x'], mesh['strainx.x'], mesh['cardDirEW.x'],
                  mesh['row'], mesh['se_y'], mesh['strainy.x'], mesh['cardDirNS.x'], mesh['se_pos'],
                  mesh['col.y'], mesh['nw_x'], mesh['strainx.y'], mesh['cardDirEW.y'],
                  mesh['row.y'], mesh['nw_y'], mesh['strainy.y'], mesh['cardDirNS.y'], mesh['nw_pos'])
    colnames(maze) <- c('e_col', 'e_x', 'strainx.x', 'cardDirEW.x',
                        's_row', 's_y', 'strainy.x', 'cardDirNS.x', 'se_pos',
                        'w_col', 'w_x', 'strainx.y', 'cardDirEW.y',
                        'n_row', 'n_y', 'strainy.y', 'cardDirNS.y', 'nw_pos')
    maze$onEdge <- 'enlarged' == maze['strainx.x'] |
      'enlarged' == maze['strainx.y'] |
      'enlarged' == maze['strainy.x'] |
      'enlarged' == maze['strainy.y']
    maze[c('strainx.x', 'strainx.y', 'strainy.x', 'strainy.y')] <- NULL

    # Note. Points (e_x, s_y) are south and east lines in their maze (cell)
    # and (w_x, n_y) are the north and west lines in that maze, and so on.
    # Note that we only track positions on the nw and se intersection points.
    # Such is sufficient (for our purposes) as the ceiling runs sw-ne.
    # w_x ~ x.low; e_x ~ x.high; s_y ~y.low; n_y ~ y.high
    # some additional info on mazes for easy identification:
    maze <- maze[with(maze, order(maze$e_x, maze$s_y)),]
    maze <- cbind(mazeID = paste0('maze_', seq.int(nrow(maze))), maze)

    # position (stance), south-east or north-west, or bisected, of meshes wrt ceiling:
    maze$stance <- as.vector(
      ifelse(maze['nw_pos'] == "nw" & maze['se_pos'] == "se", "bisect",
             ifelse(maze['nw_pos'] == "on" | maze['nw_pos'] == "se",
                    "se", "nw")))

    # Areas of meshes (with bi-sect counting for half):
    maze$area <- as.vector(
      ifelse(maze$stance == "bisect",
             (maze[['e_x']] - maze[['w_x']]) * (maze[['n_y']] - maze[['s_y']]) / 2,
             (maze[['e_x']] - maze[['w_x']]) * (maze[['n_y']] - maze[['s_y']])))
    # Do away with irrelevant 0-area meshes, probably redundant, but ... safety first:
    maze <- maze[maze['area'] > 0,]

    # Unconditional areas to north-west (for each of the 4 corners of a mesh)
    # Note. The same point, when a corner in 4 meshes, gets calculated 4 times
    maze$se_NW_den <- unlist((maze['e_x'] - br[1]) * (br[4] - maze['s_y']))
    maze$nw_NW_den <- unlist((maze['w_x'] - br[1]) * (br[4] - maze['n_y']))
    maze$sw_NW_den <- unlist((maze['w_x'] - br[1]) * (br[4] - maze['s_y']))
    maze$ne_NW_den <- unlist((maze['e_x'] - br[1]) * (br[4] - maze['n_y']))

    # and total (unconditional) area to the south-east of each point in a mesh:
    maze$se_SE_den <- unlist((br[2] - maze['e_x']) * (maze['s_y'] - br[3]))
    maze$nw_SE_den <- unlist((br[2] - maze['w_x']) * (maze['n_y'] - br[3]))
    maze$sw_SE_den <- unlist((br[2] - maze['w_x']) * (maze['s_y'] - br[3]))
    maze$ne_SE_den <- unlist((br[2] - maze['e_x']) * (maze['n_y'] - br[3]))

    # Now for in-solidity numerators (num) for each of the 4 corners of a mesh:
    maze <- cbind(maze, se_NW_num = NA, nw_NW_num = NA, sw_NW_num = NA, ne_NW_num = NA)

    # Deviance areas to north-west (for each of the 4 corners of a mesh)
    # Deviance = from what is expected for a point on the ceiling
    # (here the feasible area is to the NW of the point under consideration)
    # Note: inefficiency of a point may get (re-)calculated up to 4 times
    for (i in seq_len(nrow(maze))) { # stance in "se, bisect": 100%/50% in feas. space
      cl <- maze$e_col[i]; rw <- maze$s_row[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$se_NW_num[i] <- sum(tmp$area)

      cl <- maze$w_col[i]; rw <- maze$n_row[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$nw_NW_num[i] <- sum(tmp$area)

      rw <- maze$s_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$sw_NW_num[i] <- sum(tmp$area)


      rw <- maze$n_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$ne_NW_num[i] <- sum(tmp$area)
    }

    # # Now then for in-veracity numerators for each of the 4 corners of a maze:
    maze <- cbind(maze, se_SE_num = NA, nw_SE_num = NA, sw_SE_num = NA, ne_SE_num = NA)

    # Deviance areas to south-east (for each of the 4 corners of a maze)
    # Deviance from what is expected for a point on the ceiling
    # (here the feasible area is to the SE)
    # inefficient calculation: a point/value may get (re-)calculated up to 4 times
    for (i in seq_len(nrow(maze))) {
      rw <- maze$s_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$se_SE_num[i] <- sum(tmp$area)

      rw <- maze$n_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$nw_SE_num[i] <- sum(tmp$area)

      rw <- maze$s_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$sw_SE_num[i] <- sum(tmp$area)

      rw <- maze$n_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$ne_SE_num[i] <- sum(tmp$area)
    }

    # Finished with areas, now for the deviance calculations.
    # In-solidity and in-veracity ratios: circumvent division by zero.
    # As an in-solidity involves the north&west of the br, special care
    # (to avoid cancellation/division by zero) is needed for meshes
    # touching (osculating) the west- or north-side of the br
    maze$nw_loc <- as.character(
      ifelse((maze['w_x'] == br[1]) & (maze['n_y'] == br[4]), "north-west",
             ifelse(maze['w_x'] == br[1], "west",
                    ifelse(maze['n_y'] == br[4], "north", "other"))))
    maze$se_loc <- as.character(
      ifelse((maze['e_x'] == br[2]) & (maze['s_y'] == br[3]), "south-east",
             ifelse(maze['e_x'] == br[2], "east",
                    ifelse(maze['s_y'] == br[3], "south", "other"))))

    # vx: west-east x-value of up-going (going north(-east) ~ vertical) ceiling
    # per mesh, id-ed thru south (se-sw) row: vx = x where the ceiling crosses to ne.
    lookup_vx <- cbind(s_row = ras$colrangebyrow[, 'row'],
                       vx = ras$colrangebyrow[, 'x.high'],
                       vc = ras$colrangebyrow[, 'col.high'])
    maze <- merge(maze, lookup_vx, by = "s_row", keep = NULL)

    # the point (s_row,vc) is where the ceiling crosses the line determined
    # by the south of the mesh towards the line determined by the north of the mesh,
    # in particular (se_row,vc) is on the ceiling.

    # we need the slope for the ceiling running from the south-line to the north-line of mesh
    # if the point (se_row,vc) is on the east of the bb the slope, vs, is infinite,
    # meaning the slope is vertical, otherwise (se_row,vc) is the sw-point of a mesh in which
    # the ceiling can be assumed to run for the south-line to the north-line.
    # Search for the mesh with sw-corner (vc, se_row = sw_row):
    lookup_maze <- cbind(maze['w_col'],
                         maze['w_x'], maze['s_row'], maze['s_y'], maze['e_col'],
                         maze['e_x'], maze['n_row'], maze['n_y'], maze['stance'])
    lookup_maze <- lookup_maze[with(lookup_maze, order(lookup_maze$w_col)),]
    colnames(lookup_maze) <- c('vc', 'w_x.v', 's_row', 's_y.v', 'e_col.v',
                               'e_x.v', 'n_row.v', 'n_y.v', 'stance.v')
    xmaze <- merge(x = maze, y = lookup_maze, by = c('vc', 's_row'), all.x = TRUE)
    xmaze <- xmaze[with(xmaze, order(xmaze$e_col, xmaze$w_col)),]

    for (i in seq_len(nrow(xmaze))) {
      row <- xmaze[i,]
      tmp <- ifelse(row['stance.v'] == "bisect",
                    (row['n_y.v'] - row['s_y.v']) / (row['e_x.v'] - row['w_x.v']),
                    ifelse(is.na(row['stance.v']), Inf, Inf))
      xmaze$vslope[i] <- as.numeric(tmp)
    }
    # cleaning up superfluous values (for efficiency):
    xmaze[, colnames(xmaze)[grepl(".*\\.v", colnames(xmaze))]] <- NULL

    # Now, the analogous computation in the y-direction:
    # horizontal ceiling in south-north (hy) of maze:
    # Note. We will use south-west [] as (join_by) reference point.
    lookup_hy <- cbind(w_col = ras$rowrangebycol[, 'col'],
                       hy = ras$rowrangebycol[, 'y.high'],
                       hr = ras$rowrangebycol[, 'row.high'])
    xmaze <- merge(x = xmaze, y = lookup_hy, by = "w_col",
                   keep = NULL, all.x = TRUE)

    colnames(lookup_maze) <- c('w_col', 'w_x.h', 'hr', 's_y.h', 'e_col.h',
                               'e_x.h', 'n_row.h', 'n_y.h', 'stance.h')
    xymaze <- merge(xmaze, lookup_maze, by = c("hr", "w_col"), keep = NULL, all.x = TRUE)

    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]
      tmp <- ifelse(row['stance.h'] == "bisect",
                    (row['n_y.h'] - row['s_y.h']) / (row['e_x.h'] - row['w_x.h']),
                    ifelse(is.na(row['stance.h']), 0, 0))
      xymaze$hslope[i] <- as.numeric(tmp)
    }
    # cleaning up superfluous values (for efficiency):
    xymaze[, colnames(xymaze)[grepl(".*\\.h", colnames(xymaze))]] <- NULL

    # Note. Computing (in-)solidity: looking for feasible space to nw.
    # So, of special concern are meshes touching the bb on the north and/or west.
    # Note: when !(stance %in% c("bisect", "nw") it has to be "se".
    # So, fall-through cases stance %in% c("bisect", "nw") are "se" cases.
    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]

      sw_insol <- ifelse(row['stance'] %in% c("bisect", "nw"), 0,
                         ifelse(row['nw_loc'] == "west",
                                (row['hy'] - row['s_y']) / (br[4] - row['s_y']),
                                row['sw_NW_num'] / row['sw_NW_den']))
      xymaze$sw_insol[i] <- as.numeric(sw_insol)

      ne_insol <- ifelse(row['stance'] %in% c("bisect", "nw"), 0,
                         ifelse(row['nw_loc'] == "north",
                                (row['e_x'] - row['w_x']) / (row['e_x'] - br[1]),
                                row['ne_NW_num'] / row['ne_NW_den']))
      xymaze$ne_insol[i] <- as.numeric(ne_insol)

      nw_insol <- ifelse(row['stance'] %in% c("bisect", "nw") | (row['nw_NW_den'] == 0),
                         0,
                         row['nw_NW_num'] / row['nw_NW_den'])
      xymaze$nw_insol[i] <- as.numeric(nw_insol)

      xymaze$se_insol[i] <- as.numeric(row['se_NW_num'] / row['se_NW_den'])
    }

    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]

      sw_inver <- ifelse(row['stance'] %in% c("bisect", "se"), 0,
                         ifelse(row['se_loc'] == "south",
                                (row['e_x'] - row['w_x']) / (br[2] - row['w_x']),
                                row['sw_SE_num'] / row['sw_SE_den']))
      xymaze$sw_inver[i] <- as.numeric(sw_inver)

      ne_inver <- ifelse(row['stance'] %in% c("bisect", "se"), 0,
                         ifelse(row['se_loc'] == "east",
                                (row['n_y'] - row['s_y']) / (row['n_y'] - br[3]),
                                row['ne_SE_num'] / row['ne_SE_den']))
      xymaze$ne_inver[i] <- as.numeric(ne_inver)

      xymaze$nw_inver[i] <- as.numeric(row['nw_SE_num'] / row['nw_SE_den'])

      se_inver <- ifelse(
        row['stance'] %in% c("bisect", "se") | row['se_SE_den'] == 0,
        0, row['se_SE_num'] / row['se_SE_den'])
      xymaze$se_inver[i] <- as.numeric(se_inver)
    }

    xymaze$se_inacc <- pmax(xymaze$se_insol, xymaze$se_inver)
    xymaze$sw_inacc <- pmax(xymaze$sw_insol, xymaze$sw_inver)
    xymaze$nw_inacc <- pmax(xymaze$nw_insol, xymaze$nw_inver)
    xymaze$ne_inacc <- pmax(xymaze$ne_insol, xymaze$ne_inver)

    ras$pnts <- p_doInsolInverLookUpOrg(ras$pnts, xymaze)

    # effect size:
    se_maze <- xymaze[xymaze['s_row'] == min(xymaze['s_row']) &
                        xymaze['e_col'] == max(xymaze['e_col']),]
    nces <- 1 - se_maze$se_insol

    dossier <- list(message = "", portrayal = "")
    if (nrow(xymaze) == 1) {
      if (xymaze$stance != "bisect") { # ceiling fully in bb boundary
        if (xymaze$stance == "nw") {
          TM1 <- "\n Single mesh: Empty space goes to south-east of bounding rectangle! \n"
          TM2 <- "There is no feasible space: in-veracity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)veracity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_feasible_space"
        } else { # xymaze$stance == "se"
          TM1 <- "\n Single mesh: Feasible point at the north-west of bounding rectangle! \n"
          TM2 <- "There is no empty space and in-solidity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)solidity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_empty_space"
        }
      }
    }

    xymaze$nesw_sqr <- ifelse(
      xymaze$w_x == br[1] &
        xymaze$s_y == br[3] &
        xymaze$stance != "bisect",
      "sw_sqr",
      ifelse(
        xymaze$e_x == br[2] &
          xymaze$n_y == br[4] &
          xymaze$stance != "bisect",
        "ne_sqr",
        "other"))

    return(list(ceilings = ceilings, raster = ras, BRs = BRs, enlarged = enlarged,
                maze = xymaze, nces = nces, dossier = dossier))
  }

p_twoSidedRasterOrg <-
  function (enlceil) {
    # raster from xy-column-rows (later the meshes)
    # TODO Rename everything

    col.max <- max(enlceil['col'])
    row.max <- max(enlceil['row'])

    colx <- enlceil[c('col', 'x', 'strain')]
    distinct_col <- NULL
    for (i in seq_len(nrow(colx))) {
      row <- colx[i,]
      cardDirEW <- ifelse(row[1] == 1, "west", ifelse(row[1] == col.max, "east", "inner"))
      if (!(row[1] %in% distinct_col[, 1])) {
        distinct_col <- rbind(distinct_col, cbind(row, cardDirEW = cardDirEW))
      }
    }
    colnames(distinct_col) <- c('col', 'x', 'strainx', 'cardDirEW')

    rowy <- enlceil[c('row', 'y', 'strain')]
    distinct_row <- NULL
    for (i in seq_len(nrow(rowy))) {
      row <- rowy[i,]
      cardDirNS <- ifelse(row[1] == 1, "south", ifelse(row[1] == row.max, "north", "inner"))
      if (!(row[1] %in% distinct_row[, 1])) {
        distinct_row <- rbind(distinct_row, cbind(row, cardDirNS))
      }
    }
    colnames(distinct_row) <- c('row', 'y', 'strainy', 'cardDirNS')

    pnts <- NULL
    for (i in seq_len(nrow(distinct_col))) {
      for (j in seq_len(nrow(distinct_row))) {
        new.row <- cbind(distinct_col[i,], distinct_row[j,])
        pnts <- rbind(pnts, new.row)
      }
    }

    mincol <- pnts[pnts['col'] == 1,]
    colnames(mincol) <- paste0('m', colnames(mincol))
    maxcol <- pnts[pnts['col'] == max(pnts['col']),]
    colnames(maxcol) <- paste0('M', colnames(maxcol))
    names(maxcol)[names(maxcol) == 'Mrow'] <- 'mrow'
    horiz <- merge(mincol, maxcol, by = "mrow")

    minrow <- pnts[pnts['row'] == 1,]
    colnames(minrow) <- paste0('m', colnames(minrow))
    maxrow <- pnts[pnts['row'] == max(pnts['row']),]
    colnames(maxrow) <- paste0('M', colnames(maxrow))
    names(maxrow)[names(maxrow) == 'Mcol'] <- 'mcol'
    verti <- merge(minrow, maxrow, by = "mcol")
    # Note. horiz and verti will only be used for producing a grid in plots.

    colrangebyrow <- NULL
    for (row in seq_along(unique(enlceil[, 'row']))) {
      rows <- enlceil[enlceil['row'] == row,]
      tmp <- c(row, min(rows['col']), max(rows['col']), min(rows['x']), max(rows['x']))
      colrangebyrow <- rbind(colrangebyrow, tmp)
    }
    rownames(colrangebyrow) <- NULL
    colnames(colrangebyrow) <- c('row', 'col.low', 'col.high', 'x.low', 'x.high')

    rowrangebycol <- NULL
    for (col in seq_along(unique(enlceil[, 'col']))) {
      cols <- enlceil[enlceil['col'] == col,]
      # TODO Why is the order different?
      tmp <- c(col, min(cols['y']), max(cols['y']), min(cols['row']), max(cols['row']))
      rowrangebycol <- rbind(rowrangebycol, tmp)
    }
    rownames(rowrangebycol) <- NULL
    colnames(rowrangebycol) <- c('col', 'y.low', 'y.high', 'row.low', 'row.high')

    col.new <- NULL
    for (i in seq_len(nrow(pnts))) {
      col <- pnts[i,][['col']]
      row <- pnts[i,][['row']]
      low <- colrangebyrow[colrangebyrow[, "row"] == row][2]
      high <- colrangebyrow[colrangebyrow[, "row"] == row][3]
      tmp <- ifelse(col >= low & col <= high, 'on', ifelse(col < low, 'nw', 'se'))
      col.new <- c(col.new, tmp)
    }
    pnts$hpos <- col.new
    # TODO
    rownames(pnts) <- NULL

    return(list(pnts = pnts,
                colrangebyrow = colrangebyrow,
                rowrangebycol = rowrangebycol,
                horiz = horiz,
                verti = verti,
                Mcol = col.max,
                Mrow = row.max))
  }

p_doInsolInverLookUpOrg <-
  function (pnts, xymaze) {
    pnts$insol <- NULL
    pnts$inver <- NULL
    for (i in seq_len(nrow(pnts))) {
      row <- pnts[i,]

      if (row['cardDirEW'] != "east" & row['cardDirNS'] != "north") {
        tmp <- xymaze[xymaze['w_x'] == row[['x']] & xymaze['s_y'] == row[['y']],]
        pnts$insol[i] <- tmp$sw_insol
        pnts$inver[i] <- tmp$sw_inver
      }
      else if (row['cardDirEW'] != "east" & row['cardDirNS'] == "north") {
        tmp <- xymaze[xymaze['w_x'] == row[['x']] & xymaze['n_y'] == row[['y']],]
        pnts$insol[i] <- tmp$nw_insol
        pnts$inver[i] <- tmp$nw_inver
      }
      else if (row['cardDirEW'] == "east" & row['cardDirNS'] != "south") {
        tmp <- xymaze[xymaze['e_x'] == row[['x']] & xymaze['n_y'] == row[['y']],]
        pnts$insol[i] <- tmp$ne_insol
        pnts$inver[i] <- tmp$ne_inver
      }
      else if (row['cardDirEW'] == "east" & row['cardDirNS'] == "south") {
        tmp <- xymaze[xymaze['e_x'] == row[['x']] & xymaze['s_y'] == row[['y']],]
        pnts$insol[i] <- tmp$se_insol
        pnts$inver[i] <- tmp$se_inver
      }
      else {
        pnts$insol[i] <- NA
        pnts$inver[i] <- NA
      }
    }

    return(pnts)
  }

p_sv4PointsOrg <-
  function (xyP, wb) {
    # compute sv (raw + normalized (net)) for points

    mz <- wb$maze
    br <- wb$BRs$br # (possibly) inflated bounding rectangle
    rp <- wb$raster$pnts
    # no mixing with x and y in xyP
    colnames(rp) <- c('col', 'rx', 'strainx', 'cardDirEW', 'row', 'ry',
                      'strainy', 'cardDirNS', 'hpos', 'insol', 'inver')

    # We cannot rely on xyP having caseNo as a variable identifying a row.
    xyP <- cbind(pointID = paste0("point_", xyP$caseNo), xyP)
    xyOri <- xyP # keep (almost) original input data in xyO

    for (i in seq_len(nrow(xyP))) {
      tmp <- p_soveLookupCompsOrg(xyP[i,], rp, mz, br)
      #xyP$idRE[i] <- tmp$idRectE
      #xyP$idRW[i] <- tmp$idRectW
      #xyP$ezon[i] <- tmp$ezon
      #xyP$feas[i] <- tmp$feas
      xyP$insol[i] <- tmp$insol
      xyP$inver[i] <- tmp$inver
    }

    # names <- c('pointID', 'x', 'y', 'ezon', 'feas', 'idRE', 'idRW', 'insol', 'inver')
    names <- c('caseNo', 'pointID', 'x', 'y', 'insol', 'inver')
    xyQ <- xyP[, names]

    # add normalized (in-)solidity / (in-)veracity
    nces <- wb$nces

    xyQ$insol_net <- p_nciRaw2Norm(xyQ$insol, nces, solidity = TRUE)
    xyQ$inver_net <- p_nciRaw2Norm(xyQ$inver, nces, solidity = FALSE)
    xyQ$sol_net <- 1 - xyQ$insol_net
    xyQ$ver_net <- 1 - xyQ$inver_net

    xyR <- merge(x = xyOri, y = xyQ, by = "pointID", all.x = TRUE)
    xyR <- xyR[order(xyR$pointID),]

    res <- list(compacts = xyQ, specifics = xyR)

    return(res)
  }

p_soveLookupCompsOrg <-
  function (row, rp, mz, br) { # sove by lookup and comps
    x <- unlist(row['x'])
    y <- unlist(row['y'])

    xyS <- mz[x >= mz$w_x &
                x <= mz$e_x &
                y >= mz$s_y &
                y <= mz$n_y,]
    xyS <- xyS[order(xyS$e_x, xyS$s_y),]

    Stands <- xyS$stance
    nw <- is.element("nw", Stands)
    se <- is.element("se", Stands)

    # TODO
    # E1 <- slice(xyS[order(-xyS$e_x, xyS$s_y),], 1)
    E1 <- head(xyS, 1)

    feas <- ifelse(
      se, TRUE, ifelse(
        E1$stance == "bisect",
        (y - E1['s_y']) * (E1['e_x'] - E1['w_x']) <= (x - E1['w_x']) * (E1['n_y'] - E1['s_y']),
        FALSE))

    ezon <- ifelse(
      nw, TRUE, ifelse(
        E1$stance == "bisect",
        (y - E1['s_y']) * (E1['e_x'] - E1['w_x']) >= (x - E1['w_x']) * (E1['n_y'] - E1['s_y']),
        FALSE))

    insol <- inver <- NA
    luRP <- rp[rp['rx'] == x & rp['ry'] == y,]
    if (nrow(luRP) > 0) {
      inver <- luRP$inver
      insol <- luRP$insol
    }
    inver <- ifelse(
      !is.na(inver), inver, ifelse(
        feas, 0,
        p_deviOrg(x, y, E1['w_x'], E1['s_y'], E1['vx'], E1['hy'], E1['vslope'],
                  E1['hslope'], E1['sw_SE_num'], E1['sw_SE_den'], E1['n_y'],
                  E1['e_x'], br, E1['nesw_sqr'], ilk = "inver")))

    insol <- ifelse(
      !is.na(insol), insol, ifelse(
        ezon, 0,
        p_deviOrg(x, y, E1['w_x'], E1['s_y'], E1['vx'], E1['hy'], E1['vslope'],
                  E1['hslope'], E1['sw_NW_num'], E1['sw_NW_den'], E1['n_y'],
                  E1['e_x'], br, E1['nesw_sqr'], ilk = "insol")))

    return(list(
      inver = inver, insol = insol

      # TODO Not needed, just for development
      # , idRectE = E1$mazeID, idRectW = E1$mazeID,
      # feas = feas, ezon = ezon
    ))
  }

p_deviOrg <-
  function (x, y, x0, y0, vx, hy, sv, sh, t0, s0, ny, ex, br, sqr, ilk = "insol") {
    dev <- 0
    dx <- x - x0; dy <- y - y0 # local coordinates
    if (ilk == "inver") { # ilk = inver
      xx <- br[2]; yy <- br[3] # reference values
      # east - in-veracity, in the in-feasible space
      # y != yy unless we deal with a trivial case
      if (x == xx) {
        dev <- (y - hy) / (y - yy)
      }
      # south - in-veracity, in the in-feasible space
      # x != xx unless we deal with a trivial case
      if (y == yy) {
        dev <- (vx - x) / (xx - x)
      }
      if ((x != xx) & (y != yy)) {
        nd <- p_numdenOrg(x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy)
        dev <- ifelse(
          (sqr == "ne_sqr") & (nd$den == 0),
          dy / (y - yy),
          ifelse((sqr == "sw_sqr") & (nd$den == 0),
                 (ex - x) / (br[2] - x),
                 nd$num / nd$den))
      } # ilk = inver
    } else { # ilk = "insol"
      xx <- br[1]; yy <- br[4] # reference values
      # west - in-solidity, in the feasible space
      # y != yy unless we deal with a trivial case
      if (x == xx) {
        dev <- (y - hy) / (y - yy)
      }
      # north - in-solidity, in the feasible space
      # x != xx unless we deal with a trivial case
      if (y == yy) {
        dev <- (x - vx) / (xx - x)
      }
      if ((x != xx) & (y != yy)) {
        nd <- p_numdenOrg(x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy)
        dev <- ifelse(
          (sqr == "ne_sqr") & (nd$den == 0),
          dx / (x - xx),
          ifelse(
            (sqr == "sw_sqr") & (nd$den == 0),
            (ny - y) / (br[4] - y),
            nd$num / nd$den))
      } # ilk = "insol"
    }
    return(unlist(dev))
  }

p_numdenOrg <-
  function (x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy) {
    # num and den: ratio = num/den
    # At mesh with sw-point = (x0,y0), in local coordinates (dx = x - x0, dy = y - y0),
    # numerator = dx^2 sh/2 - dx dy + dy^2/(2 sv) + dx (hy - y0) - dy (x0 - vx) + t0
    # denominator = - dx dy + dx (ym -y0)- dy (x0 - vx) + s0
    # For iso this leads to a conic section ax^2 + 2nxy + by^2 + 2hx + 2ky + c = 0
    # reading (for in-solidity)
    # dx^2 sh/2 - (1-JJ) dx dy + dy^2/(2 sv) + dx (hy - y0  - JJ (hy -yM)) -
    #     dy (x - vx - JJ (x - xm)) + t0 - JJ s0 = 0
    # where t0 and s0 are numerator and denominator of
    # the (in-solidity) inaccuracy measure at the sw-point (x0,y0).
    dx <- x - x0; dy <- y - y0 # local coordinates
    num <- dx^2 * sh / 2 - dx * dy +
      dy^2 / (2 * sv) +
      dx * (hy - y0) - dy * (x0 - vx) + t0
    den <- -dx * dy + dx * (yy - y0) - dy * (x0 - xx) + s0
    return(list(num = num, den = den))
  }

###############################################################################
# TODO Remove above
###############################################################################

p_nciRaw2Norm <-
  function (riacc, d, solidity = TRUE) {
    # riacc = raw in-accuracy (in-solidity or in-veracity) measure
    # d = effect size; solidity = TRUE: solidity transformation

    # racc is some (solidity/veracity) raw measure
    racc <- 1 - unlist(riacc)

    if (solidity) {
      nsol <- (racc - d) / (1 - d) # nrm solidity
      return(1 - nsol)             # nrm in-solidity = (1-racc)/(1-d)
    } else {
      nver <- (racc - (1 - d)) / d # nrm veracity
      return(1 - nver)             # nrm in-veracity)
    }
  }

p_svGauge <-
  function (svP, critLevs = c(cris = .2, criv = .1)) {
    stopifnot(all(c("insol_net", "inver_net") %in% names(svP)))

    # Flags
    satS <- svP$insol_net == 0
    medS <- 0 <= svP$insol_net &
      svP$insol_net <= critLevs[["cris"]] &
      !satS
    lowS <- !medS & !satS

    satP <- svP$inver_net == 0
    medP <- 0 <= svP$inver_net &
      svP$inver_net <= critLevs[["criv"]] &
      !satP
    lowP <- !medP & !satP

    # NA-safe combos
    satD <- ifelse(is.na(satS), FALSE, satS) & ifelse(is.na(satP), FALSE, satP)
    medSsatD <- ifelse(is.na(medS), FALSE, medS) | satD

    cnt <- function (x) sum(ifelse(is.na(x), FALSE, x))
    gauge <- list(
      satD = cnt(satD),
      medS = cnt(medS),
      medP = cnt(medP),
      lowS = cnt(lowS),
      lowP = cnt(lowP),
      satS = cnt(satS),
      satP = cnt(satP),
      medSsatD = cnt(medSsatD)
    )

    # Sharpness
    denom0 <- gauge$satD + gauge$medS + gauge$medP
    sharp0 <- if (denom0 != 0) (gauge$satD + gauge$medS - gauge$medP) / denom0 else NA_real_
    denom1 <- gauge$medS + gauge$medP
    sharp1 <- if (denom1 != 0) (gauge$medS) / denom1 else NA_real_

    # Percentages & totals
    numC <- nrow(svP)
    gauge$numC <- numC
    gauge$medSpct <- if (numC > 0) 100 * gauge$medS / numC else NA_real_
    gauge$medPpct <- if (numC > 0) 100 * gauge$medP / numC else NA_real_
    gauge$medSsatDpct <- if (numC > 0) 100 * gauge$medSsatD / numC else NA_real_
    gauge$nribbon <- gauge$medSsatD + gauge$medP # number of points in ribbon
    gauge$sharp0 <- sharp0
    gauge$sharp1 <- sharp1

    # Return flags attached to svP (useful downstream)
    svP$satS <- satS
    svP$medS <- medS
    svP$lowS <- lowS
    svP$satP <- satP
    svP$medP <- medP
    svP$lowP <- lowP
    svP$satD <- satD
    svP$medSsatD <- medSsatD

    list(svP = svP, gauge = gauge)
  }

p_ensure_pct_cols <-
  function (gauge, n_points = NULL) {
    if ((!("numC" %in% names(gauge)) || is.na(gauge$numC)) && !is.null(n_points)) {
      gauge$numC <- n_points
    }
    if (is.na(gauge$numC) || gauge$numC <= 0) {
      return(gauge)
    }
    if (!("medSpct" %in% names(gauge)) && "medS" %in% names(gauge)) {
      gauge$medSpct <- 100 * gauge$medS / gauge$numC
    }
    if (!("medSsatDpct" %in% names(gauge)) && "medSsatD" %in% names(gauge)) {
      gauge$medSsatDpct <- 100 * gauge$medSsatD / gauge$numC
    }
    if (!("medPpct" %in% names(gauge)) && "medP" %in% names(gauge)) {
      gauge$medPpct <- 100 * gauge$medP / gauge$numC
    }
    return(gauge)
  }

p_get_nca_metrics <-
  function (df, condition, outcome, ceiling, test_rep = 1000) {
    model <- nca_analysis(df, condition, outcome, ceilings = ceiling, test.rep = test_rep)
    part <- function (param) nca_extract(model, x = condition, ceiling = ceiling, param = param)
    list(
      effect_size = part(param = 'Effect size'),
      p_value = part(param = 'p-value'),
      fit = part(param = 'Fit'),
      c_accuracy = part(param = 'Ceiling accuracy'),
      npeers = nrow(model$peers[[ceiling]][[condition]])
    )
  }

p_compute_spread <-
  function (svP_flags, web) {
    # x-range from the original BR (outer bounding range)
    br <- web$BRs$obr
    xmin <- br[1]
    xmax <- br[2]

    # x for rows in the ribbon: medS U satD  (NA-safe)
    in_ribbon <- ifelse(is.na(svP_flags$medSsatD), FALSE, svP_flags$medSsatD)
    x_med <- svP_flags$x[in_ribbon]

    if (length(x_med) > 0L) {
      return(p_spread_index(x_med, xmin, xmax))
    } else {
      return(NA_real_)
    }
  }

p_spread_index <-
  function (x, xmin, xmax) {
    stopifnot(is.numeric(x))
    if (!is.finite(xmin) ||
      !is.finite(xmax) ||
      xmax <= xmin) {
      stop("xmax must be > xmin, both finite.")
    }

    x <- x[is.finite(x)]
    if (length(x) == 0L) {
      return(NA_real_)
    }
    z <- sort((x - xmin) / (xmax - xmin))
    z <- pmin(pmax(z, 0), 1)
    gaps <- diff(c(0, z, 1))
    mean_gap <- mean(gaps)
    if (mean_gap == 0) {
      return(0)
    }
    cv <- stats::sd(gaps) / mean_gap
    return(1 / (1 + cv))
  }

p_abCeil4Data <-
  function (xyP, ilk = "RG") {

    if (ilk == "RG") {
      ab <- p_performRG(xyP)
    } else {
      ab <- p_performLP(xyP)
    }

    if (ab["b"] <= 0) {
      # no horizontal or downward slope
      ab["b"] <- 1e-12
    }

    return(ab)
  }

p_performLP <-
  function (xyP, weights = NULL, hallObjective = FALSE) {
    x.low <- min(xyP[, 'X'])
    y.low <- min(xyP[, 'Y'])
    xyP[, 'X'] <- xyP[, 'X'] - x.low
    xyP[, 'Y'] <- xyP[, 'Y'] - y.low
    numRows <- nrow(xyP)
    sway <- rep(1, numRows) # leads to equal weights
    # TODO I don't think this works, ordering should be over rows not columns?
    if (!(is.null(weights))) { # so we are weighing the points
      sway <- xyP[, weights]   # the weights = data in column/mast
    }
    if (hallObjective) {
      obj <- c(1, 1 / 2) # the Hall objective
    } else {
      obj <- c(sum(sway), sum(sway * xyP[, 'X']))   # "ordinary" objective coefficients
    }

    # lhs of constraints (maybe weighted)
    leftside <- cbind(rep(1, numRows), xyP[, 'X'])
    # direction of constraint(s)
    inEq <- rep(">=", numRows)
    # rhs of constraints
    rightside <- xyP[, 'Y']

    # LP with shifted data inputs
    LP_ <- lp("min", obj, leftside, inEq, rightside)
    # coefficients wrt shifted data
    a_ <- LP_$solution[1]
    b_ <- LP_$solution[2]
    # nb: a gets measured at x = 0!
    a <- a_ - b_ * x.low + y.low

    return(c(a = a, b = b_)) # line: y = a + b x, so raw ab
  }

p_performRG <-
  function (xyP) {
    # do regression of y on x
    ab <- c(a = 0, b = 1)
    xyP <- as.data.frame(xyP)
    if (nrow(xyP) > 1) {
      LM <- lm(Y ~ X, data = xyP)
      ab <- c(a = unname(LM$coefficients[1]), b = unname(LM$coefficients[2]))
    }
    return(ab)
  }

p_intersectLineBox <-
  function (ab, scope) {
    # crossings of box with line and es

    dx <- scope[2] - scope[1]
    dy <- scope[4] - scope[3]
    a <- ab[1]
    b <- ab[2]

    if (dx * dy <= 0) {
      return(NULL)
    }
    if (b <= 0) {
      return(NULL)
    }

    miss <- FALSE # does the ab-line miss the box?
    # intersection y = a + b x with x = bb$x.low:
    wi <- c(scope[1], a + b * scope[1]) # west intersection
    # intersection y = a + b x with y = bb$y.low:
    si <- c((scope[3] - a) / b, scope[3]) # south intersection

    if (wi[1] < si[1]) {
      swi <- si
    } else {
      # pick most eastern point
      swi <- wi
    }
    if (swi[1] > scope[2]) {
      # box to north-west of line
      miss <- TRUE
      es <- 1
    }
    if (swi[2] > scope[4]) {
      # box to south-east of line
      miss <- TRUE
      es <- 0
    }

    # intersection y = a + b x with x = bb$x.high:
    ei <- c(scope[2], a + b * scope[2]) # east intersection
    # intersection y = a + b x with y = bb$y.high:
    ni <- c((scope[4] - a) / b, scope[4]) # north intersection
    # pick most western point
    if (scope[2] > (scope[4] - a) / b) {
      nei <- ni
    } else {
      nei <- ei
    }
    if (nei[1] < scope[1]) {
      # box to south-east of line
      miss <- TRUE
      es <- 0
    }
    if (nei[2] < scope[3]) {
      # box (unreasonably) to north-west of line
      miss <- TRUE
      es <- 1
    }

    if (!miss) {
      # regular case: no miss and two intersections swi and nei with box's boundary
      es <- (swi[1] - scope[1]) * dy
      es <- es + (1 / 2) *
        (nei[1] - swi[1]) *
        (2 * scope[4] - swi[2] - nei[2])
    }

    # "box-line" intersections
    blis <- list(wi = wi, si = si, ei = ei, ni = ni)

    return(list(es = es / (dx * dy), swi = swi, nei = nei, blis = blis,
                inline = c(a = a, b = b), inbox = scope))
  }

p_corners <-
  function (df, condition, outcome) {
    peers <- p_get_peers(df, condition, outcome)

    # corner
    df$corner <- df$caseNo %in% rownames(peers)

    # start
    df$start <- FALSE
    for (peer.x in peers[, 1]) {
      same.x <- df[condition] == peer.x
      lowest.y <- df[outcome] == min(df[df[condition] == peer.x, outcome])
      df$start <- df$start | same.x & lowest.y
    }

    # peer
    highest.peer.idx <- as.integer(tail(rownames(peers), n = 1))
    df$peer <- data.frame(peer = rep(highest.peer.idx, nrow(df)))
    if (nrow(peers) >= 2) {
      for (row.idx in nrow(peers):2) {
        peer.x <- peers[row.idx, 1]
        peer.idx <- as.integer(rownames(peers)[row.idx - 1])
        df$peer[df[condition] < peer.x] <- peer.idx
      }
    }

    # size
    df$size <- NA
    counts <- table(df$peer)
    caseNos <- as.integer(names(counts))
    for (i in 1:nrow(counts)) {
      value <- counts[i]
      caseNo <- caseNos[i]
      df$size[df$caseNo == caseNo] <- value
    }

    return(df)
  }

p_get_peers <-
  function (df, condition, outcome, vrs = FALSE) {
    # TODO Flipping
    x <- as.matrix(df[condition])
    y <- as.matrix(df[outcome])
    loop.data <- list(x = x, y = y, flip.x = FALSE, flip.y = FALSE)
    return(p_peers(loop.data, vrs))
  }

p_fdh_nooks <-
  function (peers) {
    nooks <- matrix(nrow = 0, ncol = 2)
    for (i in 1:nrow(peers)) {
      peer <- peers[i,]
      # Add previous point if between peers
      if (i > 1) {
        nook.x <- peers[i, 1]
        nook.y <- peers[i - 1, 2]
        nooks <- rbind(nooks, c(nook.x, nook.y))
      }
      # Add the peer
      nooks <- rbind(nooks, peer)
    }

    rownames(nooks) <- NULL
    return(nooks)
  }

p_expand_fdh_nooks <-
  function (fdh_nooks) {
    if (nrow(fdh_nooks) == 0) {
      return(NULL)
    }

    fdh_nooks_expanded <- cbind(fdh_nooks, 1, 1)
    curr.col <- curr.row <- 1
    for (row.idx in seq_len(nrow(fdh_nooks_expanded))) {
      if (row.idx == 1) {
        next
      }
      prev <- fdh_nooks_expanded[row.idx - 1,]
      curr <- fdh_nooks_expanded[row.idx,]
      if (prev[1] != curr[1]) {
        curr.row <- curr.row + 1
      }
      if (prev[2] != curr[2]) {
        curr.col <- curr.col + 1
      }
      fdh_nooks_expanded[row.idx, 3] <- curr.row
      fdh_nooks_expanded[row.idx, 4] <- curr.col
    }

    colnames(fdh_nooks_expanded) <- c("x", "y", "col", "row")
    return(fdh_nooks_expanded)
  }

p_crRank <-
  function (x) {
    # x: numericals to be ranked in col/row
    d <- diff(sort(x))
    d <- ifelse(d == 0, 0, 1)
    cd <- cumsum(d)   # my rank: cumulative sum
    r <- c(1, 1 + cd) # compensate for the one value lost due to differing
  }

p_addColRow2xyPoints <-
  function (xy) {
    if (nrow(xy) == 0) {
      return(NULL)
    }

    # possibly changes order of rows
    xy <- xy[order(xy$x, xy$y),]
    xy$col <- p_crRank(xy$x) # add column data
    xy$row <- p_crRank(xy$y) # add row data

    return(xy) # Warning: the order of rows possibly changed
  }

p_closingBr <-
  function (nooks, scope) {
    ceil <- xceil <- nooks
    peers_scope <- c(min(nooks[, 1]), max(nooks[, 1]), min(nooks[, 2]), max(nooks[, 2]))

    # projections
    if (peers_scope[3] > scope[3]) {
      # nooks does not meet the bottom (south side) of scope
      xceil <- rbind(c(peers_scope[1], scope[3]), xceil)
      if (peers_scope[1] > scope[1]) {
        ceil <- rbind(c(peers_scope[1], scope[3]), ceil)
      }
    }
    if (peers_scope[2] < scope[2]) {
      # nooks does not meet the east side of bb
      xceil <- rbind(xceil, c(scope[2], peers_scope[4]))
      if (peers_scope[4] < scope[4]) {
        ceil <- rbind(ceil, c(scope[2], peers_scope[4]))
      }
    }

    # add sw and ne corners of br (if not already there):
    if (length(xceil[xceil[, 1] == scope[1] & xceil[, 2] == scope[3],]) == 0) {
      # no sw so add
      xceil <- rbind(c(scope[1], scope[3]), xceil)
    }
    if (length(xceil[xceil[, 1] == scope[2] & xceil[, 2] == scope[4],]) == 0) {
      # no ne so add
      xceil <- rbind(xceil, c(scope[2], scope[4]))
    }

    return(list(ceil = ceil, xceil = xceil))
  }

p_creduceCeil <-
  function (nooks, scope) {
    south <- nooks[nooks[, 1] == scope[3],]
    west <- nooks[nooks[, 1] == scope[1],]
    if (length(south) == (2 * ncol(nooks)) || length(west) == (2 * ncol(nooks))) {
      # delete south-west (1st_pnt)
      nooks <- matrix(nooks[-1,], ncol = ncol(nooks))
    }

    north <- nooks[nooks[, 1] == scope[4],]
    east <- nooks[nooks[, 1] == scope[2],]
    if (length(north) == (2 * ncol(nooks)) || length(east) == (2 * ncol(nooks))) {
      # delete north-east
      nooks <- matrix(nooks[1:(nrow(nooks) - 1),], ncol = ncol(nooks))
    }

    # TODO I'm not sure this is correct: shouldn't this be the 'peers' scope?
    if (nrow(nooks) == 1) {
      # TODO Why not always?
      if ((nooks[1, 1] == scope[1]) & (nooks[1, 2] == scope[4])) {
        nooks <- NULL
      }
    }
    return(nooks)
  }

p_makeStp <-
  function (peers, scope) {
    if (nrow(peers) == 0) {
      xceil <- matrix(c(scope[1], scope[1], scope[2],
                        scope[3], scope[4], scope[4]), ncol = 2)
      ceilings <- list(ceil = NULL, xceil = xceil)
      return(ceilings)
    }

    nooks <- p_fdh_nooks(peers)
    ceilings <- p_closingBr(nooks, scope)
    ceil <- p_expand_fdh_nooks(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_expand_fdh_nooks(ceilings$xceil)

    return(list(ceil = ceil, xceil = xceil))
  }

p_makeVrs <-
  function (vrs_peers, scope) {
    ceilings <- p_closingBr(vrs_peers, scope)
    ceil <- p_expand_fdh_nooks(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_expand_fdh_nooks(ceilings$xceil)

    return(list(ceil = ceil, xceil = xceil))
  }

p_doLinCeil <-
  function (xyP, scope, ilk = "RG") {
    # determine line and bounding rectangle from xy points in xyP

    # we now have the raw values for a and b
    ab <- p_abCeil4Data(xyP, ilk = ilk)
    iLB <- p_intersectLineBox(ab, scope) # non-inflated
    # intersections: south-west (swi) and north-east (nei) intersections
    swi <- unname(iLB$swi) # on boundary of BB as both ab and BR derive from the same data
    nei <- unname(iLB$nei) # on boundary of BB as both ab and BR derive from the same data
    xyP <- unique(data.frame(x = c(swi[1], nei[1]), y = c(swi[2], nei[2])))
    ceilings <- p_closingBr(xyP, scope)
    # Note. For received linear cases the extended ceiling gets traced out by 4 points.

    # add col/row information on the points in the (x)ceiling:
    ceil <- p_addColRow2xyPoints(ceilings$ceil)
    ceil <- p_creduceCeil(ceil, scope)
    xceil <- p_addColRow2xyPoints(ceilings$xceil)

    deltay <- scope[4] - scope[3]
    a1 <- (ab["a"] - scope[3] + ab["b"] * scope[1]) / deltay
    b1 <- (scope[2] - scope[1]) * ab["b"] / deltay
    nrm <- c(a = a1, b = b1)

    return(ceilings = list(ceil = ceil, xceil = xceil, ab = list(raw = ab, nrm = nrm)))
  }

###############################################################################
# HERE START
###############################################################################
p_make2SidedWeb <-
  function (ceilings, scope) {
    # compute on meshes formed by the extended ceiling's break-points
    # Meshes (maze) form a partition of the bounding rectangle based on the breakpoints
    # in the extended ceiling.

    # Note: enlceil is a dataframe, not an matrix (mixed types)
    enlceil <- cbind(data.frame(ceilings$xceil), strain = "regular")

    # if enlargement then add points to xceil to enlarge it and obtain enlceil
    if (p_ENLARGEMENT_CONSTANT > 0) {
      dx <- scope[2] - scope[1]
      dy <- scope[4] - scope[3]
      extra.row <- data.frame(x = scope[1] - p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[3] - p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = 0, row = 0, strain = "enlarged")
      enlceil <- rbind(extra.row, enlceil)

      tmp <- tail(enlceil, n = 1)
      extra.row <- data.frame(x = scope[2] + p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[4] + p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = tmp[3] + 1, row = tmp[4] + 1, strain = "enlarged")
      enlceil <- rbind(enlceil, extra.row)
      enlceil[, 3:4] <- enlceil[, 3:4] + 1
      row.names(enlceil) <- NULL
    }

    ceilings$enlceil <- enlceil

    # xy (and col-row) raster with info
    ras <- p_twoSidedRaster(enlceil)
    rp <- ras$pnts

    # TODO Rename to scope.enl
    # bounding rectangle (enlarged) coordinates from xyCR (= xceil):
    br <- c(min(rp['x']), max(rp['x']), min(rp['y']), max(rp['y']))
    # TODO Rename to scopes
    BRs <- list(obr = scope, br = br)

    # Out of the raster points we build the mesh (= partition by mazes = rectangles)
    # Meshes: raster points (x, y) are considered as south-east points in a mesh
    # below procedure may give at first some incomplete/degenerate/non-existing
    # nw points/meshes which subsequently get removed
    colnames(rp) <- c('col', 'se_x', 'strainx', 'cardDirEW',
                      'row', 'se_y', 'strainy', 'cardDirNS', 'se_pos')

    # Now compute the meshes with current points in rp considered as se-points.
    r1 <- rp
    r1 <- cbind(r1, rp[, 'row'] - 1, rp[, 'col'] + 1)
    colnames(r1) <- c('col.y', 'nw_x', 'strainx', 'cardDirEW',
                      'row.y', 'nw_y', 'strainy', 'cardDirNS',
                      'nw_pos', 'row', 'col')

    # Remove "invalid" meshes - having corners or row-col with NA-values:
    rp <- rp[complete.cases(rp[, c("row", "col")]),]
    r1 <- r1[complete.cases(r1[, c("row", "col", "row.y", "col.y", "nw_x")]),]
    rp$insol <- NA
    rp$inver <- NA

    ribbon_db <- dbConnect(SQLite(), ":memory:")

    dbWriteTable(ribbon_db, "rp", rp)
    dbWriteTable(ribbon_db, "r1", r1)
    dbWriteTable(ribbon_db, "colrangebyrow", data.frame(ras$colrangebyrow))
    dbWriteTable(ribbon_db, "rowrangebycol", data.frame(ras$rowrangebycol))

    dbExecute(ribbon_db, CREATE_MAZE,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))

    # Deviance areas to north-west
    dbExecute(ribbon_db, UPDATE_MAZE_1)

    # Deviance areas to south-east
    dbExecute(ribbon_db, UPDATE_MAZE_2)

    # Do lookup stuff (lm = lookup_maze)
    dbExecute(ribbon_db, UPDATE_MAZE_3)

    dbExecute(ribbon_db, UPDATE_MAZE_4,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))

    dbExecute(ribbon_db, UPDATE_MAZE_5)

    dbExecute(ribbon_db, UPDATE_RP)

    maze <- dbReadTable(ribbon_db, "maze")
    ras$pnts <- dbReadTable(ribbon_db, "rp")
    dbDisconnect(ribbon_db)

    # effect size:
    se_maze <- maze[maze['s_row'] == min(maze['s_row']) & maze['e_col'] == max(maze['e_col']),]
    nces <- 1 - se_maze$se_insol

    dossier <- list(message = "", portrayal = "")
    if (nrow(maze) == 1) {
      if (maze$stance != "bisect") { # ceiling fully in bb boundary
        if (maze$stance == "nw") {
          TM1 <- "\n Single mesh: Empty space goes to south-east of bounding rectangle! \n"
          TM2 <- "There is no feasible space: in-veracity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)veracity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_feasible_space"
        } else { # maze$stance == "se"
          TM1 <- "\n Single mesh: Feasible point at the north-west of bounding rectangle! \n"
          TM2 <- "There is no empty space and in-solidity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)solidity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_empty_space"
        }
      }
    }

    return(list(ceilings = ceilings, raster = ras, BRs = BRs, enlarged = p_ENLARGEMENT_CONSTANT > 0,
                maze = maze, nces = nces, dossier = dossier))
  }

p_twoSidedRaster <-
  function (enlceil) {
    # raster from xy-column-rows (later the meshes)

    col.max <- max(enlceil['col'])
    row.max <- max(enlceil['row'])

    rangeby <- NULL
    for (row_idx in seq_len(nrow(enlceil))) {
      row <- enlceil[row_idx,]
      rows <- enlceil[enlceil['row'] == row$row,]
      cols <- enlceil[enlceil['col'] == row$col,]
      rangeby <- rbind(rangeby, c(row = row$row, col = row$col,
                                  x.low = min(rows['x']), x.high = max(rows['x']),
                                  y.low = min(cols['y']), y.high = max(cols['y']),
                                  col.low = min(rows['col']), col.high = max(rows['col']),
                                  row.low = min(cols['row']), row.high = max(cols['row'])))
    }
    rownames(rangeby) <- NULL
    colrangebyrow <- rangeby[!duplicated(rangeby[, 'row']),
                             c('row', 'col.low', 'col.high', 'x.low', 'x.high')]
    rowrangebycol <- rangeby[!duplicated(rangeby[, 'col']),
                             c('col', 'row.low', 'row.high', 'y.low', 'y.high')]

    tmp <- enlceil
    tmp$cardDirEW <- ifelse(tmp$col == 1, "west", ifelse(tmp$col == col.max, "east", "inner"))
    tmp$cardDirNS <- ifelse(tmp$row == 1, "south", ifelse(tmp$row == row.max, "north", "inner"))
    distinct_col <- tmp[!duplicated(tmp$col), c('col', 'x', 'strain', 'cardDirEW')]
    colnames(distinct_col) <- c('col', 'x', 'strainx', 'cardDirEW')
    distinct_row <- tmp[!duplicated(tmp$row), c('row', 'y', 'strain', 'cardDirNS')]
    colnames(distinct_row) <- c('row', 'y', 'strainy', 'cardDirNS')

    pnts <- merge(distinct_col, distinct_row, by = NULL)
    pnts <- pnts[order(pnts$col, pnts$row),]
    pnts$hpos <- ""
    for (row_idx in seq_len(nrow(pnts))) {
      row <- pnts[row_idx,]
      tmp <- colrangebyrow[colrangebyrow[, "row"] == row$row]
      pnts[row_idx, 'hpos'] <- ifelse(row$col >= tmp[2] & row$col <= tmp[3], 'on',
                                      ifelse(row$col < tmp[2], 'nw', 'se'))
    }
    rownames(pnts) <- NULL

    return(list(pnts = pnts,
                colrangebyrow = colrangebyrow,
                rowrangebycol = rowrangebycol,
                Mcol = col.max,
                Mrow = row.max))
  }

p_sv4Points <-
  function (xyP, wb) {
    # compute sv (raw + normalized (net)) for points

    mz <- wb$maze
    br <- wb$BRs$br # (possibly) inflated bounding rectangle
    rp <- wb$raster$pnts
    # no mixing with x and y in xyP
    colnames(rp) <- c('col', 'rx', 'strainx', 'cardDirEW', 'row', 'ry',
                      'strainy', 'cardDirNS', 'hpos', 'insol', 'inver')

    # We cannot rely on xyP having caseNo as a variable identifying a row.
    xyP <- cbind(pointID = paste0("point_", xyP$caseNo), xyP)
    xyOri <- xyP # keep (almost) original input data in xyO

    xyP <- p_sv4PointsSql(xyP, mz, rp, br)

    # names <- c('pointID', 'x', 'y', 'ezon', 'feas', 'idRE', 'idRW', 'insol', 'inver')
    names <- c('caseNo', 'pointID', 'x', 'y', 'insol', 'inver')
    xyQ <- xyP[, names]

    # add normalized (in-)solidity / (in-)veracity
    nces <- wb$nces

    xyQ$insol_net <- p_nciRaw2Norm(xyQ$insol, nces, solidity = TRUE)
    xyQ$inver_net <- p_nciRaw2Norm(xyQ$inver, nces, solidity = FALSE)
    xyQ$sol_net <- 1 - xyQ$insol_net
    xyQ$ver_net <- 1 - xyQ$inver_net

    xyR <- merge(x = xyOri, y = xyQ, by = "pointID", all.x = TRUE)
    xyR <- xyR[order(xyR$pointID),]

    res <- list(compacts = xyQ, specifics = xyR)

    return(res)
  }

p_sv4PointsSql <-
  function (xyP, mz, rp, br) {
    # TODO Check what's needed
    xyP$ezon <- FALSE
    xyP$feas <- FALSE
    xyP$ver_num <- NA
    xyP$ver_den <- NA
    xyP$sol_num <- NA
    xyP$sol_den <- NA
    xyP$ver_devi <- NA
    xyP$sol_devi <- NA
    xyP$insol <- NA
    xyP$inver <- NA
    xyP$nw <- FALSE
    xyP$se <- FALSE

    sv_db <- dbConnect(SQLite(), ":memory:")
    dbWriteTable(sv_db, "xyP", xyP)
    dbWriteTable(sv_db, "rp", rp)
    dbWriteTable(sv_db, "mz", mz)

    dbExecute(sv_db, CREATE_STANCE)
    dbExecute(sv_db, CREATE_E1)
    dbExecute(sv_db, UPDATE_XYP_1)
    dbExecute(sv_db, UPDATE_XYP_2,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))
    dbExecute(sv_db, UPDATE_XYP_3, ,
              params = list(br1 = br[1], br2 = br[2], br3 = br[3], br4 = br[4]))
    dbExecute(sv_db, UPDATE_XYP_4)

    xyP <- dbReadTable(sv_db, "xyP")

    dbDisconnect(sv_db)

    return(xyP)
  }

###############################################################################
# HERE END
###############################################################################

p_make2SidedWebOrg <-
  function (ceilings, scope) {
    # compute on meshes formed by the extended ceiling's break-points
    # Meshes (maze) form a partition of the bounding rectangle based on the breakpoints
    # in the extended ceiling.

    # Note: enlceil is a dataframe, not an matrix (mixed types)
    enlceil <- cbind(data.frame(ceilings$xceil), strain = "regular")

    # if enlargement then add points to xceil to enlarge it and obtain enlceil
    enlarged <- FALSE
    if (p_ENLARGEMENT_CONSTANT > 0) {
      enlarged <- TRUE
      dx <- scope[2] - scope[1]
      dy <- scope[4] - scope[3]
      extra.row <- data.frame(x = scope[1] - p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[3] - p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = 0, row = 0, strain = "enlarged")
      enlceil <- rbind(extra.row, enlceil)


      tmp <- tail(enlceil, n = 1)
      extra.row <- data.frame(x = scope[2] + p_ENLARGEMENT_CONSTANT * dx / 2,
                              y = scope[4] + p_ENLARGEMENT_CONSTANT * dy / 2,
                              col = tmp[3] + 1, row = tmp[4] + 1, strain = "enlarged")
      enlceil <- rbind(enlceil, extra.row)
      enlceil[, 3:4] <- enlceil[, 3:4] + 1
      row.names(enlceil) <- NULL
    }

    ceilings$enlceil <- enlceil
    # xy (and col-row) raster with info
    ras <- p_twoSidedRasterOrg(enlceil)
    rp <- ras$pnts

    # TODO Rename to scope.enl
    # bounding rectangle (enlarged) coordinates from xyCR (= xceil):
    br <- c(min(rp['x']), max(rp['x']), min(rp['y']), max(rp['y']))

    # TODO Rename to scopes
    BRs <- list(obr = scope, br = br)

    # Out of the raster points we build the mesh (= partition by mazes = rectangles)
    # Meshes: raster points (x, y) are considered as south-east points in a mesh
    # below procedure may give at first some incomplete/degenerate/non-existing
    # nw points/meshes which subsequently get removed
    colnames(rp) <- c('col', 'se_x', 'strainx', 'cardDirEW',
                      'row', 'se_y', 'strainy', 'cardDirNS', 'se_pos')

    # Now compute the meshes with current points in rp considered as se-points.
    r1 <- rp
    r1 <- cbind(r1, rp[, 'row'] - 1, rp[, 'col'] + 1)
    colnames(r1) <- c('col.y', 'nw_x', 'strainx', 'cardDirEW',
                      'row.y', 'nw_y', 'strainy', 'cardDirNS',
                      'nw_pos', 'row', 'col')

    mesh <- merge(rp, r1, by = c("row", "col"))
    # Remove "invalid" meshes - having corners or row-col with NA-values:
    mesh <- mesh[complete.cases(mesh[, c("row.y", "col.y", "col", "row", "nw_x")]),]
    mesh <- mesh[order(mesh$col),]

    # fixing up remaining, valid, meshes, such as
    # using only the 4 ordinal directions (e, w, s, n) for the mesh:
    maze <- cbind(mesh['col'], mesh['se_x'], mesh['strainx.x'], mesh['cardDirEW.x'],
                  mesh['row'], mesh['se_y'], mesh['strainy.x'], mesh['cardDirNS.x'], mesh['se_pos'],
                  mesh['col.y'], mesh['nw_x'], mesh['strainx.y'], mesh['cardDirEW.y'],
                  mesh['row.y'], mesh['nw_y'], mesh['strainy.y'], mesh['cardDirNS.y'], mesh['nw_pos'])
    colnames(maze) <- c('e_col', 'e_x', 'strainx.x', 'cardDirEW.x',
                        's_row', 's_y', 'strainy.x', 'cardDirNS.x', 'se_pos',
                        'w_col', 'w_x', 'strainx.y', 'cardDirEW.y',
                        'n_row', 'n_y', 'strainy.y', 'cardDirNS.y', 'nw_pos')
    maze$onEdge <- 'enlarged' == maze['strainx.x'] |
      'enlarged' == maze['strainx.y'] |
      'enlarged' == maze['strainy.x'] |
      'enlarged' == maze['strainy.y']
    maze[c('strainx.x', 'strainx.y', 'strainy.x', 'strainy.y')] <- NULL

    # Note. Points (e_x, s_y) are south and east lines in their maze (cell)
    # and (w_x, n_y) are the north and west lines in that maze, and so on.
    # Note that we only track positions on the nw and se intersection points.
    # Such is sufficient (for our purposes) as the ceiling runs sw-ne.
    # w_x ~ x.low; e_x ~ x.high; s_y ~y.low; n_y ~ y.high
    # some additional info on mazes for easy identification:
    maze <- maze[with(maze, order(maze$e_x, maze$s_y)),]
    maze <- cbind(mazeID = paste0('maze_', seq.int(nrow(maze))), maze)

    # position (stance), south-east or north-west, or bisected, of meshes wrt ceiling:
    maze$stance <- as.vector(
      ifelse(maze['nw_pos'] == "nw" & maze['se_pos'] == "se", "bisect",
             ifelse(maze['nw_pos'] == "on" | maze['nw_pos'] == "se",
                    "se", "nw")))

    # Areas of meshes (with bi-sect counting for half):
    maze$area <- as.vector(
      ifelse(maze$stance == "bisect",
             (maze[['e_x']] - maze[['w_x']]) * (maze[['n_y']] - maze[['s_y']]) / 2,
             (maze[['e_x']] - maze[['w_x']]) * (maze[['n_y']] - maze[['s_y']])))
    # Do away with irrelevant 0-area meshes, probably redundant, but ... safety first:
    maze <- maze[maze['area'] > 0,]

    # Unconditional areas to north-west (for each of the 4 corners of a mesh)
    # Note. The same point, when a corner in 4 meshes, gets calculated 4 times
    maze$se_NW_den <- unlist((maze['e_x'] - br[1]) * (br[4] - maze['s_y']))
    maze$nw_NW_den <- unlist((maze['w_x'] - br[1]) * (br[4] - maze['n_y']))
    maze$sw_NW_den <- unlist((maze['w_x'] - br[1]) * (br[4] - maze['s_y']))
    maze$ne_NW_den <- unlist((maze['e_x'] - br[1]) * (br[4] - maze['n_y']))

    # and total (unconditional) area to the south-east of each point in a mesh:
    maze$se_SE_den <- unlist((br[2] - maze['e_x']) * (maze['s_y'] - br[3]))
    maze$nw_SE_den <- unlist((br[2] - maze['w_x']) * (maze['n_y'] - br[3]))
    maze$sw_SE_den <- unlist((br[2] - maze['w_x']) * (maze['s_y'] - br[3]))
    maze$ne_SE_den <- unlist((br[2] - maze['e_x']) * (maze['n_y'] - br[3]))

    # Now for in-solidity numerators (num) for each of the 4 corners of a mesh:
    maze <- cbind(maze, se_NW_num = NA, nw_NW_num = NA, sw_NW_num = NA, ne_NW_num = NA)

    # Deviance areas to north-west (for each of the 4 corners of a mesh)
    # Deviance = from what is expected for a point on the ceiling
    # (here the feasible area is to the NW of the point under consideration)
    # Note: inefficiency of a point may get (re-)calculated up to 4 times
    for (i in seq_len(nrow(maze))) { # stance in "se, bisect": 100%/50% in feas. space
      cl <- maze$e_col[i]; rw <- maze$s_row[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$se_NW_num[i] <- sum(tmp$area)

      cl <- maze$w_col[i]; rw <- maze$n_row[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$nw_NW_num[i] <- sum(tmp$area)

      rw <- maze$s_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$sw_NW_num[i] <- sum(tmp$area)


      rw <- maze$n_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['e_col'] <= cl &
                    maze['s_row'] >= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "se"),]
      maze$ne_NW_num[i] <- sum(tmp$area)
    }

    # # Now then for in-veracity numerators for each of the 4 corners of a maze:
    maze <- cbind(maze, se_SE_num = NA, nw_SE_num = NA, sw_SE_num = NA, ne_SE_num = NA)

    # Deviance areas to south-east (for each of the 4 corners of a maze)
    # Deviance from what is expected for a point on the ceiling
    # (here the feasible area is to the SE)
    # inefficient calculation: a point/value may get (re-)calculated up to 4 times
    for (i in seq_len(nrow(maze))) {
      rw <- maze$s_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$se_SE_num[i] <- sum(tmp$area)

      rw <- maze$n_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$nw_SE_num[i] <- sum(tmp$area)

      rw <- maze$s_row[i]; cl <- maze$w_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$sw_SE_num[i] <- sum(tmp$area)

      rw <- maze$n_row[i]; cl <- maze$e_col[i]
      tmp <- maze[maze['w_col'] >= cl &
                    maze['n_row'] <= rw &
                    (maze['stance'] == "bisect" | maze['stance'] == "nw"),]
      maze$ne_SE_num[i] <- sum(tmp$area)
    }

    # Finished with areas, now for the deviance calculations.
    # In-solidity and in-veracity ratios: circumvent division by zero.
    # As an in-solidity involves the north&west of the br, special care
    # (to avoid cancellation/division by zero) is needed for meshes
    # touching (osculating) the west- or north-side of the br
    maze$nw_loc <- as.character(
      ifelse((maze['w_x'] == br[1]) & (maze['n_y'] == br[4]), "north-west",
             ifelse(maze['w_x'] == br[1], "west",
                    ifelse(maze['n_y'] == br[4], "north", "other"))))
    maze$se_loc <- as.character(
      ifelse((maze['e_x'] == br[2]) & (maze['s_y'] == br[3]), "south-east",
             ifelse(maze['e_x'] == br[2], "east",
                    ifelse(maze['s_y'] == br[3], "south", "other"))))

    # vx: west-east x-value of up-going (going north(-east) ~ vertical) ceiling
    # per mesh, id-ed thru south (se-sw) row: vx = x where the ceiling crosses to ne.
    lookup_vx <- cbind(s_row = ras$colrangebyrow[, 'row'],
                       vx = ras$colrangebyrow[, 'x.high'],
                       vc = ras$colrangebyrow[, 'col.high'])
    maze <- merge(maze, lookup_vx, by = "s_row", keep = NULL)

    # the point (s_row,vc) is where the ceiling crosses the line determined
    # by the south of the mesh towards the line determined by the north of the mesh,
    # in particular (se_row,vc) is on the ceiling.

    # we need the slope for the ceiling running from the south-line to the north-line of mesh
    # if the point (se_row,vc) is on the east of the bb the slope, vs, is infinite,
    # meaning the slope is vertical, otherwise (se_row,vc) is the sw-point of a mesh in which
    # the ceiling can be assumed to run for the south-line to the north-line.
    # Search for the mesh with sw-corner (vc, se_row = sw_row):
    lookup_maze <- cbind(maze['w_col'],
                         maze['w_x'], maze['s_row'], maze['s_y'], maze['e_col'],
                         maze['e_x'], maze['n_row'], maze['n_y'], maze['stance'])
    lookup_maze <- lookup_maze[with(lookup_maze, order(lookup_maze$w_col)),]
    colnames(lookup_maze) <- c('vc', 'w_x.v', 's_row', 's_y.v', 'e_col.v',
                               'e_x.v', 'n_row.v', 'n_y.v', 'stance.v')
    xmaze <- merge(x = maze, y = lookup_maze, by = c('vc', 's_row'), all.x = TRUE)
    xmaze <- xmaze[with(xmaze, order(xmaze$e_col, xmaze$w_col)),]

    for (i in seq_len(nrow(xmaze))) {
      row <- xmaze[i,]
      tmp <- ifelse(row['stance.v'] == "bisect",
                    (row['n_y.v'] - row['s_y.v']) / (row['e_x.v'] - row['w_x.v']),
                    ifelse(is.na(row['stance.v']), Inf, Inf))
      xmaze$vslope[i] <- as.numeric(tmp)
    }
    # cleaning up superfluous values (for efficiency):
    xmaze[, colnames(xmaze)[grepl(".*\\.v", colnames(xmaze))]] <- NULL

    # Now, the analogous computation in the y-direction:
    # horizontal ceiling in south-north (hy) of maze:
    # Note. We will use south-west [] as (join_by) reference point.
    lookup_hy <- cbind(w_col = ras$rowrangebycol[, 'col'],
                       hy = ras$rowrangebycol[, 'y.high'],
                       hr = ras$rowrangebycol[, 'row.high'])
    xmaze <- merge(x = xmaze, y = lookup_hy, by = "w_col",
                   keep = NULL, all.x = TRUE)

    colnames(lookup_maze) <- c('w_col', 'w_x.h', 'hr', 's_y.h', 'e_col.h',
                               'e_x.h', 'n_row.h', 'n_y.h', 'stance.h')
    xymaze <- merge(xmaze, lookup_maze, by = c("hr", "w_col"), keep = NULL, all.x = TRUE)

    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]
      tmp <- ifelse(row['stance.h'] == "bisect",
                    (row['n_y.h'] - row['s_y.h']) / (row['e_x.h'] - row['w_x.h']),
                    ifelse(is.na(row['stance.h']), 0, 0))
      xymaze$hslope[i] <- as.numeric(tmp)
    }
    # cleaning up superfluous values (for efficiency):
    xymaze[, colnames(xymaze)[grepl(".*\\.h", colnames(xymaze))]] <- NULL

    # Note. Computing (in-)solidity: looking for feasible space to nw.
    # So, of special concern are meshes touching the bb on the north and/or west.
    # Note: when !(stance %in% c("bisect", "nw") it has to be "se".
    # So, fall-through cases stance %in% c("bisect", "nw") are "se" cases.
    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]

      sw_insol <- ifelse(row['stance'] %in% c("bisect", "nw"), 0,
                         ifelse(row['nw_loc'] == "west",
                                (row['hy'] - row['s_y']) / (br[4] - row['s_y']),
                                row['sw_NW_num'] / row['sw_NW_den']))
      xymaze$sw_insol[i] <- as.numeric(sw_insol)

      ne_insol <- ifelse(row['stance'] %in% c("bisect", "nw"), 0,
                         ifelse(row['nw_loc'] == "north",
                                (row['e_x'] - row['w_x']) / (row['e_x'] - br[1]),
                                row['ne_NW_num'] / row['ne_NW_den']))
      xymaze$ne_insol[i] <- as.numeric(ne_insol)

      nw_insol <- ifelse(row['stance'] %in% c("bisect", "nw") | (row['nw_NW_den'] == 0),
                         0,
                         row['nw_NW_num'] / row['nw_NW_den'])
      xymaze$nw_insol[i] <- as.numeric(nw_insol)

      xymaze$se_insol[i] <- as.numeric(row['se_NW_num'] / row['se_NW_den'])
    }

    for (i in seq_len(nrow(xymaze))) {
      row <- xymaze[i,]

      sw_inver <- ifelse(row['stance'] %in% c("bisect", "se"), 0,
                         ifelse(row['se_loc'] == "south",
                                (row['e_x'] - row['w_x']) / (br[2] - row['w_x']),
                                row['sw_SE_num'] / row['sw_SE_den']))
      xymaze$sw_inver[i] <- as.numeric(sw_inver)

      ne_inver <- ifelse(row['stance'] %in% c("bisect", "se"), 0,
                         ifelse(row['se_loc'] == "east",
                                (row['n_y'] - row['s_y']) / (row['n_y'] - br[3]),
                                row['ne_SE_num'] / row['ne_SE_den']))
      xymaze$ne_inver[i] <- as.numeric(ne_inver)

      xymaze$nw_inver[i] <- as.numeric(row['nw_SE_num'] / row['nw_SE_den'])

      se_inver <- ifelse(
        row['stance'] %in% c("bisect", "se") | row['se_SE_den'] == 0,
        0, row['se_SE_num'] / row['se_SE_den'])
      xymaze$se_inver[i] <- as.numeric(se_inver)
    }

    xymaze$se_inacc <- pmax(xymaze$se_insol, xymaze$se_inver)
    xymaze$sw_inacc <- pmax(xymaze$sw_insol, xymaze$sw_inver)
    xymaze$nw_inacc <- pmax(xymaze$nw_insol, xymaze$nw_inver)
    xymaze$ne_inacc <- pmax(xymaze$ne_insol, xymaze$ne_inver)

    ras$pnts <- p_doInsolInverLookUpOrg(ras$pnts, xymaze)

    # effect size:
    se_maze <- xymaze[xymaze['s_row'] == min(xymaze['s_row']) &
                        xymaze['e_col'] == max(xymaze['e_col']),]
    nces <- 1 - se_maze$se_insol

    dossier <- list(message = "", portrayal = "")
    if (nrow(xymaze) == 1) {
      if (xymaze$stance != "bisect") { # ceiling fully in bb boundary
        if (xymaze$stance == "nw") {
          TM1 <- "\n Single mesh: Empty space goes to south-east of bounding rectangle! \n"
          TM2 <- "There is no feasible space: in-veracity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)veracity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_feasible_space"
        } else { # xymaze$stance == "se"
          TM1 <- "\n Single mesh: Feasible point at the north-west of bounding rectangle! \n"
          TM2 <- "There is no empty space and in-solidity equals 1 (except on ceiling). \n"
          TM3 <- "Any iso-(in-)solidity-line will be meaningless! \n \n"
          dossier$message <- paste0(TM1, TM2, TM3)
          dossier$portrayal <- "no_empty_space"
        }
      }
    }

    xymaze$nesw_sqr <- ifelse(
      xymaze$w_x == br[1] &
        xymaze$s_y == br[3] &
        xymaze$stance != "bisect",
      "sw_sqr",
      ifelse(
        xymaze$e_x == br[2] &
          xymaze$n_y == br[4] &
          xymaze$stance != "bisect",
        "ne_sqr",
        "other"))

    return(list(ceilings = ceilings, raster = ras, BRs = BRs, enlarged = enlarged,
                maze = xymaze, nces = nces, dossier = dossier))
  }

p_twoSidedRasterOrg <-
  function (enlceil) {
    # raster from xy-column-rows (later the meshes)
    # TODO Rename everything

    col.max <- max(enlceil['col'])
    row.max <- max(enlceil['row'])

    colx <- enlceil[c('col', 'x', 'strain')]
    distinct_col <- NULL
    for (i in seq_len(nrow(colx))) {
      row <- colx[i,]
      cardDirEW <- ifelse(row[1] == 1, "west", ifelse(row[1] == col.max, "east", "inner"))
      if (!(row[1] %in% distinct_col[, 1])) {
        distinct_col <- rbind(distinct_col, cbind(row, cardDirEW = cardDirEW))
      }
    }
    colnames(distinct_col) <- c('col', 'x', 'strainx', 'cardDirEW')

    rowy <- enlceil[c('row', 'y', 'strain')]
    distinct_row <- NULL
    for (i in seq_len(nrow(rowy))) {
      row <- rowy[i,]
      cardDirNS <- ifelse(row[1] == 1, "south", ifelse(row[1] == row.max, "north", "inner"))
      if (!(row[1] %in% distinct_row[, 1])) {
        distinct_row <- rbind(distinct_row, cbind(row, cardDirNS))
      }
    }
    colnames(distinct_row) <- c('row', 'y', 'strainy', 'cardDirNS')

    pnts <- NULL
    for (i in seq_len(nrow(distinct_col))) {
      for (j in seq_len(nrow(distinct_row))) {
        new.row <- cbind(distinct_col[i,], distinct_row[j,])
        pnts <- rbind(pnts, new.row)
      }
    }

    mincol <- pnts[pnts['col'] == 1,]
    colnames(mincol) <- paste0('m', colnames(mincol))
    maxcol <- pnts[pnts['col'] == max(pnts['col']),]
    colnames(maxcol) <- paste0('M', colnames(maxcol))
    names(maxcol)[names(maxcol) == 'Mrow'] <- 'mrow'
    horiz <- merge(mincol, maxcol, by = "mrow")

    minrow <- pnts[pnts['row'] == 1,]
    colnames(minrow) <- paste0('m', colnames(minrow))
    maxrow <- pnts[pnts['row'] == max(pnts['row']),]
    colnames(maxrow) <- paste0('M', colnames(maxrow))
    names(maxrow)[names(maxrow) == 'Mcol'] <- 'mcol'
    verti <- merge(minrow, maxrow, by = "mcol")
    # Note. horiz and verti will only be used for producing a grid in plots.

    colrangebyrow <- NULL
    for (row in seq_along(unique(enlceil[, 'row']))) {
      rows <- enlceil[enlceil['row'] == row,]
      tmp <- c(row, min(rows['col']), max(rows['col']), min(rows['x']), max(rows['x']))
      colrangebyrow <- rbind(colrangebyrow, tmp)
    }
    rownames(colrangebyrow) <- NULL
    colnames(colrangebyrow) <- c('row', 'col.low', 'col.high', 'x.low', 'x.high')

    rowrangebycol <- NULL
    for (col in seq_along(unique(enlceil[, 'col']))) {
      cols <- enlceil[enlceil['col'] == col,]
      # TODO Why is the order different?
      tmp <- c(col, min(cols['y']), max(cols['y']), min(cols['row']), max(cols['row']))
      rowrangebycol <- rbind(rowrangebycol, tmp)
    }
    rownames(rowrangebycol) <- NULL
    colnames(rowrangebycol) <- c('col', 'y.low', 'y.high', 'row.low', 'row.high')

    col.new <- NULL
    for (i in seq_len(nrow(pnts))) {
      col <- pnts[i,][['col']]
      row <- pnts[i,][['row']]
      low <- colrangebyrow[colrangebyrow[, "row"] == row][2]
      high <- colrangebyrow[colrangebyrow[, "row"] == row][3]
      tmp <- ifelse(col >= low & col <= high, 'on', ifelse(col < low, 'nw', 'se'))
      col.new <- c(col.new, tmp)
    }
    pnts$hpos <- col.new
    # TODO
    rownames(pnts) <- NULL

    return(list(pnts = pnts,
                colrangebyrow = colrangebyrow,
                rowrangebycol = rowrangebycol,
                horiz = horiz,
                verti = verti,
                Mcol = col.max,
                Mrow = row.max))
  }

p_doInsolInverLookUpOrg <-
  function (pnts, xymaze) {
    pnts$insol <- NULL
    pnts$inver <- NULL
    for (i in seq_len(nrow(pnts))) {
      row <- pnts[i,]

      if (row['cardDirEW'] != "east" & row['cardDirNS'] != "north") {
        tmp <- xymaze[xymaze['w_x'] == row[['x']] & xymaze['s_y'] == row[['y']],]
        pnts$insol[i] <- tmp$sw_insol
        pnts$inver[i] <- tmp$sw_inver
      }
      else if (row['cardDirEW'] != "east" & row['cardDirNS'] == "north") {
        tmp <- xymaze[xymaze['w_x'] == row[['x']] & xymaze['n_y'] == row[['y']],]
        pnts$insol[i] <- tmp$nw_insol
        pnts$inver[i] <- tmp$nw_inver
      }
      else if (row['cardDirEW'] == "east" & row['cardDirNS'] != "south") {
        tmp <- xymaze[xymaze['e_x'] == row[['x']] & xymaze['n_y'] == row[['y']],]
        pnts$insol[i] <- tmp$ne_insol
        pnts$inver[i] <- tmp$ne_inver
      }
      else if (row['cardDirEW'] == "east" & row['cardDirNS'] == "south") {
        tmp <- xymaze[xymaze['e_x'] == row[['x']] & xymaze['s_y'] == row[['y']],]
        pnts$insol[i] <- tmp$se_insol
        pnts$inver[i] <- tmp$se_inver
      }
      else {
        pnts$insol[i] <- NA
        pnts$inver[i] <- NA
      }
    }

    return(pnts)
  }

p_sv4PointsOrg <-
  function (xyP, wb) {
    # compute sv (raw + normalized (net)) for points

    mz <- wb$maze
    br <- wb$BRs$br # (possibly) inflated bounding rectangle
    rp <- wb$raster$pnts
    # no mixing with x and y in xyP
    colnames(rp) <- c('col', 'rx', 'strainx', 'cardDirEW', 'row', 'ry',
                      'strainy', 'cardDirNS', 'hpos', 'insol', 'inver')

    # We cannot rely on xyP having caseNo as a variable identifying a row.
    xyP <- cbind(pointID = paste0("point_", xyP$caseNo), xyP)
    xyOri <- xyP # keep (almost) original input data in xyO

    for (i in seq_len(nrow(xyP))) {
      tmp <- p_soveLookupCompsOrg(xyP[i,], rp, mz, br)
      #xyP$idRE[i] <- tmp$idRectE
      #xyP$idRW[i] <- tmp$idRectW
      #xyP$ezon[i] <- tmp$ezon
      #xyP$feas[i] <- tmp$feas
      xyP$insol[i] <- tmp$insol
      xyP$inver[i] <- tmp$inver
    }

    # names <- c('pointID', 'x', 'y', 'ezon', 'feas', 'idRE', 'idRW', 'insol', 'inver')
    names <- c('caseNo', 'pointID', 'x', 'y', 'insol', 'inver')
    xyQ <- xyP[, names]

    # add normalized (in-)solidity / (in-)veracity
    nces <- wb$nces

    xyQ$insol_net <- p_nciRaw2Norm(xyQ$insol, nces, solidity = TRUE)
    xyQ$inver_net <- p_nciRaw2Norm(xyQ$inver, nces, solidity = FALSE)
    xyQ$sol_net <- 1 - xyQ$insol_net
    xyQ$ver_net <- 1 - xyQ$inver_net

    xyR <- merge(x = xyOri, y = xyQ, by = "pointID", all.x = TRUE)
    xyR <- xyR[order(xyR$pointID),]

    res <- list(compacts = xyQ, specifics = xyR)

    return(res)
  }

p_soveLookupCompsOrg <-
  function (row, rp, mz, br) { # sove by lookup and comps
    x <- unlist(row['x'])
    y <- unlist(row['y'])

    xyS <- mz[x >= mz$w_x &
                x <= mz$e_x &
                y >= mz$s_y &
                y <= mz$n_y,]
    xyS <- xyS[order(xyS$e_x, xyS$s_y),]

    Stands <- xyS$stance
    nw <- is.element("nw", Stands)
    se <- is.element("se", Stands)

    # TODO
    # E1 <- slice(xyS[order(-xyS$e_x, xyS$s_y),], 1)
    E1 <- head(xyS, 1)

    feas <- ifelse(
      se, TRUE, ifelse(
        E1$stance == "bisect",
        (y - E1['s_y']) * (E1['e_x'] - E1['w_x']) <= (x - E1['w_x']) * (E1['n_y'] - E1['s_y']),
        FALSE))

    ezon <- ifelse(
      nw, TRUE, ifelse(
        E1$stance == "bisect",
        (y - E1['s_y']) * (E1['e_x'] - E1['w_x']) >= (x - E1['w_x']) * (E1['n_y'] - E1['s_y']),
        FALSE))

    insol <- inver <- NA
    luRP <- rp[rp['rx'] == x & rp['ry'] == y,]
    if (nrow(luRP) > 0) {
      inver <- luRP$inver
      insol <- luRP$insol
    }
    inver <- ifelse(
      !is.na(inver), inver, ifelse(
        feas, 0,
        p_deviOrg(x, y, E1['w_x'], E1['s_y'], E1['vx'], E1['hy'], E1['vslope'],
                  E1['hslope'], E1['sw_SE_num'], E1['sw_SE_den'], E1['n_y'],
                  E1['e_x'], br, E1['nesw_sqr'], ilk = "inver")))

    insol <- ifelse(
      !is.na(insol), insol, ifelse(
        ezon, 0,
        p_deviOrg(x, y, E1['w_x'], E1['s_y'], E1['vx'], E1['hy'], E1['vslope'],
                  E1['hslope'], E1['sw_NW_num'], E1['sw_NW_den'], E1['n_y'],
                  E1['e_x'], br, E1['nesw_sqr'], ilk = "insol")))

    return(list(
      inver = inver, insol = insol

      # TODO Not needed, just for development
      # , idRectE = E1$mazeID, idRectW = E1$mazeID,
      # feas = feas, ezon = ezon
    ))
  }

p_deviOrg <-
  function (x, y, x0, y0, vx, hy, sv, sh, t0, s0, ny, ex, br, sqr, ilk = "insol") {
    dev <- 0
    dx <- x - x0; dy <- y - y0 # local coordinates
    if (ilk == "inver") { # ilk = inver
      xx <- br[2]; yy <- br[3] # reference values
      # east - in-veracity, in the in-feasible space
      # y != yy unless we deal with a trivial case
      if (x == xx) {
        dev <- (y - hy) / (y - yy)
      }
      # south - in-veracity, in the in-feasible space
      # x != xx unless we deal with a trivial case
      if (y == yy) {
        dev <- (vx - x) / (xx - x)
      }
      if ((x != xx) & (y != yy)) {
        nd <- p_numdenOrg(x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy)
        dev <- ifelse(
          (sqr == "ne_sqr") & (nd$den == 0),
          dy / (y - yy),
          ifelse((sqr == "sw_sqr") & (nd$den == 0),
                 (ex - x) / (br[2] - x),
                 nd$num / nd$den))
      } # ilk = inver
    } else { # ilk = "insol"
      xx <- br[1]; yy <- br[4] # reference values
      # west - in-solidity, in the feasible space
      # y != yy unless we deal with a trivial case
      if (x == xx) {
        dev <- (y - hy) / (y - yy)
      }
      # north - in-solidity, in the feasible space
      # x != xx unless we deal with a trivial case
      if (y == yy) {
        dev <- (x - vx) / (xx - x)
      }
      if ((x != xx) & (y != yy)) {
        nd <- p_numdenOrg(x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy)
        dev <- ifelse(
          (sqr == "ne_sqr") & (nd$den == 0),
          dx / (x - xx),
          ifelse(
            (sqr == "sw_sqr") & (nd$den == 0),
            (ny - y) / (br[4] - y),
            nd$num / nd$den))
      } # ilk = "insol"
    }
    return(unlist(dev))
  }

p_numdenOrg <-
  function (x, y, x0, y0, vx, hy, sv, sh, t0, s0, xx, yy) {
    # num and den: ratio = num/den
    # At mesh with sw-point = (x0,y0), in local coordinates (dx = x - x0, dy = y - y0),
    # numerator = dx^2 sh/2 - dx dy + dy^2/(2 sv) + dx (hy - y0) - dy (x0 - vx) + t0
    # denominator = - dx dy + dx (ym -y0)- dy (x0 - vx) + s0
    # For iso this leads to a conic section ax^2 + 2nxy + by^2 + 2hx + 2ky + c = 0
    # reading (for in-solidity)
    # dx^2 sh/2 - (1-JJ) dx dy + dy^2/(2 sv) + dx (hy - y0  - JJ (hy -yM)) -
    #     dy (x - vx - JJ (x - xm)) + t0 - JJ s0 = 0
    # where t0 and s0 are numerator and denominator of
    # the (in-solidity) inaccuracy measure at the sw-point (x0,y0).
    dx <- x - x0; dy <- y - y0 # local coordinates
    num <- dx^2 * sh / 2 - dx * dy +
      dy^2 / (2 * sv) +
      dx * (hy - y0) - dy * (x0 - vx) + t0
    den <- -dx * dy + dx * (yy - y0) - dy * (x0 - xx) + s0
    return(list(num = num, den = den))
  }

###############################################################################
# TODO Remove above
###############################################################################

p_nciRaw2Norm <-
  function (riacc, d, solidity = TRUE) {
    # riacc = raw in-accuracy (in-solidity or in-veracity) measure
    # d = effect size; solidity = TRUE: solidity transformation

    # racc is some (solidity/veracity) raw measure
    racc <- 1 - unlist(riacc)

    if (solidity) {
      nsol <- (racc - d) / (1 - d) # nrm solidity
      return(1 - nsol)             # nrm in-solidity = (1-racc)/(1-d)
    } else {
      nver <- (racc - (1 - d)) / d # nrm veracity
      return(1 - nver)             # nrm in-veracity)
    }
  }

p_svGauge <-
  function (svP, critLevs = c(cris = .2, criv = .1)) {
    stopifnot(all(c("insol_net", "inver_net") %in% names(svP)))

    # Flags
    satS <- svP$insol_net == 0
    medS <- 0 <= svP$insol_net &
      svP$insol_net <= critLevs[["cris"]] &
      !satS
    lowS <- !medS & !satS

    satP <- svP$inver_net == 0
    medP <- 0 <= svP$inver_net &
      svP$inver_net <= critLevs[["criv"]] &
      !satP
    lowP <- !medP & !satP

    # NA-safe combos
    satD <- ifelse(is.na(satS), FALSE, satS) & ifelse(is.na(satP), FALSE, satP)
    medSsatD <- ifelse(is.na(medS), FALSE, medS) | satD

    cnt <- function (x) sum(ifelse(is.na(x), FALSE, x))
    gauge <- list(
      satD = cnt(satD),
      medS = cnt(medS),
      medP = cnt(medP),
      lowS = cnt(lowS),
      lowP = cnt(lowP),
      satS = cnt(satS),
      satP = cnt(satP),
      medSsatD = cnt(medSsatD)
    )

    # Sharpness
    denom0 <- gauge$satD + gauge$medS + gauge$medP
    sharp0 <- if (denom0 != 0) (gauge$satD + gauge$medS - gauge$medP) / denom0 else NA_real_
    denom1 <- gauge$medS + gauge$medP
    sharp1 <- if (denom1 != 0) (gauge$medS) / denom1 else NA_real_

    # Percentages & totals
    numC <- nrow(svP)
    gauge$numC <- numC
    gauge$medSpct <- if (numC > 0) 100 * gauge$medS / numC else NA_real_
    gauge$medPpct <- if (numC > 0) 100 * gauge$medP / numC else NA_real_
    gauge$medSsatDpct <- if (numC > 0) 100 * gauge$medSsatD / numC else NA_real_
    gauge$nribbon <- gauge$medSsatD + gauge$medP # number of points in ribbon
    gauge$sharp0 <- sharp0
    gauge$sharp1 <- sharp1

    # Return flags attached to svP (useful downstream)
    svP$satS <- satS
    svP$medS <- medS
    svP$lowS <- lowS
    svP$satP <- satP
    svP$medP <- medP
    svP$lowP <- lowP
    svP$satD <- satD
    svP$medSsatD <- medSsatD

    list(svP = svP, gauge = gauge)
  }

p_ensure_pct_cols <-
  function (gauge, n_points = NULL) {
    if ((!("numC" %in% names(gauge)) || is.na(gauge$numC)) && !is.null(n_points)) {
      gauge$numC <- n_points
    }
    if (is.na(gauge$numC) || gauge$numC <= 0) {
      return(gauge)
    }
    if (!("medSpct" %in% names(gauge)) && "medS" %in% names(gauge)) {
      gauge$medSpct <- 100 * gauge$medS / gauge$numC
    }
    if (!("medSsatDpct" %in% names(gauge)) && "medSsatD" %in% names(gauge)) {
      gauge$medSsatDpct <- 100 * gauge$medSsatD / gauge$numC
    }
    if (!("medPpct" %in% names(gauge)) && "medP" %in% names(gauge)) {
      gauge$medPpct <- 100 * gauge$medP / gauge$numC
    }
    return(gauge)
  }

p_get_nca_metrics <-
  function (df, condition, outcome, ceiling, test_rep = 1000) {
    model <- nca_analysis(df, condition, outcome, ceilings = ceiling, test.rep = test_rep)
    part <- function (param) nca_extract(model, x = condition, ceiling = ceiling, param = param)
    list(
      effect_size = part(param = 'Effect size'),
      p_value = part(param = 'p-value'),
      fit = part(param = 'Fit'),
      c_accuracy = part(param = 'Ceiling accuracy'),
      npeers = nrow(model$peers[[ceiling]][[condition]])
    )
  }

p_compute_spread <-
  function (svP_flags, web) {
    # x-range from the original BR (outer bounding range)
    br <- web$BRs$obr
    xmin <- br[1]
    xmax <- br[2]

    # x for rows in the ribbon: medS U satD  (NA-safe)
    in_ribbon <- ifelse(is.na(svP_flags$medSsatD), FALSE, svP_flags$medSsatD)
    x_med <- svP_flags$x[in_ribbon]

    if (length(x_med) > 0L) {
      return(p_spread_index(x_med, xmin, xmax))
    } else {
      return(NA_real_)
    }
  }

p_spread_index <-
  function (x, xmin, xmax) {
    stopifnot(is.numeric(x))
    if (!is.finite(xmin) ||
      !is.finite(xmax) ||
      xmax <= xmin) {
      stop("xmax must be > xmin, both finite.")
    }

    x <- x[is.finite(x)]
    if (length(x) == 0L) {
      return(NA_real_)
    }
    z <- sort((x - xmin) / (xmax - xmin))
    z <- pmin(pmax(z, 0), 1)
    gaps <- diff(c(0, z, 1))
    mean_gap <- mean(gaps)
    if (mean_gap == 0) {
      return(0)
    }
    cv <- stats::sd(gaps) / mean_gap
    return(1 / (1 + cv))
  }

p_abCeil4Data <-
  function (xyP, ilk = "RG") {
    if (ilk == "RG") {
      ab <- p_performRG(xyP)
    } else {
      ab <- p_performLP(xyP)
    }

    if (is.na(ab["b"]) || ab["b"] <= 0) {
      # no horizontal or downward slope
      ab["b"] <- 1e-12
    }

    return(ab)
  }

p_performLP <-
  function (xyP, weights = NULL, hallObjective = FALSE) {
    x.low <- min(xyP[, 'X'])
    y.low <- min(xyP[, 'Y'])
    xyP[, 'X'] <- xyP[, 'X'] - x.low
    xyP[, 'Y'] <- xyP[, 'Y'] - y.low
    numRows <- nrow(xyP)
    sway <- rep(1, numRows) # leads to equal weights
    # TODO I don't think this works, ordering should be over rows not columns?
    if (!(is.null(weights))) { # so we are weighing the points
      sway <- xyP[, weights]   # the weights = data in column/mast
    }
    if (hallObjective) {
      obj <- c(1, 1 / 2) # the Hall objective
    } else {
      obj <- c(sum(sway), sum(sway * xyP[, 'X']))   # "ordinary" objective coefficients
    }

    # lhs of constraints (maybe weighted)
    leftside <- cbind(rep(1, numRows), xyP[, 'X'])
    # direction of constraint(s)
    inEq <- rep(">=", numRows)
    # rhs of constraints
    rightside <- xyP[, 'Y']

    # LP with shifted data inputs
    LP_ <- lp("min", obj, leftside, inEq, rightside)
    # coefficients wrt shifted data
    a_ <- LP_$solution[1]
    b_ <- LP_$solution[2]
    # nb: a gets measured at x = 0!
    a <- a_ - b_ * x.low + y.low

    return(c(a = a, b = b_)) # line: y = a + b x, so raw ab
  }

p_performRG <-
  function (xyP) {
    # do regression of y on x
    ab <- c(a = 0, b = 1)
    xyP <- as.data.frame(xyP)
    if (nrow(xyP) > 1) {
      LM <- lm(Y ~ X, data = xyP)
      ab <- c(a = unname(LM$coefficients[1]), b = unname(LM$coefficients[2]))
    }
    return(ab)
  }

p_intersectLineBox <-
  function (ab, scope) {
    # crossings of box with line and es

    dx <- scope[2] - scope[1]
    dy <- scope[4] - scope[3]
    a <- ab[1]
    b <- ab[2]

    if (dx * dy <= 0) {
      return(NULL)
    }
    if (b <= 0) {
      return(NULL)
    }

    miss <- FALSE # does the ab-line miss the box?
    # intersection y = a + b x with x = bb$x.low:
    wi <- c(scope[1], a + b * scope[1]) # west intersection
    # intersection y = a + b x with y = bb$y.low:
    si <- c((scope[3] - a) / b, scope[3]) # south intersection

    if (wi[1] < si[1]) {
      swi <- si
    } else {
      # pick most eastern point
      swi <- wi
    }
    if (swi[1] > scope[2]) {
      # box to north-west of line
      miss <- TRUE
      es <- 1
    }
    if (swi[2] > scope[4]) {
      # box to south-east of line
      miss <- TRUE
      es <- 0
    }

    # intersection y = a + b x with x = bb$x.high:
    ei <- c(scope[2], a + b * scope[2]) # east intersection
    # intersection y = a + b x with y = bb$y.high:
    ni <- c((scope[4] - a) / b, scope[4]) # north intersection
    # pick most western point
    if (scope[2] > (scope[4] - a) / b) {
      nei <- ni
    } else {
      nei <- ei
    }
    if (nei[1] < scope[1]) {
      # box to south-east of line
      miss <- TRUE
      es <- 0
    }
    if (nei[2] < scope[3]) {
      # box (unreasonably) to north-west of line
      miss <- TRUE
      es <- 1
    }

    if (!miss) {
      # regular case: no miss and two intersections swi and nei with box's boundary
      es <- (swi[1] - scope[1]) * dy
      es <- es + (1 / 2) *
        (nei[1] - swi[1]) *
        (2 * scope[4] - swi[2] - nei[2])
    }

    # "box-line" intersections
    blis <- list(wi = wi, si = si, ei = ei, ni = ni)

    return(list(es = es / (dx * dy), swi = swi, nei = nei, blis = blis,
                inline = c(a = a, b = b), inbox = scope))
  }

p_test <-
function (analyses, loop.data, test.params, effect_aggregation) {
  if (test.params$rep < 1) {
    return (NULL)
  }
  test.params$rep <- round(test.params$rep)
  test.params$p_confidence <- max(0, min(1, test.params$p_confidence))
  test.params$p_threshold <- max(0, min(1, test.params$p_threshold))

  # Add statistics tests
  test <- list()
  h <- length(loop.data$x)
  y_org <- loop.data$y
  effect.sims <- list()
  # But not for OLS
  ceilings <- names(analyses)[names(analyses) != "ols"]
  # Purity not needed
  loop.data[[p_skip_purity]] <- TRUE

  # Do you remember?
  start <- Sys.time()
  # Shorten the X name
  x.name <- paste0(strtrim(colnames(loop.data$x), 25),
    ifelse(nchar(colnames(loop.data$x)) > 25, "...", ""))

  # Create a unique list of samples
  offset <- sample.int(2^24, 1)
  i <- NULL
  samples <- foreach (i=1:test.params$rep) %dopar% {
    set.seed(offset + i)
    sample(1:h, h, replace = FALSE)
  }

  # For small sample sizes, make sure they are unique
  if (test.params$rep <= 720) {
    while (length(unique(samples)) < test.params$rep) {
      set.seed(offset + length(samples))
      samples[[length(samples) + 1]] <- sample(1:h, h, replace = FALSE)
    }
    samples <- unique(samples)
  }

  # Unset the seed, and we're done
  set.seed(NULL)

  for (ceiling in ceilings) {
    cat("Do test for  : ", ceiling, "-", x.name)

    effect.sims[[ceiling]] <- foreach (sample=iter(samples)) %dopar% {
      loop.data$y <- y_org[unlist(sample)]
      analysis <- p_nca_wrapper(ceiling, loop.data, NULL, effect_aggregation)
      return (analysis$effect)
    }

    # Convert to a vector, remove NA
    effect.sims[[ceiling]] <- unlist(effect.sims[[ceiling]])
    effect.sims[[ceiling]] <- effect.sims[[ceiling]][!is.na(effect.sims[[ceiling]])]

    cat("\rDone test for: ", ceiling, "-", x.name,  strrep(" ", 5), "\n")
  }

  for (ceiling in ceilings) {
    observed <- analyses[[ceiling]]$effect
    data <- effect.sims[[ceiling]]

    if (is.null(data)) {
      fmt <- "No permutation test for %s on %s\n"
      cat(sprintf(fmt, loop.data$names[1], ceiling))
    }

    # Add threshold- and P-value for displaying in summary
    threshold.value <- as.numeric(quantile(sort(data), 1 - test.params$p_threshold))
    # https://doi.org/10.1086/341527
    tmp <- data >= observed
    p_value <- (sum(tmp) + 1) / (length(tmp) + 1)
    p_value <- max(p_value, 1/test.params$rep)

    # Add MOE
    uns <- test.params$p_confidence + 0.5 * (1 - test.params$p_confidence)
    test.params$p_accuracy <- qnorm(uns) * sqrt(p_value * (1 - p_value) / test.params$rep)

    if (h <= 6 && test.params$rep == factorial(h)) {
      test.params$p_accuracy <- 0
    }

    names <- c(colnames(loop.data$x), colnames(y_org))
    test[[ceiling]] <- list(data = data, observed = observed, test.params=test.params,
                            p_value = p_value, threshold.value = threshold.value,
                            names = names)
  }

  return (list(test=test, test.time=difftime(Sys.time(), start, units="secs")))
}

p_test_time <-
function (test.time) {
  test.time <- as.numeric(test.time)
  if (test.time > 86400) {
    duration <- sprintf("%s days, %02d:%02d:%02d",
      test.time %/% 86400,
      test.time %% 86400 %/% 3600,
      test.time %% 3600 %/% 60,
      test.time %% 60 %/% 1)
  }
  else {
    duration <- sprintf("%02d:%02d:%02d",
      test.time %/% 3600,
      test.time %% 3600 %/% 60,
      test.time %% 60 %/% 1)
  }
  return (paste("Test done in", duration))
}

p_display_test <-
function (test, pdf=FALSE, path=NULL) {
  for (ceiling in rev(names(test))) {
    p_display_ceiling_test(ceiling, test[[ceiling]], pdf, path)
  }
}

p_display_ceiling_test <-
function (ceiling, ceiling_test, pdf=FALSE, path=NULL) {
  ceiling <- p_pretty_name(ceiling)
  data <- ceiling_test$data
  observed <- ceiling_test$observed
  p_value <- ceiling_test$p_value
  threshold.value <- ceiling_test$threshold.value
  p_threshold <- ceiling_test$test.params$p_threshold
  names <- ceiling_test$names

  # All Y-values are equal
  if (is.null(data)) {
    return(NULL)
  }

  bin.count <- 30
  x.low <- - 0.05
  x.high <- max(data, observed) + 0.01

  label.main <- sprintf("X=%s  Y=%s  %s\n", names[1], names[2], ceiling)
  label.threshold <- ifelse(p_threshold > 0, sprintf(
      "----  threshold (d = %.2f, p_threshold %.2f)", threshold.value, p_threshold), "")
  label.p <- ""
  if (!is.na(p_value)) {
    rep <- ceiling_test$test.params$rep
    p_accuracy <- ceiling_test$test.params$p_accuracy
    if (p_accuracy > 1e-6) {
      p.min <- max(0, p_value - p_accuracy)
      p.max <- min(1, p_value + p_accuracy)
      label.p <- sprintf(", p = %.3f [%.3f, %.3f], rep = %d",
                       p_value, p.min, p.max, rep)
    } else {
      label.p <- sprintf(", p = %.3f, rep = %d", p_value, rep)
    }
  }
  label.observed <- sprintf("observed (d = %.2f%s)", observed, label.p)

  # Add data plot and title
  df <- data.frame(x=data)
  obj <- ggplot(df, aes(x=.data$x)) + xlim(x.low, x.high)
  obj <- obj + geom_histogram(bins = bin.count, fill = I("white"), colour = I("black"), na.rm = TRUE)
  obj <- obj + labs(x = "Permutated effect sizes", y = "Frequency",
                    title = label.main, subtitle = "")
  obj <- obj + theme(plot.title=element_text(hjust=0.5))

  # Add lines
  if (p_threshold > 0) {
    obj <- obj + geom_vline(xintercept = threshold.value, col = "darkgreen", lwd = .5, lty = 5)
  }
  obj <- obj + geom_vline(xintercept = observed, col = "red", lwd = 0.5, lty = 1)

  # Open new window or PDF
  if (pdf) {
    title <- paste(paste(ceiling_test$names, collapse="-"), "random test")
    p_new_pdf("plot", title, path)
  } else {
    p_new_window(title = paste("Random test ", ceiling))
    par(family = "")
    par(mfrow = c(1, 1))
  }

  # Plot all
  plot(obj)

  # Add legend for lines
  p_add_legend(label.threshold, label.observed, p_threshold)

  if (pdf) {
    dev.off()
    cat("")
  }
}

p_add_legend <-
function (label.threshold, label.observed, p_threshold) {
  gp.r <- gpar(col = "darkgreen", fontsize = 11)
  gp.o <- gpar(col = "red", fontsize = 11)
  gp.o1 <- gpar(col = "red", fontsize = 10, fontface="bold")
  if (p_threshold > 0) {
    pos.x  <- unit(0.5, "npc") - unit(5, "cm")
    pos.x1 <- pos.x + unit(.75, "cm")
    pos.y  <- unit(1, "npc") - unit(2, "line")
    pos.y1 <- unit(1, "npc") - unit(3.0, "line")
    pos.y2 <- unit(1, "npc") - unit(3.0, "line")

    grid.text(label.threshold,   pos.x,  pos.y,  just="left", gp = gp.r)
    grid.text("___",          pos.x,  pos.y1, just="left", gp = gp.o1)
    grid.text(label.observed, pos.x1, pos.y2, just="left", gp = gp.o)
  }
  else {
    pos.x <- unit(0.5, "npc")
    pos.y <- unit(1, "npc") - unit(2, "line")
    grid.text(label.observed, pos.x, pos.y, just="centre", gp = gp.o)
  }
}

nca_output <-
  function (model, plots = TRUE, plotly = FALSE, bottlenecks = FALSE, summaries = TRUE,
            test = FALSE, pdf = FALSE, path = NULL, selection = NULL, plot_bottlenecks = 0) {
    # model needs to be the output of the NCA command

    result <- NULL

    # We have a numeric vector
    if (is.numeric(selection)) {
      selection <- intersect(selection, seq_along(model$summaries))
    }
      # We have a string (names) vector
    else {
      selection <- match(selection, names(model$summaries))
      selection <- selection[!is.na(selection)]
    }
    # Just take all
    if (is.null(selection) || length(selection) == 0) {
      selection <- seq_along(model$summaries)
    }

    if (bottlenecks) {
      # Insert the Y-column as first
      selection.bottlenecks <- c(1, selection + 1)

      bn <- list()
      for (method in names(model$bottlenecks)) {
        bn.method <- model$bottlenecks[[method]]
        bn[[method]] <- bn.method[, selection.bottlenecks]

        atts <- c("bn.x", "bn.y", "bn.y.id", "size", "cutoff")
        for (att in atts) {
          attr(bn[[method]], att) <- attr(model$bottlenecks[[method]], att)
        }
      }
      p_display_bottleneck(bn, pdf = pdf, path = path)
    }

    if (summaries) {
      for (i in selection) {
        summary <- model$summaries[[i]]
        p_display_summary(summary, pdf = pdf, path = path)
      }
    }

    if (plots) {
      for (i in selection) {
        plot <- model$plots[[i]]
        bottleneck.list <- p_get_bottleneck_list(model, i, plot_bottlenecks)
        p <- p_display_plot(plot, bottleneck.list = bottleneck.list, pdf = pdf, path = path)
        name <- names(model$summaries)[i]
        result$plots[[name]] <- p
      }
    }

    if (test) {
      if (length(model$tests) == 0) {
        warning(paste("Tests are selected in the output,",
                      "but non were supplied by the analysis!"), call. = F)
      }
      else {
        for (i in selection) {
          test <- model$tests[[i]]
          p_display_test(test, pdf = pdf, path = path)
        }
      }
    }

    if (!isFALSE(plotly)) {
      for (i in selection) {
        labels <- NULL
        if (!isTRUE(plotly)) {
          labels <- plotly
        }
        peers <- p_aggregate_peers(model$peers, i)
        bottleneck.list <- p_get_bottleneck_list(model, i, plot_bottlenecks)
        p_display_plotly(model$plots[[i]], peers, labels, bottleneck.list = bottleneck.list)
      }
    }

    cat("")
    invisible(result)
  }

p_get_bottleneck_list <-
  function (model, i, plot_bottlenecks) {
    bottleneck.list <- list(ver = list(), hor = list())
    if (plot_bottlenecks %in% c(1, 2)) {
      mpy <- attr(model$bottlenecks, "mpy")
      mpy_labels <- model$bottlenecks[[1]][, 1]
      precision <- median(sapply(mpy_labels, p_get_precision), na.rm = TRUE)
      precision <- ifelse(is.na(precision), 0, precision)
      x0 <- model$plots[[i]]$scope.theo[1]
      x1 <- model$plots[[i]]$scope.theo[2]
      for (j in seq_along(mpy)) {
        label <- format(round(mpy_labels[j], precision), nsmall = precision)
        coord <- list(x0, x1, mpy[j], mpy[j])
        bottleneck.list[['hor']][[label]] <- coord
      }

      # Aggregate over all methods, make sure the lines are unique
      if (plot_bottlenecks == 2) {
        y0 <- model$plots[[i]]$scope.theo[3]
        y1 <- model$plots[[i]]$scope.theo[4]

        for (method in names(model$bottlenecks)) {
          tmp <- model$bottlenecks[[method]][, i + 1]
          mpx.actual <- attr(tmp, "mpx.actual")
          tmp <- suppressWarnings(as.numeric(tmp))
          precision <- median(sapply(tmp, p_get_precision), na.rm = TRUE)
          precision <- ifelse(is.na(precision), 0, precision)
          for (j in seq_len(nrow(mpx.actual))) {
            if (is.finite(mpx.actual[j])) {
              value <- ifelse(is.na(tmp[j]), 0, tmp[j])
              label <- format(round(value, precision), nsmall = precision)
              coord <- list(mpx.actual[j], mpx.actual[j], y0, y1)
              bottleneck.list[['ver']][[label]] <- coord
            }
          }
        }
      }
    }
    return(bottleneck.list)
  }

print.nca_result <-
  function (x, ...) {
    p_display_summary_simple(x$summaries)
    if (attr(x, "show.plots")) {
      nca_output(x, plots = TRUE, summaries = FALSE, bottlenecks = FALSE)
    }
  }

summary.nca_result <-
  function (object, columns = NULL, ...) {
    if (!is.null(columns)) {
      # Columns can be indexes or names
      if (!is.numeric(columns)) {
        columns <- match(c(columns), names(object$summaries))
        columns <- columns[!is.na(columns)]
      }
      else {
        columns <- columns[columns > 0]
        columns <- columns[columns < length(object$summaries)]
      }

      # Make sure user actually selected columns
      tmp <- object$summaries[columns]
      if (length(tmp) > 0) {
        object$summaries <- tmp
      }
    }

    nca_output(object)
  }

plot.nca_result <-
  function (x, ...) {
    nca_output(x, plots = TRUE, summaries = FALSE, bottlenecks = FALSE)
  }

p_create_loop_data <-
function (x, y, scope, flip.x, flip.y, id.x, qr.tau, custom) {
  names <- c(colnames(x), colnames(y))
  x <- x[id.x]
  scope <- scope[[id.x]]
  flip.x <- flip.x[id.x]

  # Remove not complete cases
  tmp <- cbind(x, y)
  tmp <- tmp[complete.cases(tmp),]
  x <- tmp[-length(tmp)]
  y <- tmp[length(tmp)]

  # Define the scope params
  scope.emp <- c(min(x), max(x), min(y), max(y))
  if (is.null(scope)) {
    scope.theo <-  scope.emp
  } else {
    scope.theo <- c(min(scope.emp[1], scope[c(1, 2)], na.rm=TRUE),
                    max(scope.emp[2], scope[c(1, 2)], na.rm=TRUE),
                    min(scope.emp[3], scope[c(3, 4)], na.rm=TRUE),
                    max(scope.emp[4], scope[c(3, 4)], na.rm=TRUE))
    if (any(sort(scope[1:2]) != sort(scope.theo[1:2])) ||
        any(sort(scope[3:4]) != sort(scope.theo[3:4]))) {
      # Needed to collapse multiple scope warnings
      if (is.null(.GlobalEnv$scope_warnings_nca)) {
        .GlobalEnv$scope_warnings_nca <- TRUE
        warning(
          "Theorectical scope has been adjusted to include all observations", call.=FALSE)
      }
    }
  }
  scope.area  <- (scope.theo[2] - scope.theo[1]) * (scope.theo[4] - scope.theo[3])

  return (list(x=as.matrix(x), y=as.matrix(y), idx=id.x,
               scope.emp=scope.emp, scope.theo=scope.theo,
               scope.area=scope.area, names=names,
               flip.x=flip.x, flip.y=flip.y,
               qr.tau=qr.tau, custom=custom))
}

nca_random <- function (n, intercepts, slopes, corner = 1,
                        distribution.x = "uniform", distribution.y = "uniform",
                        mean.x = 0.5, mean.y = 0.5, sd.x = 0.2, sd.y = 0.2) {
  # Validate the inputs
  tmp <- p_validate_inputs(n, intercepts, slopes, corner, distribution.x, distribution.y)
  error <- tmp$error
  intercepts <- tmp$intercepts
  slopes <- tmp$slopes

  if (!is.null(error)) {
    stop(p_errors[error])
  }

  data <- NULL
  if (length(slopes) == 1) {
    data[['X']] <- NA
  }
  else {
    for (i in seq_along(slopes)) {
      data[[paste0('X', i)]] <- NA
    }
  }
  data[['Y']] <- NA
  data <- data.frame(data)
  for (idx in 1:n) {
    while (TRUE) {
      # Generate random point within [0, 1], check if they conform to the line
      y.value <- p_value(distribution.y, mean.y, sd.y)
      data[idx, length(slopes) + 1] <- y.value

      all.good <- TRUE
      for (x.idx in seq_along(slopes)) {
        x.value <- p_value(distribution.x, mean.x, sd.x)
        data[idx, x.idx] <- x.value

        slope <- slopes[x.idx]
        intercept <- intercepts[x.idx]
        if (corner %in% c(1, 2) && y.value >= min(1, (intercept + slope * x.value))) {
          all.good <- FALSE
        }
        if (corner %in% c(3, 4) && y.value <= max(0, (intercept + slope * x.value))) {
          all.good <- FALSE
        }
      }
      if (all.good) {
        break
      }
    }
  }
  data[order(data$Y),]
  rownames(data) <- 1:n

  attr(data, "distribution.x") <- distribution.x
  attr(data, "distribution.y") <- distribution.y
  if (distribution.x == "normal") {
    attr(data, "mean.x") <- mean.x
    attr(data, "sd.x") <- sd.x
  }
  if (distribution.y == "normal") {
    attr(data, "mean.y") <- mean.y
    attr(data, "sd.y") <- sd.y
  }

  return(data)
}

p_value <- function (distribution, mean, sd)
{
  if (distribution == "uniform") {
    return(runif(1, 0, 1))
  }
  else {
    return(rtruncnorm(1, a = 0, b = 1, mean = mean, sd = sd))
  }
}

p_errors <- list(
  n = "n should be an integer > 1!",
  combination = "The combination of slope and intercept does not provide points in the [(0, 0), (1, 1)] area!",
  length = "The length of the slopes and intercepts should be equal!",
  corner_23 = "Upward slope can not provide empty corners 2 (upper right) or 3 (lower left)!",
  corner_14 = "Downward slope can not provide empty corners 1 (upper left) or 4 (lower right)!",
  distribution = "The distribution types need to be 'uniform' or 'normal'!"
)

p_validate_inputs <- function (n, intercepts, slopes, corner, distribution.x, distribution.y)
{
  if (n < 1 || n != round(n)) {
    return(list(error = 'n', intercepts = intercepts, slopes = slopes))
  }

  if (length(intercepts) == 1 && length(slopes) > 1)
  {
    intercepts <- rep(intercepts, length(slopes))
  }
  else if (length(intercepts) > 1 && length(slopes) == 1) {
    slopes <- rep(slopes, length(intercepts))
  }

  if (length(intercepts) != length(slopes)) {
    return(list(error = 'length', intercepts = intercepts, slopes = slopes))
  }

  for (idx in seq_along(slopes)) {
    slope <- slopes[idx]
    intercept <- intercepts[idx]
    cond.1 <- slope > 0 && (intercept >= 1 || (intercept + slope) <= 0)
    cond.2 <- slope < 0 && (intercept <= 0 || (intercept + slope) >= 1)
    cond.3 <- slope == 0 && (intercept <= 0 || intercept >= 1)
    if (cond.1 || cond.2 || cond.3) {
      return(list(error = 'combination', intercepts = intercepts, slopes = slopes))
    }

    if (slope > 0 && (corner %in% c(2, 3))) {
      return(list(error = 'corner_23', intercepts = intercepts, slopes = slopes))
    }
    if (slope < 0 && (corner %in% c(1, 4))) {
      return(list(error = 'corner_14', intercepts = intercepts, slopes = slopes))
    }
  }

  if (length(setdiff(c(distribution.x, distribution.y), c("uniform", "normal"))) != 0) {
    return(list(error = 'distribution', intercepts = intercepts, slopes = slopes))
  }

  return(list(error = NULL, intercepts = intercepts, slopes = slopes))
}

# 0 : NN / NA
# 1 : NN / highest observed values
# 2 : calculated values

p_bottleneck <-
  function (loop.data, bn.data, slope, intercept) {
    if (is.null(bn.data)) {
      return(NULL)
    }

    theo <- loop.data$scope.theo
    flip.x <- loop.data$flip.x
    precision.x <- ifelse(bn.data$bn.x.id %in% c(1, 2, 4), 1, 3)

    if (is.na(intercept) || is.na(slope)) {
      mpx <- p_mpx_single_peer(bn.data, theo, flip.x)
    } else {
      mpx <- (bn.data$mpy - intercept) / slope
      mpx <- p_edge_cases(mpx, bn.data, theo, flip.x)
    }
    cases <- p_cases(loop.data, mpx)

    nn.value <- p_nn_value(mpx, loop.data, bn.data)
    na.value <- p_na_value(mpx, loop.data, bn.data)

    # Display Xs as percentage (either cutoff or 0-high) or percentile
    mpx.actual <- mpx
    mpx <- p_transform_mpx(loop.data, mpx, bn.data$bn.x.id)
    pretty_mpx <- p_pretty_mpx(loop.data, mpx, nn.value, na.value, precision.x)
    attr(pretty_mpx, "mpx.actual") <- mpx.actual
    if (bn.data$bn.x.id == 4) {
      attr(pretty_mpx, "cases") <- cases
    }

    return(pretty_mpx)
  }

p_bottleneck_ce <-
  function (loop.data, bn.data, peers, type) {
    if (is.null(bn.data)) {
      return(NULL)
    }

    theo <- loop.data$scope.theo
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y
    precision.x <- ifelse(bn.data$bn.x.id %in% c(1, 2, 4), 1, 3)

    if (type == "fdh") {
      mpx <- p_bottleneck_fdh(bn.data, peers, flip.y)
    } else if (type == "vrs") {
      mpx <- p_bottleneck_vrs(bn.data, peers, flip.y)
    }
    cases <- p_cases(loop.data, mpx)

    mpx <- p_edge_cases(mpx, bn.data, theo, flip.x, FALSE)
    nn.value <- p_nn_value(mpx, loop.data, bn.data)
    na.value <- p_na_value(mpx, loop.data, bn.data)

    # Display Xs as percentage (either cutoff or 0-high) or percentile
    mpx.actual <- mpx
    mpx <- p_transform_mpx(loop.data, mpx, bn.data$bn.x.id)
    pretty_mpx <- p_pretty_mpx(loop.data, mpx, nn.value, na.value, precision.x)
    attr(pretty_mpx, "mpx.actual") <- mpx.actual
    if (bn.data$bn.x.id == 4) {
      attr(pretty_mpx, "cases") <- cases
    }

    return(pretty_mpx)
  }

p_bottleneck_fdh <-
  function (bn.data, peers, flip.y) {
    mpy <- bn.data$mpy
    mpx <- matrix(nrow = length(mpy), ncol = 1)
    x.peers <- peers[, 1]
    y.peers <- peers[, 2]

    for (j in seq_along(mpy)) {
      # search the peer that is closest above to the desired outcome,
      # and select it corresponding x value of that peer
      if (flip.y) {
        index <- which(y.peers < (mpy[j, 1] + epsilon))[1]
      } else {
        index <- which(y.peers > (mpy[j, 1] - epsilon))[1]
      }

      if (is.na(index)) {
        mpx[j, 1] <- NA
      } else {
        mpx[j, 1] <- x.peers[index]
      }
    }

    return(mpx)
  }

p_bottleneck_vrs <-
  function (bn.data, peers, flip.y) {
    mpy <- bn.data$mpy
    mpx <- matrix(nrow = length(mpy), ncol = 1)
    x.peers <- peers[, 1]
    y.peers <- peers[, 2]

    calculate_x <- function (y, peers, index) {
      p1 <- peers[index - 1,]
      p2 <- peers[index,]
      return(p1[1] + (y - p1[2]) * (p1[1] - p2[1]) / (p1[2] - p2[2]))
    }

    for (j in seq_along(mpy)) {
      if (flip.y) {
        index <- which(y.peers < (mpy[j, 1] + epsilon))[1]
      } else {
        index <- which(y.peers > (mpy[j, 1] - epsilon))[1]
      }

      if (is.na(index)) {
        mpx[j, 1] <- NA
      } else if (index == 1) {
        mpx[j, 1] <- x.peers[index]
      } else {
        mpx[j, 1] <- calculate_x(mpy[j, 1], peers, index)
      }
    }

    return(mpx)
  }

p_mpx_single_peer <-
  function (bn.data, theo, flip.x) {
    mpy <- bn.data$mpy
    cutoff <- bn.data$cutoff

    # if peers is a vector there is only one peer
    if (cutoff == 0) {
      mpx <- matrix(Inf, nrow = length(mpy), ncol = 1)
    } else {
      mpx <- matrix(theo[1 + flip.x], nrow = length(mpy), ncol = 1)
    }
    return(mpx)
  }

p_transform_mpx <-
  function (loop.data, mpx, bn.x.id) {
    flip.x <- loop.data$flip.x
    theo <- loop.data$scope.theo

    # Display Xs as percentage (either cutoff or 0-high) or percentile
    if (bn.x.id == 1) {
      mpx <- 100 * (mpx - theo[1]) / (theo[2] - theo[1])
    } else if (bn.x.id == 2) {
      mpx <- 100 * mpx / theo[2]
    } else if (bn.x.id == 4) {
      if (flip.x) {
        percentile <- ecdf(sort(-loop.data$x))
        tmp <- -mpx - epsilon
      }
      else {
        percentile <- ecdf(sort(loop.data$x))
        tmp <- mpx - epsilon
      }
      mpx <- matrix(100 * percentile(tmp), ncol = 1)
      mpx[mpx == 0] <- Inf
    }

    return(mpx)
  }

p_nn_value <-
  function (mpx, loop.data, bn.data) {
    if (bn.data$cutoff %in% c(0, 1)) {
      return("NN")
    }

    flip.x <- loop.data$flip.x
    theo <- loop.data$scope.emp
    nn.value <- p_if_min_else_max(!flip.x, mpx, na.rm = TRUE)

    return(p_transform_value(loop.data, nn.value, theo, bn.data$bn.x.id))
  }

p_na_value <-
  function (mpx, loop.data, bn.data) {
    if (bn.data$cutoff == 0) {
      return("NA")
    }

    flip.x <- loop.data$flip.x
    theo <- loop.data$scope.theo
    na.value <- p_if_min_else_max(flip.x, mpx, na.rm = TRUE)

    return(p_transform_value(loop.data, na.value, theo, bn.data$bn.x.id))
  }

p_transform_value <-
  function (loop.data, value, theo, bn.x.id) {
    if (bn.x.id == 1) {
      value <- 100 * (value - theo[1]) / (theo[2] - theo[1])
    } else if (bn.x.id == 2) {
      value <- 100 * value / theo[2]
    } else if (bn.x.id == 4) {
      percentile <- ecdf(sort(loop.data$x))
      value <- 100 * percentile(value)
    }

    return(value)
  }

p_pretty_mpx <-
  function (loop.data, mpx, nn.value, na.value, precision.x) {

    subst <- function (x) {
      if (is.na(x)) {
        return(p_pretty_number(na.value, na.value, prec = precision.x))
      } else if (is.infinite(x)) {
        if (is.infinite(nn.value)) { nn.value <- "NN" }
        return(p_pretty_number(nn.value, nn.value, prec = precision.x))
      } else {
        return(p_pretty_number(x, prec = precision.x))
      }
    }

    mpx <- matrix(mapply(subst, mpx), ncol = 1)
    colnames(mpx) <- c(loop.data$names[loop.data$idx])

    return(mpx)
  }

p_edge_cases <-
  function (mpx, bn.data, theo, flip.x, use.epsilon = FALSE) {
    tmp <- ifelse(use.epsilon, epsilon, 0)
    mpx[mpx < (theo[1] + epsilon)] <- ifelse(flip.x, NA, -Inf)
    if (bn.data$cutoff == 0) {
      mpx[mpx > (theo[2] - tmp)] <- ifelse(flip.x, Inf, NA)
    } else if (bn.data$cutoff == 1) {
      mpx[mpx > (theo[2] - tmp)] <- theo[2]
    }

    return(mpx)
  }

p_cases <-
  function (loop.data, mpx) {
    x <- loop.data$x
    flip.x <- loop.data$flip.x
    mpx.cases <- mpx + ifelse(flip.x, epsilon, -epsilon)
    factor <- ifelse(flip.x, -1, 1)
    percentile <- ecdf(sort(factor * x))
    cases <- percentile(factor * mpx.cases) * nrow(x)
    return(cases)
  }

p_plot <-
  function (analyses, loop.data, corner) {
    plot <- list()

    plot$x <- loop.data$x
    plot$y <- loop.data$y
    plot$scope.theo <- loop.data$scope.theo
    plot$names <- loop.data$names
    plot$corner <- corner
    plot$flip.x <- ifelse(is.null(corner), loop.data$flip.x, FALSE)
    plot$flip.y <- ifelse(is.null(corner), loop.data$flip.y, FALSE)
    plot$conf <- loop.data$conf
    plot$title <- p_generate_title(colnames(plot$x), colnames(plot$y))
    plot$methods <- names(analyses)
    plot$lines <- list()
    plot$line_matrices <- list()

    for (method in plot$methods) {
      analysis <- analyses[[method]]
      plot$lines[[method]] <- analysis$line
      plot$line_matrices[[method]] <- analysis$line_matrix
    }

    return(plot)
  }

p_display_plot <-
  function (plot, bottleneck.list, pdf = FALSE, path = NULL) {
    # Get the params for plotting
    params <- get_plot_params()
    line_colors <- params[[1]]
    line_types <- params[[2]]
    line_width <- params[[3]]
    point_type <- params[[4]]
    point_color <- params[[5]]
    point_size <- params[[6]]

    corner <- plot$corner
    flip.x <- plot$flip.x
    flip.y <- plot$flip.y

    # Open new window or PDF
    if (pdf) {
      p_new_pdf("plot", plot$title, path)
    } else {
      p_new_window(title = plot$title)
      par(family = "")
      par(mfrow = c(1, 1))
    }

    # Add extra width / height for bottleneck data
    xlim <- c(plot$scope.theo[1 + flip.x], plot$scope.theo[2 - flip.x])
    ylim <- c(plot$scope.theo[3 + flip.y], plot$scope.theo[4 - flip.y])
    if (!is.null(names(bottleneck.list$hor))) {
      xlim <- c(xlim[1] - diff(xlim) / 10, xlim[2])
    }
    if (!is.null(names(bottleneck.list$ver))) {
      ylim <- c(ylim[1] - diff(ylim) / 25, ylim[2])
    }

    # Create main data frame
    df_points <- data.frame(x = plot$x, y = plot$y)
    names(df_points) <- c("x", "y")

    # Plot the data points
    p <- ggplot() +
      geom_point(data = df_points, aes(x = .data[['x']], y = .data[['y']]),
                 shape = point_type, color = point_color, size = point_size, fill = NA) +
      labs(x = colnames(plot$x), y = tail(plot$names, n = 1),
           title = paste0("NCA Plot : ", plot$title)) +
      coord_cartesian(xlim = xlim, ylim = ylim) +
      theme(
        plot.title = element_text(hjust = 0.5, vjust = 2, size = 14),
        text = element_text(size = 12),
        axis.text = element_text(size = 12),

        panel.border = element_rect(colour = "black", fill = NA),
        panel.grid.major = element_blank(),
        panel.grid.minor = element_blank(),
        panel.background = element_blank(),

        legend.position = c(0.00125, 0.9989),
        legend.justification = c(0, 1),
        legend.background = element_rect(fill = "white"),
        legend.box.background = element_rect(colour = "black", linewidth=.5),
        legend.title = element_blank(),
        legend.key.height = unit(0.45, "cm"),
        legend.text = element_text(size = 8),
      )

    # Plot the scope outline
    p <- p_ggplot_outline(p, plot)

    # Plot the bottlenecks
    p <- p_ggplot_bottlenecks(p, bottleneck.list, flip.x, flip.y)

    # Process Lines
    lines_data <- data.frame()
    for (method in plot$methods) {
      line_matrix <- plot$line_matrices[[method]]
      if (is_infinite_null(line_matrix)) {
        next
      }

      # Sort matrix depending on corner / flipping
      on.x <- line_matrix[, 1]
      on.y <- line_matrix[, 2]
      # Corner 2 or 3 || flip.x XOR flip.y
      if ((!is.null(corner) && corner %in% c(2, 3)) || xor(flip.x, flip.y)) {
        on.y <- -1 * on.y
      }
      line_matrix <- line_matrix[order(on.x, on.y),]

      tmp_df <- as.data.frame(line_matrix)
      names(tmp_df) <- c("x", "y")
      tmp_df$method <- p_pretty_name(method)
      tmp_df$color <- line_colors[[method]]
      tmp_df$line <- line_types[[method]]
      lines_data <- rbind(lines_data, tmp_df)
    }

    if (nrow(lines_data) > 0) {
      aesthetics <- unique(lines_data[c("method", "color", "line")])
      method_colors <- setNames(aesthetics$color, aesthetics$method)
      method_lines <- setNames(aesthetics$line, aesthetics$method)

      p <- p +
        geom_line(data = lines_data, linewidth = line_width,
                  aes(x = .data[['x']], y = .data[['y']],
                      color = .data[['method']], linetype = .data[['method']]))
      p <- p + scale_color_manual(values = method_colors)
      p <- p + scale_linetype_manual(values = method_lines)
    }

    print(p)
    if (pdf) {
      dev.off()
      cat("")
    }

    return(p)
  }

p_ggplot_bottlenecks <-
  function (p, bottleneck.list, flip.x, flip.y) {
    segments_hor <- NULL
    for (name in names(bottleneck.list[['hor']])) {
      coord <- unlist(bottleneck.list[['hor']][[name]])
      segments_hor <- rbind(segments_hor, coord)

      adjust <- abs(coord[2] - coord[1]) / 30
      coord.x <- ifelse(flip.x, coord[2] + adjust, coord[1] - adjust)
      p <- p + annotate(geom = "text", x = coord.x, y = coord[3], label = name, color = "grey")
    }
    if (!is.null(segments_hor)) {
      colnames(segments_hor) <- c("x", "xend", "y", "yend")
      p <- p + geom_segment(data = segments_hor, linetype = "dashed", color = "grey",
                            aes(x = .data[['x']], y = .data[['y']],
                                xend = .data[['xend']], yend = .data[['yend']]))
    }

    segments_ver <- NULL
    for (name in names(bottleneck.list[['ver']])) {
      coord <- unlist(bottleneck.list[['ver']][[name]])
      segments_ver <- rbind(segments_ver, coord)

      adjust <- abs(coord[4] - coord[3]) / 50
      coord.y <- ifelse(flip.y, coord[4] + adjust, coord[3] - adjust)
      p <- p + annotate(geom = "text", x = coord[1], y = coord.y, label = name, color = "grey")
    }
    if (!is.null(segments_ver)) {
      colnames(segments_ver) <- c("x", "xend", "y", "yend")
      p <- p + geom_segment(data = segments_ver, linetype = "dashed", color = "grey",
                            aes(x = .data[['x']], y = .data[['y']],
                                xend = .data[['xend']], yend = .data[['yend']]))
    }

    return(p)
  }

p_ggplot_outline <-
  function (p, plot) {
    p <- p + geom_vline(xintercept = plot$scope.theo[1], linetype = "dashed", color = "grey")
    p <- p + geom_vline(xintercept = plot$scope.theo[2], linetype = "dashed", color = "grey")
    p <- p + geom_hline(yintercept = plot$scope.theo[3], linetype = "dashed", color = "grey")
    p <- p + geom_hline(yintercept = plot$scope.theo[4], linetype = "dashed", color = "grey")
    return(p)
  }

get_plot_params <-
  function () {
    # Get the line colors and types from the global env, allowing user changes
    line_colors <- mget("line.colors", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    line_types <- mget("line.types", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    # Append default for missing values
    line_colors <- append(line_colors, line.colors)
    line_types <- append(line_types, line.types)
    # The same for line width and point-type/-color
    line_width <- mget("line.width", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(line_width)) {
      line_width <- line.width
    }
    point_type <- mget("point.type", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(point_type)) {
      point_type <- point.type
    }
    point_color <- mget("point.color", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    point_color <- tryCatch({
      col2rgb(point_color); point_color
    }, error = function (e) return(point.color))
    point_size <- mget("point.SIZE", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(point_size)) {
      point_size <- point.size
    }

    return(list(line_colors, line_types, line_width, point_type, point_color, point_size))
  }

is_infinite_null <-
  function (line) {
    if (is.null(line)) {
      return(TRUE)
    }

    # LH line
    if (typeof(line) == "double" && !all(is.finite(line))) {
      return(TRUE)
    }
    # 'LM' style line
    if (typeof(line) == "list" && !all(is.finite(line$coefficients))) {
      return(TRUE)
    }
    return(FALSE)
  }

p_plot <-
  function (analyses, loop.data, corner) {
    plot <- list()

    plot$x <- loop.data$x
    plot$y <- loop.data$y
    plot$scope.theo <- loop.data$scope.theo
    plot$names <- loop.data$names
    plot$corner <- corner[loop.data$idx]
    plot$flip.x <- ifelse(is.null(corner), loop.data$flip.x, FALSE)
    plot$flip.y <- ifelse(is.null(corner), loop.data$flip.y, FALSE)
    plot$conf <- loop.data$conf
    plot$title <- p_generate_title(colnames(plot$x), colnames(plot$y))
    plot$methods <- names(analyses)
    plot$lines <- list()
    plot$line_matrices <- list()

    for (method in plot$methods) {
      analysis <- analyses[[method]]
      plot$lines[[method]] <- analysis$line
      plot$line_matrices[[method]] <- analysis$line_matrix
    }

    return(plot)
  }

p_display_plot <-
  function (plot, bottleneck.list, pdf = FALSE, path = NULL) {
    # Get the params for plotting
    params <- get_plot_params()
    line_colors <- params[[1]]
    line_types <- params[[2]]
    line_width <- params[[3]]
    point_type <- params[[4]]
    point_color <- params[[5]]
    point_size <- params[[6]]

    corner <- plot$corner
    flip.x <- plot$flip.x
    flip.y <- plot$flip.y

    # Open new window or PDF
    if (pdf) {
      p_new_pdf("plot", plot$title, path)
    } else {
      p_new_window(title = plot$title)
      par(family = "")
      par(mfrow = c(1, 1))
    }

    # Add extra width / height for bottleneck data
    xlim <- c(plot$scope.theo[1 + flip.x], plot$scope.theo[2 - flip.x])
    ylim <- c(plot$scope.theo[3 + flip.y], plot$scope.theo[4 - flip.y])
    if (!is.null(names(bottleneck.list$hor))) {
      xlim <- c(xlim[1] - diff(xlim) / 10, xlim[2])
    }
    if (!is.null(names(bottleneck.list$ver))) {
      ylim <- c(ylim[1] - diff(ylim) / 25, ylim[2])
    }

    # Create main data frame
    df_points <- data.frame(x = plot$x, y = plot$y)
    names(df_points) <- c("x", "y")

    # Plot the data points
    p <- ggplot() +
      geom_point(data = df_points, aes(x = .data[['x']], y = .data[['y']]),
                 shape = point_type, color = point_color, size = point_size, fill = NA) +
      labs(x = colnames(plot$x), y = tail(plot$names, n = 1),
           title = paste0("NCA Plot : ", plot$title)) +
      coord_cartesian(xlim = xlim, ylim = ylim) +
      theme(
        plot.title = element_text(hjust = 0.5, vjust = 2, size = 14),
        text = element_text(size = 12),
        axis.text = element_text(size = 12),

        panel.border = element_rect(colour = "black", fill = NA),
        panel.grid.major = element_blank(),
        panel.grid.minor = element_blank(),
        panel.background = element_blank(),

        legend.position = c(0.00125, 0.9989),
        legend.justification = c(0, 1),
        legend.background = element_rect(fill = "white"),
        legend.box.background = element_rect(colour = "black", linewidth=.5),
        legend.title = element_blank(),
        legend.key.height = unit(0.45, "cm"),
        legend.text = element_text(size = 8),
      )

    # Plot the scope outline
    p <- p_ggplot_outline(p, plot)

    # Plot the bottlenecks
    p <- p_ggplot_bottlenecks(p, bottleneck.list, flip.x, flip.y)

    # Process Lines
    lines_data <- data.frame()
    for (method in plot$methods) {
      line_matrix <- plot$line_matrices[[method]]
      if (is_infinite_null(line_matrix)) {
        next
      }

      # Sort matrix depending on corner / flipping
      on.x <- line_matrix[, 1]
      on.y <- line_matrix[, 2]
      # Corner 2 or 3 || flip.x XOR flip.y
      if ((!is.null(corner) && corner %in% c(2, 3)) || xor(flip.x, flip.y)) {
        on.y <- -1 * on.y
      }
      line_matrix <- line_matrix[order(on.x, on.y),]

      tmp_df <- as.data.frame(line_matrix)
      names(tmp_df) <- c("x", "y")
      tmp_df$method <- p_pretty_name(method)
      tmp_df$color <- line_colors[[method]]
      tmp_df$line <- line_types[[method]]
      lines_data <- rbind(lines_data, tmp_df)
    }

    if (nrow(lines_data) > 0) {
      aesthetics <- unique(lines_data[c("method", "color", "line")])
      method_colors <- setNames(aesthetics$color, aesthetics$method)
      method_lines <- setNames(aesthetics$line, aesthetics$method)

      p <- p +
        geom_line(data = lines_data, linewidth = line_width,
                  aes(x = .data[['x']], y = .data[['y']],
                      color = .data[['method']], linetype = .data[['method']]))
      p <- p + scale_color_manual(values = method_colors)
      p <- p + scale_linetype_manual(values = method_lines)
    }

    print(p)
    if (pdf) {
      dev.off()
      cat("")
    }

    return(p)
  }

p_ggplot_bottlenecks <-
  function (p, bottleneck.list, flip.x, flip.y) {
    segments_hor <- NULL
    for (name in names(bottleneck.list[['hor']])) {
      coord <- unlist(bottleneck.list[['hor']][[name]])
      segments_hor <- rbind(segments_hor, coord)

      adjust <- abs(coord[2] - coord[1]) / 30
      coord.x <- ifelse(flip.x, coord[2] + adjust, coord[1] - adjust)
      p <- p + annotate(geom = "text", x = coord.x, y = coord[3], label = name, color = "grey")
    }
    if (!is.null(segments_hor)) {
      colnames(segments_hor) <- c("x", "xend", "y", "yend")
      p <- p + geom_segment(data = segments_hor, linetype = "dashed", color = "grey",
                            aes(x = .data[['x']], y = .data[['y']],
                                xend = .data[['xend']], yend = .data[['yend']]))
    }

    segments_ver <- NULL
    for (name in names(bottleneck.list[['ver']])) {
      coord <- unlist(bottleneck.list[['ver']][[name]])
      segments_ver <- rbind(segments_ver, coord)

      adjust <- abs(coord[4] - coord[3]) / 50
      coord.y <- ifelse(flip.y, coord[4] + adjust, coord[3] - adjust)
      p <- p + annotate(geom = "text", x = coord[1], y = coord.y, label = name, color = "grey")
    }
    if (!is.null(segments_ver)) {
      colnames(segments_ver) <- c("x", "xend", "y", "yend")
      p <- p + geom_segment(data = segments_ver, linetype = "dashed", color = "grey",
                            aes(x = .data[['x']], y = .data[['y']],
                                xend = .data[['xend']], yend = .data[['yend']]))
    }

    return(p)
  }

p_ggplot_outline <-
  function (p, plot) {
    p <- p + geom_vline(xintercept = plot$scope.theo[1], linetype = "dashed", color = "grey")
    p <- p + geom_vline(xintercept = plot$scope.theo[2], linetype = "dashed", color = "grey")
    p <- p + geom_hline(yintercept = plot$scope.theo[3], linetype = "dashed", color = "grey")
    p <- p + geom_hline(yintercept = plot$scope.theo[4], linetype = "dashed", color = "grey")
    return(p)
  }

get_plot_params <-
  function () {
    # Get the line colors and types from the global env, allowing user changes
    line_colors <- mget("line.colors", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    line_types <- mget("line.types", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    # Append default for missing values
    line_colors <- append(line_colors, line.colors)
    line_types <- append(line_types, line.types)
    # The same for line width and point-type/-color
    line_width <- mget("line.width", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(line_width)) {
      line_width <- line.width
    }
    point_type <- mget("point.type", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(point_type)) {
      point_type <- point.type
    }
    point_color <- mget("point.color", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    point_color <- tryCatch({
      col2rgb(point_color); point_color
    }, error = function (e) return(point.color))
    point_size <- mget("point.SIZE", envir = .GlobalEnv, ifnotfound = "notfound")[[1]]
    if (!is.numeric(point_size)) {
      point_size <- point.size
    }

    return(list(line_colors, line_types, line_width, point_type, point_color, point_size))
  }

is_infinite_null <-
  function (line) {
    if (is.null(line)) {
      return(TRUE)
    }

    # LH line
    if (typeof(line) == "double" && !all(is.finite(line))) {
      return(TRUE)
    }
    # 'LM' style line
    if (typeof(line) == "list" && !all(is.finite(line$coefficients))) {
      return(TRUE)
    }
    return(FALSE)
  }

p_display_summary <-
  function (summary, pdf = FALSE, path = NULL) {
    if (pdf) {
      p_display_summary_pdf(summary, path)
    } else {
      p_display_summary_screen(summary)
    }
  }

p_display_summary_pdf <-
  function (summary, path) {
    x.name <- summary$names[1]
    y.name <- summary$names[2]

    p_new_pdf("summary", p_generate_title(x.name, y.name), path, paper = "A4r")

    # Set page layout
    layout(matrix(c(1, 2), ncol = 1), heights = c(1, 3))
    par(family = "mono")

    # Plot the 2 dataframes
    textplot(p_pretty_global(summary$global),
             cex = 1, halign = "left", mar = c(0, 0, 4, 0), rmar = 1)
    title <- paste("NCA Parameters :", p_generate_title(x.name, y.name))
    par(family = "")
    title(title, cex.main = 1)
    par(family = "mono")

    if (ncol(summary$params) == 0) {
      textplot(" No NCA parameters available because only OLS selected\n",
               cex = 1, halign = "left")
    } else {
      textplot(p_pretty_params(summary$params),
               cex = 1, halign = "left", mar = c(0, 0, 0, 0), rmar = 1)
    }

    dev.off()
    cat("")
  }

p_display_summary_screen <-
  function (summary) {
    x.name <- summary$names[1]
    y.name <- summary$names[2]
    title <- paste("NCA Parameters :", p_generate_title(x.name, y.name))

    cat("\n", strrep('-', dash.count), "\n", sep = "")
    message(title)
    cat(strrep('-', dash.count), "\n", sep = "")
    print(p_pretty_global(summary$global))
    cat("\n")

    if (ncol(summary$params) == 0) {
      message(" No NCA parameters available because only OLS selected\n")
    } else {
      print(p_pretty_params(summary$params))
    }
    cat("\n")
  }

p_display_summary_simple <-
  function (summaries) {
    if (ncol(summaries[[1]]$params) == 0) {
      warning("\n No effect sizes available because only OLS selected\n", call. = FALSE)
      return()
    }

    simple <- matrix(nrow = length(names(summaries)),
                     ncol = 2 * length(colnames(summaries[[1]]$params)))

    for (i in seq_along(names(summaries))) {
      x.name <- names(summaries)[i]
      tmp <- summaries[[x.name]]$params
      for (j in seq_along(colnames(tmp))) {
        tmp.2 <- unlist(tmp[2, j])
        tmp.6 <- unlist(tmp[6, j])
        simple[i, 2 * j - 1] <- ifelse(is.nan(tmp.2), NA, sprintf("%.2f", tmp.2))
        simple[i, 2 * j] <- ifelse(is.nan(tmp.6), NA, sprintf("%.3f", tmp.6))
      }
    }

    colnames <- colnames(summaries[[1]]$params)
    colnames(simple) <- c(rbind(colnames, rep("p", length(colnames))))
    rownames(simple) <- names(summaries)

    # Remove columns if no tests are present
    simple[simple == "NA"] <- NA
    colnames(simple)[colSums(!is.na(simple)) == 0] <- ''

    cat("\n", strrep('-', dash.count), "\n", sep = "")
    message("Effect size(s):")
    print(simple, na.print = "", quote = FALSE)
    cat(strrep('-', dash.count), "\n\n", sep = "")
  }

p_summary <-
  function (analyses, loop.data) {
    # Do the 'global statistics'
    obs <- min(length(loop.data$x), length(loop.data$y))
    emp <- loop.data$scope.emp
    scope.emp.area <- (emp[2] - emp[1]) * (emp[4] - emp[3])

    if (identical(loop.data$scope.theo, loop.data$scope.emp)) {
      mat1 <- matrix(ncol = 1, nrow = 6)
      mat1[1, 1] <- obs
      mat1[2, 1] <- loop.data$scope.area
      mat1[3:6] <- loop.data$scope.emp
      colnames(mat1) <- c("")
      rownames(mat1) <- p_GLOBAL_NAMES
    } else {
      mat1 <- matrix(ncol = 2, nrow = 6)
      mat1[1,] <- c(obs, NA)
      mat1[2,] <- c(scope.emp.area, loop.data$scope.area)
      mat1[3:6, 1] <- loop.data$scope.emp
      mat1[3:6, 2] <- loop.data$scope.theo
      colnames(mat1) <- c("", " ")
      rownames(mat1) <- replace(p_GLOBAL_NAMES, 2, "Scope  emp / theo")
    }

    # And the parameter statistics (order corresponds to p_PARAM_NAMES)
    data2 <- c(sapply(analyses, function (x) x$ceiling),
               sapply(analyses, function (x) x$effect),
               sapply(analyses, function (x) x$above),
               sapply(analyses, function (x) x$slope),
               sapply(analyses, function (x) x$intercept),
               sapply(analyses, function (x) x$p),
               sapply(analyses, function (x) x$p_accuracy),
               sapply(analyses, function (x) NA),
               sapply(analyses, function (x) x$complexity),
               sapply(analyses, function (x) x$fit),
               sapply(analyses, function (x) x$accuracy),
               sapply(analyses, function (x) x$accuracy_nof),
               sapply(analyses, function (x) x$noise_pct),
               sapply(analyses, function (x) x$noise_nof),
               sapply(analyses, function (x) x$exceptions_pct),
               sapply(analyses, function (x) x$exceptions_nof),
               sapply(analyses, function (x) x$support_pct),
               sapply(analyses, function (x) x$support_nof),
               sapply(analyses, function (x) x$spread),
               sapply(analyses, function (x) x$sharpness),
               sapply(analyses, function (x) NA),
               sapply(analyses, function (x) x$ineffs$abs),
               sapply(analyses, function (x) x$ineffs$rel),
               sapply(analyses, function (x) x$ineffs$x),
               sapply(analyses, function (x) x$ineffs$y))
    mat2 <- t(matrix(data2[names(data2) != "ols"], ncol = length(p_PARAM_NAMES)))
    colnames(mat2) <- names(analyses)[names(analyses) != "ols"]
    rownames(mat2) <- p_PARAM_NAMES

    names <- c(colnames(loop.data$x), colnames(loop.data$y))

    return(list(global = mat1, params = mat2, names = names))
  }

p_pretty_global <-
  function (global) {
    digits <- p_get_digits(global[3:6,])

    pretty <- matrix(ncol = ncol(global), nrow = 6)
    for (row in 1:6) {
      pretty[row, 1] <- ifelse(row == 1,
                               p_pretty_number(global[row, 1], " ", prec = 0, useSpaces = TRUE),
                               p_pretty_number(global[row, 1], "", digits, TRUE))
      if (ncol(global) > 1) {
        pretty[row, 2] <- ifelse(row == 1, "",
                                 p_pretty_number(global[row, 2], "", digits, TRUE))
      }
    }

    colnames(pretty) <- rep(" ", ncol(pretty))
    rownames(pretty) <- rownames(global)

    return(as.data.frame(pretty))
  }

p_pretty_params <-
  function (params) {
    # Extra columns for the NOF items
    pretty <- data.frame(params)
    ids <- newnames <- NULL
    for (i in seq_len(ncol(params))) {
      ids <- c(ids, i, i)
      newnames <- c(newnames, colnames(params)[i], " ")
    }
    pretty <- pretty[, ids]
    pretty[, c(FALSE, TRUE)] <- ""
    colnames(pretty) <- newnames

    for (row in seq_len(nrow(params))) {
      rowname <- rownames(params)[row]
      for (col in seq_len(ncol(params))) {
        item <- unlist(params[row, col])
        if (rowname %in% c('Complexity', '# above')) {
          pretty[row, 2 * col - 1] <- p_pretty_number(item, prec = 0, useSpaces = TRUE)
        }
        else if (rowname == "Accuracy nof") {
          pretty["Ceiling accuracy", 2 * col] <- p_nof(item)
        }
        else if (rowname == "Noise nof") {
          pretty["Noise", 2 * col] <- p_nof(item)
        }
        else if (rowname == "Exceptions nof") {
          pretty["Exceptions", 2 * col] <- p_nof(item)
        }
        else if (rowname == "Support nof") {
          pretty["Support", 2 * col] <- p_nof(item)
        }
        else if (rowname %in% c('Ceiling accuracy', 'Fit', 'Noise', 'Exceptions', 'Support')) {
          if (!p_is_number(item)) {
            pretty[row, 2 * col - 1] <- ifelse(rowname == 'Fit' && is.na(item), "NA    ", "")
          } else if (item %% 1 == 0) {
            item <- p_pretty_number(item, "", 0)
            pretty[row, 2 * col - 1] <- sprintf("%s  %% ", item)
          } else {
            item <- p_pretty_number(item, "", 1)
            pretty[row, 2 * col - 1] <- sprintf("%s%% ", item)
          }
        }
        else if (rowname %in% c("Spread", "Sharpness")) {
          pretty[row, 2 * col - 1] <- p_pretty_number(item, prec = 2, useSpaces = TRUE)
        }
        else {
          pretty[row, 2 * col - 1] <- p_pretty_number(item)
        }
      }
    }

    drop.names <- c('Noise nof', 'Exceptions nof', 'Support nof', 'Accuracy nof')
    if (all(pretty['p-value',] == "")) {
      drop.names <- c(drop.names, 'p-value', 'p-accuracy')
    }
    pretty <- pretty[!(rownames(pretty) %in% drop.names), , drop = FALSE]

    return(pretty)
  }

HIDDEN <- 'hidden'
SHOWN <- 'shown'

nca_outliers <- function (data, x, y, ceiling = NULL,
                          corner = NULL, flip.x = FALSE, flip.y = FALSE,
                          scope = NULL,
                          k = 1, min.dif = 1e-2, max.results = 25,
                          plotly = FALSE, condensed = FALSE) {
  # Cleans up any cluster registration
  p_cluster_cleanup()

  input.ok <- p_check_input(x, y, ceiling)
  if (isFALSE(input.ok)) {
    return()
  }
  if (is.null(input.ok)) {
    ceiling <- "ce_fdh"
  }
  data <- as.data.frame(p_validate_clean(data, x, y, TRUE))
  x <- colnames(data)[1]
  y <- colnames(data)[2]

  # Never do purity
  p_setenv(p_skip_purity)

  model <- nca_analysis(data, x, y, ceilings = ceiling, corner = corner,
                        flip.x = flip.x, flip.y = flip.y, scope = scope)
  eff.or <- nca_extract(model, x, ceiling, 'Effect size')

  params <- list(model = model, x = x, y = y, ceiling = ceiling,
                 corner = corner, flip.x = flip.x, flip.y = flip.y,
                 scope = scope, eff.or = eff.or,
                 Scope = nca_extract(model, x, ceiling, 'Scope'),
                 x.min = nca_extract(model, x, ceiling, 'Xmin'),
                 x.max = nca_extract(model, x, ceiling, 'Xmax'),
                 y.min = nca_extract(model, x, ceiling, 'Ymin'),
                 y.max = nca_extract(model, x, ceiling, 'Ymax'),
                 min.dif = min.dif, peers = p_aggregate_peers(model$peers, 1))

  org_outliers <- p_get_outliers(data, params, 1)
  if (k == 1 && is.null(org_outliers)) {
    message("\nNo outliers identified")
    Sys.unsetenv(p_skip_purity)
    return()
  }
  outliers <- p_format_outliers(org_outliers, max.results, 1, min.dif, condensed)

  if (plotly && !is.null(outliers)) {
    points <- data[outliers[, 1], c(x, y)]
    labels <- paste0(rownames(points), '<br>diff ', outliers[, 5])
    marks <- paste0(outliers[, 6], outliers[, 7])
    points <- cbind(points, labels = labels, marks)
    points <- points[points[, 4] != '', 1:3]
    p_display_plotly(model$plots[[1]], points, NULL, name = 'outlier')
  }

  if (k == 1) {
    # If k == 1, all outliers are shown in plotly so we have to slice here
    outliers <- outliers[1:min(nrow(outliers), max.results),]
    # Don't show 'scope' column if a theoretical scope is defined
    if (!is.null(scope)) {
      outliers <- outliers[, names(outliers) != 'scope']
    }
    Sys.unsetenv(p_skip_purity)
    return(outliers)
  }

  outliers <- p_get_outliers(data, params, k, org_outliers)
  if (is.null(outliers) || nrow(outliers) == 0) {
    message("\nNo outliers identified")
    Sys.unsetenv(p_skip_purity)
    return()
  }

  outliers <- p_format_outliers(outliers, max.results, k, min.dif, condensed)
  # Don't show 'scope' column if a theoretical scope is defined
  if (!is.null(scope)) {
    outliers <- outliers[, names(outliers) != 'scope']
  }
  Sys.unsetenv(p_skip_purity)
  return(outliers)
}


p_check_input <- function (x, y, ceiling) {
  if (length(x) != 1) {
    message()
    message("Outlier detection needs a single independent variable")
    return(FALSE)
  }
  if (length(y) != 1) {
    message()
    message("Outlier detection needs a single dependent variable")
    return(FALSE)
  }
  if (is.null(ceiling)) {
    return(NULL)
  }
  if (length(ceiling) != 1) {
    message()
    message("Outlier detection needs a single ceiling")
    return(FALSE)
  }

  tmp <- c(p_ceilings_step, p_ceilings_line)
  allowed <- tmp[tmp != "ols"]
  if (!(ceiling %in% allowed)) {
    message()
    message(paste("Outlier detection does not work with", ceiling))
    return(FALSE)
  }

  return(TRUE)
}


p_get_outliers <- function (data, params, k, org_outliers = NULL) {
  if (k > length(rownames(data))) {
    k <- length(rownames(data))
    warning(paste("Reduced k to", length(rownames(data))), call. = F)
  }
  combos <- p_get_combos(data, params, k)

  # Start a cluster if needed
  condition <- p_dopar_condition(nrow(combos) > 250)
  p_start_cluster(condition)

  idx <- NULL
  ids <- seq(1, nrow(combos), max(1, round(nrow(combos) / 50)))
  outliers <- foreach(idx = seq_len(nrow(combos)), .combine = rbind) %dopar% {
    if (condition && idx %in% ids) {
      cat(".")
    }
    return(p_get_outlier(data, combos[idx,], params, k))
  }
  if (condition) {
    message("\rDone", strrep(' ', 50), "\n")
  }

  if (is.null(outliers)) {
    p_stop_cluster()

    return(org_outliers)
  }

  new_outliers <- foreach(idx = seq_len(nrow(outliers)), .combine = rbind) %dopar% {
    outlier <- outliers[idx,]
    old_combo <- outlier[8]
    tmp <- org_outliers

    f <- function (n) {
      dif.rel <- tmp[tmp[, "combo"] %in% n,]$dif.rel
      return(abs(ifelse(is.null(dif.rel), 0, dif.rel)))
    }

    ord <- unlist(sapply(unlist(old_combo), f))
    tmp <- list(unlist(old_combo)[order(-ord)])
    outlier[1] <- p_get_names(unlist(tmp), k)$outliers

    return(outlier)
  }

  p_stop_cluster()

  return(rbind(org_outliers, new_outliers))
}


p_get_all_names <- function (data, peers, params, k) {
  all.names <- rownames(peers)
  tmp.var <- c(params$x, params$x, params$y, params$y)
  scope <- c(params$x.min, params$x.max, params$y.min, params$y.max)
  for (idx in 1:4) {
    tmp <- data[data[[tmp.var[idx]]] %in% scope[[idx]],]
    if (k >= nrow(tmp)) {
      all.names <- c(all.names, rownames(tmp))
    }
  }

  return(all.names)
}


p_get_combos <- function (data, params, k) {
  # For COLS and QR we need all the points
  if (params$ceiling[1] %in% p_no_peer_line) {
    return(t(combn(rownames(data), k)))
  }

  all.names <- p_get_all_names(data, params$peers, params, k)

  counter <- k
  while (counter > 1) {
    data <- data[!(row.names(data) %in% all.names),]
    model <- nca_analysis(data, params$x, params$y, ceilings = params$ceiling,
                          corner = params$corner, flip.x = params$flip.x,
                          flip.y = params$flip.y, scope = params$scope)

    all.names <- c(
      all.names,
      p_get_all_names(data, p_aggregate_peers(model$peers, 1), params, k)
    )

    counter <- counter - 1
  }

  all.names <- unique(all.names)
  if (k == 1) {
    return(matrix(all.names, ncol = 1))
  }

  combos <- NULL
  counter <- k
  while (counter > 1) {
    tmp <- t(combn(all.names, counter))
    tmp <- cbind(tmp, matrix(NA, ncol = k - counter, nrow = nrow(tmp)))
    combos <- rbind(combos, tmp)
    counter <- counter - 1
  }

  return(combos)
}


p_get_outlier <- function (data, combo, params, k) {
  combo <- combo[!is.na(combo)]
  data.new <- data[-(which(rownames(data) %in% combo)),]
  values <- p_get_values(data.new, params)
  eff.nw <- values[[1]]; dif.abs <- values[[2]]; dif.rel <- values[[3]]
  global.new <- values[[4]]

  if (round(abs(dif.rel), digits = 2) < params$min.dif) {
    return()
  }

  zone_scope <- p_zone_scope(combo, params, global.new)
  zone <- zone_scope[1]; scope <- zone_scope[2]

  # Extra check: 'ceiling' outliers must be on COLS and C_LP lines
  if (all(k == 1, params$ceiling[1] %in% c("cols", "c_lp"), scope == "")) {
    line <- params$model$plots[[1]]$lines[[1]]
    if (params$ceiling == "cols") {
      line <- unname(coef(line))
    }
    intercept <- line[1]
    slope <- line[2]
    x.value <- data[combo[k], params$x]
    y.value <- data[combo[k], params$y]
    y.calc <- intercept + x.value * slope

    # If not on line, ignore ceiling zone outlier
    if (all.equal(y.value, y.calc) != TRUE) {
      return()
    }
  }

  names <- p_get_names(combo, k)
  values <- list(eff.or = params$eff.or, eff.nw = eff.nw,
                 dif.abs = dif.abs, dif.rel = dif.rel,
                 ceiling = zone, scope = scope, combo = combo)
  return(append(names, values))
}


p_get_values <- function (data.new, params) {
  model.new <- nca_analysis(data.new, params$x, params$y,
                            ceilings = params$ceiling, corner = params$corner,
                            flip.x = params$flip.x, flip.y = params$flip.y,
                            scope = params$scope)
  eff.nw <- nca_extract(model.new, params$x, params$ceiling, 'Effect size')

  dif.abs <- ifelse(is.na(eff.nw), 0, eff.nw - params$eff.or)
  zero.dif.rel <- ifelse(dif.abs < epsilon, 0, Inf)
  dif.rel <- ifelse(params$eff.or < epsilon, zero.dif.rel, 100 * dif.abs / params$eff.or)
  global.new <- c(
    nca_extract(model.new, params$x, params$ceiling, 'Scope'),
    nca_extract(model.new, params$x, params$ceiling, 'Xmin'),
    nca_extract(model.new, params$x, params$ceiling, 'Xmax'),
    nca_extract(model.new, params$x, params$ceiling, 'Ymin'),
    nca_extract(model.new, params$x, params$ceiling, 'Ymax')
  )

  return(list(eff.nw, dif.abs, dif.rel, global.new))
}


p_zone_scope <- function (combo, params, global.new) {
  global.org <- c(params$Scope,
                  params$x.min, params$x.max, params$y.min, params$y.max)
  found_scope <- !all(global.new == global.org)

  if (params$ceiling[1] %in% p_no_peer_line) {
    found_ceiling <- TRUE
  }
  else {
    found_ceiling <- any(combo %in% rownames(params$peers))
  }

  return(c(ifelse(found_ceiling, "X", ""), ifelse(found_scope, "X", "")))
}


p_get_names <- function (combo, k) {
  if (k == 1) {
    name <- combo[1]
  }
  else {
    name <- paste(combo, '-', collapse = " ")
    name <- substr(name, 0, nchar(name) - 2)
  }
  return(list(outliers = name))
}


p_format_outliers <- function (outliers, max.results, k, min.dif, condensed) {
  # The dataset / param combination doesn't produce outliers
  if (is.null(outliers) || nrow(outliers) == 0) {
    return()
  }

  outliers <- data.frame(outliers)

  for (i in 1:7) { outliers[[i]] <- unlist(outliers[[i]]) }
  for (row_idx in seq_len(nrow(outliers))) {
    len <- length(unlist(strsplit(outliers[row_idx, 1], ' - ', fixed = T)))
    outliers[row_idx, 9] <- len
  }
  outliers <- outliers[order(-abs(outliers$dif.rel), -abs(outliers$dif.abs), outliers[, 9]),]
  org.length <- nrow(outliers)

  # If k == 1: all outliers are shown on Plotly
  if (k != 1) {
    # Strip combos where extra observations do not change outcome
    if (nrow(outliers) > 1 && condensed) {
      keep <- 1
      # Remove outliers which are not larger than previous (but keep singles)
      for (row_idx in 2:nrow(outliers)) {
        parts <- unlist(strsplit(outliers[row_idx, 1], ' - ', fixed = T))
        current <- abs(outliers[row_idx, 5])
        prev <- abs(outliers[row_idx - 1, 5])
        if (length(parts) == 1 || abs(current - prev) > min.dif) {
          keep <- c(keep, row_idx)
        }
      }
      outliers <- outliers[keep,]
    }

    outliers <- outliers[1:min(nrow(outliers), max.results),]
  }

  outliers <- outliers[, -c(8, 9)]
  outliers[, c(2, 3, 4)] <- round(outliers[, c(2, 3, 4)], digits = 2)
  outliers[, 5] <- round(outliers[, 5], digits = 1)

  if (org.length > nrow(outliers)) {
    attr(outliers, SHOWN) <- nrow(outliers)
    attr(outliers, HIDDEN) <- org.length - nrow(outliers)
  }
  class(outliers) <- c("nca_outliers", class(outliers))

  rownames(outliers) <- NULL
  return(outliers)
}


print.nca_outliers <- function (x, ...) {
  # NextMethod()
  print(format(x, justify = "left"))

  hidden <- attr(x, HIDDEN)
  shown <- attr(x, SHOWN)
  # Only show when 'full' outliers, not when selection
  if (!is.null(hidden) && shown == nrow(x)) {
    message(sprintf("# Not showing %s possible outliers", hidden))
  }
}

p_scope <-
  function (x, scope) {
    if (is.null(scope)) {
      return(replicate(length(x), NULL))
    } else if (typeof(scope) == "double") {
      return(p_validate_scope_vector(x, scope))
    } else if (typeof(scope) == "list") {
      return(p_validate_scope_list(x, scope))
    }
  }

p_validate_scope_list <-
  function (x, scope) {
    if (length(scope) != length(x)) {
      message()
      stop("The length of scope needs to be equal to the length of x \n",
           "       or a single vector with 4 values!\n\n", call. = F)
    }

    for (i in scope) {
      if (length(i) != 4) {
        message()
        stop("The length of each scope segment needs to be 4!\n\n", call. = F)
      }
    }

    for (i in scope) {
      same <- i[c(3, 4)] == scope[[1]][c(3, 4)]
      same <- same | (is.na(i[c(3, 4)]) & is.na(scope[[1]][c(3, 4)]))
      same[is.na(same)] <- FALSE
      if (!all(same)) {
        warning("Scope is using different Ymin and Ymax for the same Y", call. = F)
        break
      }
    }

    return(scope)
  }

p_validate_scope_vector <-
  function (x, scope) {
    # Duplicate single vector for all Xs
    if (length(scope) == 4) {
      scope <- rep(scope, length(x))
    }

    # Vector with all values, transform to list of vectors
    if (length(scope) == 4 * length(x)) {
      return(split(matrix(scope, nrow = 4), rep(seq_along(x), each = 4)))
    }

    message()
    stop("The length of scope needs to be equal to the length of x \n",
         "       or a single vector with 4 values!\n\n", call. = F)
  }

p_scope <-
  function (x, scope) {
    if (typeof(scope) == "double") {
      return(p_validate_scope_vector(x, scope))
    } else if (typeof(scope) == "list") {
      return(p_validate_scope_list(x, scope))
    } else {
      return(replicate(length(x), NULL))
    }
  }

p_validate_scope_list <-
  function (x, scope) {
    if (length(scope) != length(x)) {
      message()
      stop("The length of scope needs to be equal to the length of x \n",
           "       or a single vector with 4 values!\n\n", call. = F)
    }

    for (i in scope) {
      if (length(i) != 4) {
        message()
        stop("The length of each scope segment needs to be 4!\n\n", call. = F)
      }
    }

    for (i in scope) {
      same <- i[c(3, 4)] == scope[[1]][c(3, 4)]
      same <- same | (is.na(i[c(3, 4)]) & is.na(scope[[1]][c(3, 4)]))
      same[is.na(same)] <- FALSE
      if (!all(same)) {
        warning("Scope is using different Ymin and Ymax for the same Y", call. = F)
        break
      }
    }

    return(scope)
  }

p_validate_scope_vector <-
  function (x, scope) {
    # Duplicate single vector for all Xs
    if (length(scope) == 4) {
      scope <- rep(scope, length(x))
    }

    # Vector with all values, transform to list of vectors
    if (length(scope) == 4 * length(x)) {
      return(split(matrix(scope, nrow = 4), rep(seq_along(x), each = 4)))
    }

    message()
    stop("The length of scope needs to be equal to the length of x \n",
         "       or a single vector with 4 values!\n\n", call. = F)
  }

p_above <-
  function (loop.data, slope, intercept) {
    x <- loop.data$x
    y <- loop.data$y
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y

    # Upper left and lower right
    if ((flip.x == flip.y) && (is.na(slope) || slope < 0)) {
      return(NaN)
    }
    # Lower left and upper right
    if ((flip.x != flip.y) && (is.na(slope) || slope > 0)) {
      return(NaN)
    }

    # Vertical difference between observations and ceiling line
    y.c <- slope * x + intercept
    if (!flip.y) {
      y.diff <- y - y.c
    } else {
      y.diff <- y.c - y
    }

    return(sum(y.diff > 1e-07, na.rm = TRUE))
  }

p_accuracy <-
  function (loop.data, above) {
    nObservations <- min(length(loop.data$x), length(loop.data$y))
    return(100 * (nObservations - above) / nObservations)
  }

p_fit <-
  function (ceiling, fdh_ceiling) {
    if (is.nan(ceiling) || is.nan(fdh_ceiling)) {
      return(NA)
    }
    if (ceiling > fdh_ceiling) {
      return(NA)
    }
    return(100 - 100 * abs(ceiling - fdh_ceiling) / fdh_ceiling)
  }

.onAttach <- function(...) {

    meta  <- packageDescription("NCA")
    year  <- sub("-.*", "", meta$Date)
    title <- meta$Title

    msg1  <- sprintf("Dul, J. & Buijs, G. %s.", year)
    msg2  <- sprintf("%s.", title)
    msg3  <- sprintf("R Package Version %s.\n", meta$Version)
    msg4  <- "URL: https://cran.r-project.org/web/packages/NCA/"

    msg5  <- "This package is based on:"
    msg6  <- "Dul, J. (2026). Necessary Condition Analysis - NCA."
    msg7  <- "Principles and Application. Chapman & Hall/CRC Press."
    msg8  <- "An online version is here:"
    msg9  <- "https://jandul.github.io/NCA/"
    
    msg10 <- "For general information about NCA see:"
    msg11 <- "https://www.erim.nl/nca"

    packageStartupMessage("\nPlease cite the NCA package as:\n\n",
                          strwrap(msg1, indent = 2, exdent = 2), "\n",
                          strwrap(msg2, indent = 2, exdent = 2), "\n",
                          strwrap(msg3, indent = 2, exdent = 2), "\n",
                          strwrap(msg4, indent = 2, exdent = 2), "\n",
                          "\n",
                          strwrap(msg5, indent = 0, exdent = 2), "\n",
                          strwrap(msg6, indent = 2, exdent = 2), "\n",
                          strwrap(msg7, indent = 2, exdent = 2), "\n",
                          strwrap(msg8, indent = 2, exdent = 2), "\n",
                          strwrap(msg9, indent = 4, exdent = 2), "\n",
                          "\n",
                          strwrap(msg10, indent = 0, exdent = 2), "\n",
                          strwrap(msg11, indent = 2, exdent = 2), "\n")
}

# TODO Use functions everywhere

p_get_x <-
  function (intercept, slope, y.value) {
    return((y.value - intercept) / slope)
  }

p_get_y <-
  function (intercept, slope, x.value) {
    return(intercept + slope * x.value)
  }

p_x_on_scope <-
  function (x.value, scope) {
    return(x.value >= min(scope[1:2]) && x.value <= max(scope[1:2]))
  }

p_y_on_scope <-
  function (y.value, scope) {
    return(y.value >= min(scope[3:4]) && y.value <= max(scope[3:4]))
  }

p_through_scope <-
  function (intercept, slope, scope) {
    y.1 <- p_get_y(intercept, slope, scope[1])
    left_side <- p_y_on_scope(y.1, scope)
    y.2 <- p_get_y(intercept, slope, scope[2])
    right_side <- p_y_on_scope(y.2, scope)
    x.3 <- p_get_x(intercept, slope, scope[3])
    floor <- p_x_on_scope(x.3, scope)
    x.4 <- p_get_x(intercept, slope, scope[4])
    roof <- p_x_on_scope(x.4, scope)

    return(left_side || floor || roof || right_side)
  }

p_get_line_matrix <-
  function (intercept, slope, loop.data) {
    scope <- loop.data$scope.theo
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y
    corner <- 1 + flip.x + 2 * flip.y

    y.1 <- p_get_y(intercept, slope, scope[1])
    left_side <- p_y_on_scope(y.1, scope)
    y.2 <- p_get_y(intercept, slope, scope[2])
    right_side <- p_y_on_scope(y.2, scope)
    x.3 <- p_get_x(intercept, slope, scope[3])
    floor <- p_x_on_scope(x.3, scope)
    x.4 <- p_get_x(intercept, slope, scope[4])
    roof <- p_x_on_scope(x.4, scope)

    line <- matrix(ncol = 2, nrow = 0)

    ############################################################################
    # Add start point
    ############################################################################
    if (corner == 1) {
      if (left_side) {
        line <- rbind(line, c(scope[1], scope[3]))
      }
      else if (slope < 0 && roof) {
        line <- rbind(line, c(x.4, scope[3]))
      }
    }
    else if (corner == 2) {
      if (slope > 0 && floor) {
        line <- rbind(line, c(scope[1], scope[3]))
      }
      else if (slope < 0 && roof) {
        line <- rbind(line, c(scope[1], scope[4]))
      }
    }
    else if (corner == 3) {
      if (left_side) {
        line <- rbind(line, c(scope[1], scope[4]))
      }
      if (slope > 0 && floor) {
        line <- rbind(line, c(x.3, scope[4]))
      }
    }
    else if (corner == 4) {
      if (slope > 0 && floor) {
        line <- rbind(line, c(scope[1], scope[3]))
      }
      else if (slope < 0 && roof) {
        line <- rbind(line, c(scope[1], scope[4]))
      }
    }

    ############################################################################
    # Add points on the scope
    ############################################################################
    if (left_side) {
      line <- rbind(line, c(scope[1], y.1))
    }
    if (right_side) {
      line <- rbind(line, c(scope[2], y.2))
    }
    if (floor) {
      line <- rbind(line, c(x.3, scope[3]))
    }
    if (roof) {
      line <- rbind(line, c(x.4, scope[4]))
    }

    ############################################################################
    # Add end point
    ############################################################################
    if (corner == 1) {
      if (slope > 0 && roof) {
        line <- rbind(line, c(scope[2], scope[4]))
      }
    }
    else if (corner == 2) {
      if (slope > 0) {
        if (roof) {
          line <- rbind(line, c(x.4, scope[3]))
        }
        else if (right_side) {
          line <- rbind(line, c(scope[2], scope[3]))
        }
      }
      else if (slope < 0 && right_side) {
        line <- rbind(line, c(scope[2], scope[3]))
      }
    }
    else if (corner == 3) {
      if (slope > 0 && roof) {
        line <- rbind(line, c(scope[2], scope[4]))
      }
      else if (slope < 0 && floor) {
        line <- rbind(line, c(scope[2], scope[3]))
      }
    }
    else if (corner == 4) {
      if (slope > 0 && right_side) {
        line <- rbind(line, c(scope[2], scope[4]))
      }
      else if (slope < 0) {
        if (floor) {
          line <- rbind(line, c(x.3, scope[4]))
        }
        else if (right_side) {
          line <- rbind(line, c(scope[2], scope[4]))
        }
      }
    }

    line <- line[order(line[, 1]),]

    return(line)
  }

p_nca_custom <-
  function (loop.data, bn.data) {
    intercept <- loop.data$custom[loop.data$idx, 1]
    slope <- loop.data$custom[loop.data$idx, 2]

    line <- c(intercept, slope)
    line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
    ceiling <- p_ceiling(loop.data, slope, intercept)
    above <- p_above(loop.data, slope, intercept)

    effect <- ceiling / loop.data$scope.area
    accuracy <- p_accuracy(loop.data, above)
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1
    metrics <- p_purity_empty

    if (!p_through_scope(intercept, slope, loop.data$scope.theo)) {
      name.x <- colnames(loop.data$x)
      name.y <- colnames(loop.data$y)
      warning("Custom line (", intercept, ", ", slope,
              ") is outside the scope of ", name.x, " - ", name.y, call. = FALSE)
    }

    return(list(line = line, line_matrix = line_matrix, peers = NULL,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

nca_difference <-
  function (data1, data2 = NULL, x, y,
            ceilings = c("ce_fdh", "cr_fdh"), common_scope, test.rep = 1000,
            test.type = c("contrast", "independent", "paired")) {
    test.type <- match.arg(test.type)
    p_test_difference_args(data1, data2, x, y, test.rep, test.type)
    if (test.type == "contrast") {
      data2 <- data1
    }
    else {
      x <- c(x, x)
    }
    ceilings <- p_validate_ceilings(ceilings)
    common_scope <- p_validate_common_scope(data1, data2, x, y, common_scope)

    # Never do purity
    p_setenv(p_skip_purity)
    # Always do parallel
    if (p_dopar_condition(test.rep >= 50)) {
      p_setenv(p_parallel_force)
      p_start_cluster(TRUE)
    }

    differences <- list()
    for (ceiling in ceilings) {
      # Compute observations
      observations <- p_observed(data1, data2, x, y, ceiling, common_scope)
      obs <- observations$obs
      obs1 <- observations$obs1
      obs2 <- observations$obs2

      # Compute p-value
      if (test.rep > 0) {
        seeds <- sample(1:2^15, test.rep)
        perm <- p_perm_test(data1, data2, x, y, ceiling, common_scope, test.rep, test.type, seeds)
        tmp <- abs(perm) >= abs(obs)
        p_value <- (sum(tmp) + 1) / (length(tmp) + 1)
        p_value <- max(p_value, 1 / test.rep)
      }
      else {
        p_value <- NA
      }

      differences[[ceiling]] <- list(
        observed1 = obs1, observed2 = obs2, observed = obs, p_value = p_value)
    }

    Sys.unsetenv(p_parallel_start)
    Sys.unsetenv(p_parallel_force)
    p_stop_cluster()
    Sys.unsetenv(p_skip_purity)

    attr(differences, "test.type") <- test.type
    attr(differences, "x") <- x
    class(differences) <- c("nca_difference", class(differences))

    return(differences)
  }

p_test_difference_args <-
  function (data1, data2, x, y, test.rep, test.type) {
    # --- correctness assertions (not robustness) ---
    if (test.type == "contrast") {
      if (length(x) != 2) {
        stop("For 'contrast', all 'x' must be a vector with 2 values.", call. = F)
      }
      if (p_invalid_variable(x, data1)) {
        stop("For 'contrast', all 'x' must be columns in data1.", call. = F)
      }
      if (p_invalid_variable(y, data1)) {
        stop("For 'contrast', 'y' must be a column in data1.", call. = F)
      }
    }
    if (test.type == "independent") {
      if (length(x) != 1) {
        stop("For 'independent', 'x' must be a single column name.", call. = F)
      }
      if (p_invalid_variable(x, data1) || p_invalid_variable(x, data2)) {
        stop("For 'independent', both data1 and data2 must",
             "contain columns 'x'.", call. = F)
      }
      if (p_invalid_variable(y, data1) || p_invalid_variable(y, data2)) {
        stop("For 'independent', both data1 and data2 must",
             "contain columns 'y'.", call. = F)
      }
      if (!identical(names(data1), names(data2))) {
        stop("For 'independent', data1 and data2 must have",
             "identical column sets and order (so rbind aligns).", call. = F)
      }
    }
    if (test.type == "paired") {
      if (length(x) != 1) {
        stop("For 'paired', 'x' must be a single column name.", call. = F)
      }
      if (p_invalid_variable(x, data1) || p_invalid_variable(x, data2)) {
        stop("For 'paired', both data1 and data2 must",
             "contain columns 'x'.", call. = F)
      }
      if (p_invalid_variable(y, data1) || p_invalid_variable(y, data2)) {
        stop("For 'paired', both data1 and data2 must",
             "contain columns 'y'.", call. = F)
      }
      if (nrow(data1) != nrow(data2)) {
        stop("For 'paired', data1 and data2 must have the same number of ",
             "rows (same cases).", call. = F)
      }
      if (!identical(names(data1), names(data2))) {
        stop("For 'paired', data1 and data2 must ",
             "have identical column sets and order.", call. = F)
      }
    }

    if (test.rep <= 0) {
      warning("No permutation test: test.rep is 0")
    }
  }

p_observed <-
  function (data1, data2, x, y, ceiling, common_scope) {
    model1 <- nca_analysis(data1, x[1], y, ceilings = ceiling, scope = common_scope)
    effect1 <- model1$`Effect size`[[1]][[1]]
    model2 <- nca_analysis(data2, x[2], y, ceilings = ceiling, scope = common_scope)
    effect2 <- model2$`Effect size`[[1]][[1]]

    obs1 <- effect1
    obs2 <- effect2

    return(list(obs = obs2 - obs1, obs1 = obs1, obs2 = obs2))
  }

p_perm_test <-
  function (data1, data2, x, y, ceiling, common_scope, test.rep, test.type, seeds) {
    ids <- seq(1, test.rep, max(1, round(test.rep / 50)))
    cat(sprintf("Doing %-6s ", ceiling))

    i <- NULL
    perm <- foreach(i = 1:test.rep) %dopar% {
      if (i %in% ids) {
        cat(".")
      }
      set.seed(seeds[i])
      if (test.type == "contrast") {
        # Permutation logic: randomly swap X1 and X2 within each case.
        data1_perm <- data1
        swap_mask <- sample(c(TRUE, FALSE), nrow(data1), replace = TRUE)
        data1_perm[swap_mask, c(x[1], x[2])] <- data1_perm[swap_mask, c(x[2], x[1])]
        data2_perm <- data1_perm
      }
      else if (test.type == "independent") {
        # Permutation logic: randomly shuffle group labels across pooled sample
        combined_data <- rbind(data1, data2)
        n1 <- nrow(data1); n2 <- nrow(data2)
        labels <- c(rep(1, n1), rep(2, n2))
        lab_perm <- sample(labels)
        data1_perm <- combined_data[lab_perm == 1, , drop = FALSE]
        data2_perm <- combined_data[lab_perm == 2, , drop = FALSE]
      }
      else if (test.type == "paired") {
        # Permutation logic: randomly swap time labels within each case
        flips <- sample(c(TRUE, FALSE), nrow(data1), replace = TRUE)
        data1_perm <- data1
        data2_perm <- data2
        data1_perm[flips,] <- data2[flips,]
        data2_perm[flips,] <- data1[flips,]
      }

      model1 <- nca_analysis(data1_perm, x[1], y, ceilings = ceiling, scope = common_scope)
      eff1 <- model1$`Effect size`[[1]][[1]]
      model2 <- nca_analysis(data2_perm, x[2], y, ceilings = ceiling, scope = common_scope)
      eff2 <- model2$`Effect size`[[1]][[1]]

      return(eff2 - eff1)
    }
    cat("\n")

    return(unlist(perm))
  }

print.nca_difference <-
  function (x, ...) {
    differences <- x
    test.type <- attr(x, "test.type")
    x <- attr(differences, "x")
    ceilings <- names(differences)
    if (test.type != "contrast") {
      x <- paste0(x, c(".1", ".2"))
    }

    simple <- matrix(nrow = 3, ncol = 2 * length(differences))
    col.names <- NULL
    for (i in seq_along(differences)) {
      ceiling <- ceilings[i]
      difference <- differences[[ceiling]]
      col.names <- c(col.names, ceiling, 'p')

      simple[1, 2 * i - 1] <- round(difference$observed1, digits = 2)
      simple[2, 2 * i - 1] <- round(difference$observed2, digits = 2)
      simple[3, 2 * i - 1] <- round(difference$observed, digits = 2)
      simple[3, 2 * i] <- round(difference$p_value, digits = 3)
    }
    colnames(simple) <- col.names
    rownames(simple) <- c(x, 'Difference')

    cat("\n", strrep('-', dash.count), "\n", sep = "")
    message("Effect size (difference):")
    print(simple, na.print = "", quote = FALSE)
    cat(strrep('-', dash.count), "\n\n", sep = "")
  }

p_invalid_variable <-
  function (vv, df) {
    invalid <- FALSE
    for (v in vv) {
      if (is.numeric(v)) {
        invalid <- invalid || (v < 1 || v > ncol(df))
      }
      else {
        invalid <- invalid || !(v %in% names(df))
      }
    }
    return(invalid)
  }

p_validate_common_scope <-
  function (data1, data2, x, y, common_scope) {
    message <- paste0("common_scope should be a vector of 4 numeric values,\n",
                      "       encapsulating both datasets")
    if (is.vector(common_scope, mode = "list")) {
      stop(message, call. = FALSE)
    }
    common_scope <- as.vector(common_scope)
    if (length(common_scope) != 4 || !is.numeric(common_scope)) {
      stop(message, call. = FALSE)
    }
    if (common_scope[1] > min(data1[, x[1]], data2[, x[2]]) ||
      common_scope[2] < max(data1[, x[1]], data2[, x[2]]) ||
      common_scope[3] > min(data1[, y], data2[, y]) ||
      common_scope[4] < max(data1[, y], data2[, y])) {
      stop(message, call. = FALSE)
    }
    return(common_scope)
  }

nca_difference <-
  function (data1, data2 = NULL, x, y,
            ceilings = c("ce_fdh", "cr_fdh"), scope = NULL, test.rep = 1000,
            test.type = c("contrast", "independent", "paired")) {
    test.type <- match.arg(test.type)
    p_test_difference_args(data1, data2, x, y, test.rep, test.type)
    if (test.type == "contrast") {
      data2 <- data1
    }
    else {
      x <- c(x, x)
    }
    ceilings <- p_validate_ceilings(ceilings)
    scopes <- p_scope(x, scope)

    # Always do parallel
    if (p_dopar_condition(test.rep >= 50)) {
      p_setenv(p_parallel_force)
      p_start_cluster(TRUE)
    }
    # Never do purity
    p_setenv(p_skip_purity)

    differences <- list()
    for (ceiling in ceilings) {
      # Compute observations
      observations <- p_observed(data1, data2, x, y, ceiling, scopes)
      obs <- observations$obs
      obs1 <- observations$obs1
      obs2 <- observations$obs2

      # Compute p-value
      if (test.rep > 0) {
        perm <- p_perm_test(data1, data2, x, y, ceiling, scopes, test.rep, test.type)
        tmp <- abs(perm) >= abs(obs)
        p_value <- (sum(tmp) + 1) / (length(tmp) + 1)
        p_value <- max(p_value, 1/test.rep)
      }
      else {
        p_value <- NA
      }

      differences[[ceiling]] <- list(
        observed1 = obs1, observed2 = obs2, observed = obs, p_value = p_value)
    }

    Sys.unsetenv(p_parallel_start)
    Sys.unsetenv(p_parallel_force)
    p_stop_cluster()
    Sys.unsetenv(p_skip_purity)

    attr(differences, "test.type") <- test.type
    attr(differences, "x") <- x
    class(differences) <- c("nca_difference", class(differences))

    return(differences)
  }

p_test_difference_args <-
  function (data1, data2, x, y, test.rep, test.type) {
    # --- correctness assertions (not robustness) ---
    if (test.type == "contrast") {
      if (length(x) != 2) {
        stop("For 'contrast', all 'x' must be a vector with 2 values.", call. = F)
      }
      if (p_invalid_variable(x, data1)) {
        stop("For 'contrast', all 'x' must be columns in data1.", call. = F)
      }
      if (p_invalid_variable(y, data1)) {
        stop("For 'contrast', 'y' must be a column in data1.", call. = F)
      }
    }
    if (test.type == "independent") {
      if (length(x) != 1) {
        stop("For 'independent', 'x' must be a single column name.", call. = F)
      }
      if (p_invalid_variable(x, data1) || p_invalid_variable(x, data2)) {
        stop("For 'independent', both data1 and data2 must",
             "contain columns 'x'.", call. = F)
      }
      if (p_invalid_variable(y, data1) || p_invalid_variable(y, data2)) {
        stop("For 'independent', both data1 and data2 must",
             "contain columns 'y'.", call. = F)
      }
      if (!identical(names(data1), names(data2))) {
        stop("For 'independent', data1 and data2 must have",
             "identical column sets and order (so rbind aligns).", call. = F)
      }
    }
    if (test.type == "paired") {
      if (length(x) != 1) {
        stop("For 'paired', 'x' must be a single column name.", call. = F)
      }
      if (p_invalid_variable(x, data1) || p_invalid_variable(x, data2)) {
        stop("For 'paired', both data1 and data2 must",
             "contain columns 'x'.", call. = F)
      }
      if (p_invalid_variable(y, data1) || p_invalid_variable(y, data2)) {
        stop("For 'paired', both data1 and data2 must",
             "contain columns 'y'.", call. = F)
      }
      if (nrow(data1) != nrow(data2)) {
        stop("For 'paired', data1 and data2 must have the same number of ",
             "rows (same cases).", call. = F)
      }
      if (!identical(names(data1), names(data2))) {
        stop("For 'paired', data1 and data2 must ",
             "have identical column sets and order.", call. = F)
      }
    }

    if (test.rep <= 0) {
      warning("No permutation test: test.rep is 0")
    }
  }

p_observed <-
  function (data1, data2, x, y, ceiling, scopes) {
    model1 <- nca_analysis(data1, x[1], y, ceilings = ceiling, scope = scopes[[1]])
    effect1 <- model1$`Effect size`[[1]][[1]]
    model2 <- nca_analysis(data2, x[2], y, ceilings = ceiling, scope = scopes[[2]])
    effect2 <- model2$`Effect size`[[1]][[1]]

    obs1 <- effect1
    obs2 <- effect2

    return(list(obs = obs2 - obs1, obs1 = obs1, obs2 = obs2))
  }

p_perm_test <-
  function (data1, data2, x, y, ceiling, scopes, test.rep, test.type) {
    ids <- seq(1, test.rep, max(1, round(test.rep / 50)))
    cat(sprintf("Doing %-6s ", ceiling))

    i <- NULL
    perm <- foreach(i = 1:test.rep) %dopar% {
      if (i %in% ids) {
        cat(".")
      }
      if (test.type == "contrast") {
        # Permutation logic: randomly swap X1 and X2 within each case.
        data1_perm <- data1
        swap_mask <- sample(c(TRUE, FALSE), nrow(data1), replace = TRUE)
        # temp <- data1_perm[[x[1]]]
        # data1_perm[[x[1]]][swap_mask] <- data1_perm[[x[2]]][swap_mask]
        # data1_perm[[x[2]]][swap_mask] <- temp[swap_mask]
        data1_perm[swap_mask, c(x[1], x[2])] <- data1_perm[swap_mask, c(x[2], x[1])]
        data2_perm <- data1_perm
      }
      else if (test.type == "independent") {
        # Permutation logic: randomly shuffle group labels across pooled sample
        combined_data <- rbind(data1, data2)
        # perm_indices <- sample(nrow(combined_data))
        # data1_perm <- combined_data[perm_indices[1:nrow(data1)],]
        # data2_perm <- combined_data[perm_indices[(nrow(data1) + 1):nrow(combined_data)],]
        n1 <- nrow(data1); n2 <- nrow(data2)
        labels <- c(rep(1, n1), rep(2, n2))
        lab_perm <- sample(labels)
        data1_perm <- combined_data[lab_perm == 1, , drop = FALSE]
        data2_perm <- combined_data[lab_perm == 2, , drop = FALSE]
      }
      else if (test.type == "paired") {
        # Permutation logic: randomly swap time labels within each case
        # sign_flip <- sample(c(-1, 1), nrow(data1), replace = TRUE)
        # data1_perm <- data1
        # data2_perm <- data1 + sign_flip * (data2 - data1)
        flips <- sample(c(TRUE, FALSE), nrow(data1), replace = TRUE)
        data1_perm <- data1
        data2_perm <- data2
        data1_perm[flips,] <- data2[flips,]
        data2_perm[flips,] <- data1[flips,]
      }

      model1 <- nca_analysis(data1_perm, x[1], y, ceilings = ceiling, scope = scopes[[1]])
      eff1 <- model1$`Effect size`[[1]][[1]]
      model2 <- nca_analysis(data2_perm, x[2], y, ceilings = ceiling, scope = scopes[[2]])
      eff2 <- model2$`Effect size`[[1]][[1]]

      return(eff2 - eff1)
    }
    cat("\n")

    return(unlist(perm))
  }

print.nca_difference <-
  function (x, ...) {
    differences <- x
    test.type <- attr(x, "test.type")
    x <- attr(differences, "x")
    ceilings <- names(differences)
    if (test.type != "contrast") {
      x <- paste0(x, c(".1", ".2"))
    }

    simple <- matrix(nrow = 3, ncol = 2 * length(differences))
    col.names <- NULL
    for (i in seq_along(differences)) {
      ceiling <- ceilings[i]
      difference <- differences[[ceiling]]
      col.names <- c(col.names, ceiling, 'p')

      simple[1, 2 * i - 1] <- round(difference$observed1, digits = 2)
      simple[2, 2 * i - 1] <- round(difference$observed2, digits = 2)
      simple[3, 2 * i - 1] <- round(difference$observed, digits = 2)
      simple[3, 2 * i] <- round(difference$p_value, digits = 3)
    }
    colnames(simple) <- col.names
    rownames(simple) <- c(x, 'Difference')

    cat("\n", strrep('-', dash.count), "\n", sep = "")
    message("Effect size (difference):")
    print(simple, na.print = "", quote = FALSE)
    cat(strrep('-', dash.count), "\n\n", sep = "")
  }

p_invalid_variable <-
  function (vv, df) {
    invalid <- FALSE
    for (v in vv) {
      if (is.numeric(v)) {
        invalid <- invalid || (v < 1 || v > ncol(df))
      }
      else {
        invalid <- invalid || !(v %in% names(df))
      }
    }
    return(invalid)
  }

p_ineffs <-
function (loop.data, slope, intercept) {
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y

  # Upper left and lower right
  if ( (flip.x == flip.y) && (is.na(slope) || slope <= 1E-3) ) {
    return( list(x=NA, y=NA, abs=NA, rel=NA) )
  }
  # Lower left and upper right
  if ( (flip.x != flip.y) && (is.na(slope) || slope >= -1E-3) ) {
    return( list(x=NA, y=NA, abs=NA, rel=NA) )
  }

  # Get the x-value of the line crossing the upper or low boundry
  y.xlim <- loop.data$scope.theo[4 - flip.y]
  x.lim <- (y.xlim - intercept) / slope

  # Get the y-value of the line crossing the left or right boundry
  x.ylim <- loop.data$scope.theo[1 + flip.x]
  y.lim <- slope * x.ylim + intercept

  return ( p_ineff(loop.data, x.lim, y.lim) )
}

p_ineffs_ce <-
function (loop.data, peers) {
  # if there is only one peer, the ceiling zone is zero
  if (nrow(unique(peers)) == 1) {
    return( list(x=NA, y=NA, abs=NA, rel=NA) )
  }

  # Get the x-value of the line crossing the upper or low boundry
  x.lim <- tail(peers, n=1)[1]

  # Get the y-value of the line crossing the left or right boundry
  y.lim <- head(peers, n=1)[2]

  return ( p_ineff(loop.data, x.lim, y.lim) )
}

p_ineff <-
function (loop.data, x.lim, y.lim) {
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y

  if (flip.x) {
    x.lim <- max(loop.data$scope.theo[1], x.lim)
    x.eff <- x.lim - loop.data$scope.theo[1]
  } else {
    x.lim <- min(loop.data$scope.theo[2], x.lim)
    x.eff <- loop.data$scope.theo[2] - x.lim
  }

  if (flip.y) {
    y.lim <- min(loop.data$scope.theo[4], y.lim)
    y.eff <- loop.data$scope.theo[4] - y.lim
  } else {
    y.lim <- max(loop.data$scope.theo[3], y.lim)
    y.eff <- y.lim - loop.data$scope.theo[3]
  }

  ineffs.x    <- x.eff / (loop.data$scope.theo[2] - loop.data$scope.theo[1])
  ineffs.y    <- y.eff / (loop.data$scope.theo[4] - loop.data$scope.theo[3])
  ineffs.rel  <- ineffs.x + ineffs.y - ineffs.x * ineffs.y
  ineffs.abs  <- loop.data$scope.area * ineffs.rel

  return( list(x=ineffs.x * 100, y=unname(ineffs.y) * 100,
               abs=ineffs.abs, rel=ineffs.rel * 100) )
}

CREATE_MAZE <- paste(
  "CREATE TABLE maze",
  "AS SELECT",

  "'maze_' || ROW_NUMBER() OVER (ORDER by '') AS mazeID,",

  "rp.col AS e_col,",
  "rp.se_x AS e_x,",
  "rp.cardDirEW AS 'cardDirEW.x',",

  "rp.row AS s_row,",
  "rp.se_y AS s_y,",
  "rp.cardDirNS AS 'cardDirNS.x',",
  "rp.se_pos,",

  "r1.'col.y' AS w_col,",
  "r1.nw_x AS w_x,",
  "r1.cardDirEW AS 'cardDirEW.y',",

  "r1.'row.y' AS 'n_row',",
  "r1.nw_y AS n_y,",
  "r1.cardDirNS AS 'cardDirNS.y',",
  "r1.nw_pos,",

  # TODO Not needed?
  "(CASE WHEN 'enlarged' IN (rp.strainx, rp.strainy, r1.strainx, r1.strainy)
    THEN TRUE ELSE FALSE END) AS onEdge,",
  "(CASE WHEN (nw_pos = 'nw' AND se_pos = 'se') THEN 'bisect'
         WHEN (nw_pos = 'on' OR nw_pos = 'se') THEN 'se'
         ELSE 'nw' END) AS stance,",
  "(CASE WHEN (nw_pos = 'nw' AND se_pos = 'se')
         THEN (rp.se_x - r1.nw_x) * (r1.nw_y - rp.se_y) / 2
         ELSE (rp.se_x - r1.nw_x) * (r1.nw_y - rp.se_y)
         END) AS area,",

  "(rp.se_x - :br1) * (:br4 - rp.se_y) AS se_NW_den,",
  "(r1.nw_x - :br1) * (:br4 - r1.nw_y) AS nw_NW_den,",
  "(r1.nw_x - :br1) * (:br4 - rp.se_y) AS sw_NW_den,",
  "(rp.se_x - :br1) * (:br4 - r1.nw_y) AS ne_NW_den,",

  "(:br2 - rp.se_x) * (rp.se_y - :br3) AS se_SE_den,",
  "(:br2 - r1.nw_x) * (r1.nw_y - :br3) AS nw_SE_den,",
  "(:br2 - r1.nw_x) * (rp.se_y - :br3) AS sw_SE_den,",
  "(:br2 - rp.se_x) * (r1.nw_y - :br3) AS ne_SE_den,",

  "NULL AS se_NW_num,",
  "NULL AS nw_NW_num,",
  "NULL AS sw_NW_num,",
  "NULL AS ne_NW_num,",

  "NULL AS se_SE_num,",
  "NULL AS nw_SE_num,",
  "NULL AS sw_SE_num,",
  "NULL AS ne_SE_num,",

  "(CASE WHEN (r1.nw_x = :br1 AND r1.nw_y = :br4) THEN 'north-west'
         WHEN r1.nw_x = :br1 THEN 'west'
         WHEN r1.nw_y = :br4 THEN 'north'
         ELSE 'other' END) AS nw_loc,",
  "(CASE WHEN (rp.se_x = :br2 AND rp.se_y = :br3) THEN 'south-east'
         WHEN rp.se_x = :br2 THEN 'east'
         WHEN rp.se_y = :br3 THEN 'south'
         ELSE 'other' END) AS se_loc,",

  "colrangebyrow.'x.high' AS vx,",
  "colrangebyrow.'col.high' AS vc,",
  "rowrangebycol.'y.high' AS hy,",
  "rowrangebycol.'row.high' AS hr,",

  "NULL AS vslope,",
  "NULL AS hslope,",

  "NULL AS sw_insol,",
  "NULL AS ne_insol,",
  "NULL AS nw_insol,",
  "NULL AS se_insol,",

  "NULL AS sw_inver,",
  "NULL AS ne_inver,",
  "NULL AS nw_inver,",
  "NULL AS se_inver,",

  # TODO Not needed?
  "NULL AS se_inacc,",
  "NULL AS sw_inacc,",
  "NULL AS nw_inacc,",
  "NULL AS ne_inacc,",

  "NULL AS nesw_sqr",

  "\n",
  "FROM rp",
  "JOIN r1 ON rp.col = r1.col AND rp.row = r1.row",
  "JOIN colrangebyrow ON rp.row = colrangebyrow.row",
  "JOIN rowrangebycol ON r1.'col.y' = rowrangebycol.col",

  "WHERE area > 0",
  "ORDER BY e_col, e_x, s_y;")

# TODO Merge updates for efficiency?
UPDATE_MAZE_1 <- "UPDATE maze SET
    se_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.e_col AND m2.s_row >= maze.s_row
                       AND m2.stance IN ('bisect', 'se')),
    nw_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.w_col AND m2.s_row >= maze.n_row
                       AND m2.stance IN ('bisect', 'se')),
    sw_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.w_col AND m2.s_row >= maze.s_row
                       AND m2.stance IN ('bisect', 'se')),
    ne_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.e_col AND m2.s_row >= maze.n_row
                       AND m2.stance IN ('bisect', 'se'));"

UPDATE_MAZE_2 <- "UPDATE maze SET
    se_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.e_col AND m2.n_row <= maze.s_row
                       AND m2.stance IN ('bisect', 'nw')),
    nw_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.w_col AND m2.n_row <= maze.n_row
                       AND m2.stance IN ('bisect', 'nw')),
    sw_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.w_col AND m2.n_row <= maze.s_row
                       AND m2.stance IN ('bisect', 'nw')),
    ne_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.e_col AND m2.n_row <= maze.n_row
                       AND m2.stance IN ('bisect', 'nw'));"

UPDATE_MAZE_3 <- "UPDATE maze SET
    vslope = (SELECT (CASE WHEN lm.stance = 'bisect'
                           THEN (lm.n_y - lm.s_y) / (lm.e_x - lm.w_x)
                           ELSE 1e999 END)
              FROM maze lm
              WHERE maze.vc = lm.w_col AND maze.s_row = lm.s_row),
    hslope = (SELECT (CASE WHEN lm.stance = 'bisect'
                           THEN (lm.n_y - lm.s_y) / (lm.e_x - lm.w_x)
                           ELSE 0 END)
              FROM maze lm
              WHERE maze.hr = lm.s_row AND maze.w_col = lm.w_col);"

UPDATE_MAZE_4 <- "UPDATE maze SET
    sw_insol = (CASE WHEN stance IN ('bisect', 'nw') THEN 0
                     WHEN nw_loc = 'west' THEN (hy - s_y) / (:br4 - s_y)
                     ELSE sw_NW_num / sw_NW_den END),
    ne_insol = (CASE WHEN stance IN ('bisect', 'nw') THEN 0
                     WHEN nw_loc = 'north' THEN (e_x - w_x) / (e_x - :br1)
                     ELSE ne_NW_num / ne_NW_den END),
    nw_insol = (CASE WHEN (stance IN ('bisect', 'nw') OR nw_NW_den = 0) THEN 0
                     ELSE nw_NW_num / nw_NW_den END),
    se_insol = se_NW_num / se_NW_den,

    sw_inver = (CASE WHEN stance IN ('bisect', 'se') THEN 0
                     WHEN se_loc = 'south' THEN (e_x - w_x) / (:br2 - w_x)
                     ELSE sw_SE_num / sw_SE_den END),
    ne_inver = (CASE WHEN stance IN ('bisect', 'se') THEN 0
                     WHEN se_loc = 'east' THEN (n_y - s_y) / (n_y - :br3)
                     ELSE ne_SE_num / ne_SE_den END),
    nw_inver = nw_SE_num / nw_SE_den,
    se_inver = (CASE WHEN (stance IN ('bisect', 'se') OR se_SE_den = 0) THEN 0
                     ELSE se_SE_num / se_SE_den END),

    nesw_sqr = (CASE WHEN (w_x = :br1 AND s_y = :br3 AND stance != 'bisect') THEN 'sw_sqr'
                     WHEN (e_x = :br2 AND n_y = :br4 AND stance != 'bisect') THEN 'ne_sqr'
                     ELSE 'other' END);"

UPDATE_MAZE_5 <- "UPDATE maze SET
    se_inacc  = MAX(se_insol, se_inver),
    sw_inacc  = MAX(sw_insol, sw_inver),
    nw_inacc  = MAX(nw_insol, nw_inver),
    ne_inacc  = MAX(ne_insol, ne_inver);"

UPDATE_RP <- "UPDATE rp SET
      insol = (CASE WHEN (cardDirEW != 'east' AND cardDirNS != 'north')
                    THEN (SELECT sw_insol FROM maze WHERE maze.w_x = se_x AND maze.s_y = se_y)
                    WHEN (cardDirEW != 'east' AND cardDirNS = 'north')
                    THEN (SELECT nw_insol FROM maze WHERE maze.w_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS != 'south')
                    THEN (SELECT ne_insol FROM maze WHERE maze.e_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS = 'south')
                    THEN (SELECT se_insol FROM maze WHERE maze.e_x = se_x AND maze.s_y = se_y)
                    END),
      inver = (CASE WHEN (cardDirEW != 'east' AND cardDirNS != 'north')
                    THEN (SELECT sw_inver FROM maze WHERE maze.w_x = se_x AND maze.s_y = se_y)
                    WHEN (cardDirEW != 'east' AND cardDirNS = 'north')
                    THEN (SELECT nw_inver FROM maze WHERE maze.w_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS != 'south')
                    THEN (SELECT ne_inver FROM maze WHERE maze.e_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS = 'south')
                    THEN (SELECT se_inver FROM maze WHERE maze.e_x = se_x AND maze.s_y = se_y)
                    END);"

##############################################################################
# SV 4 POINTS
##############################################################################
CREATE_STANCE <- "
    CREATE TABLE stance AS
      SELECT xyP.x, xyP.y, GROUP_CONCAT(mz.stance) as stance
			FROM mz, xyP
			WHERE xyP.x >= mz.w_x AND xyP.x <= mz.e_x AND xyP.y >= mz.s_y AND xyP.y <= mz.n_y
			GROUP BY xyP.x, xyP.y;"

CREATE_E1 <- "
    CREATE TABLE e1 AS
      SELECT *
      FROM (
        SELECT xyP.x, xyP.y, mz.*, ROW_NUMBER() OVER (PARTITION BY xyP.x, xyP.y ORDER BY mz.e_x, mz.s_y) rn
        FROM mz, xyP
        WHERE xyP.x >= mz.w_x AND xyP.x <= mz.e_x AND xyP.y >= mz.s_y AND xyP.y <= mz.n_y
      )
      WHERE rn == 1;"

UPDATE_XYP_1 <- "
    UPDATE xyP SET
    se = (SELECT stance.stance FROM stance WHERE stance.x = xyP.x AND stance.y = xyP.y) LIKE '%se%',
    nw = (SELECT stance.stance FROM stance WHERE stance.x = xyP.x AND stance.y = xyP.y) LIKE '%nw%';"

UPDATE_XYP_2 <- "
    UPDATE xyP SET
    feas = (CASE WHEN se = 1 THEN TRUE
                 WHEN e1.stance = 'bisect'
                 THEN (xyP.y - e1.s_y) * (e1.e_x - e1.w_x) <= (xyP.x - e1.w_x) * (e1.n_y - e1.s_y)
                 ELSE FALSE END),
    ezon = (CASE WHEN nw = 1 THEN TRUE
                 WHEN e1.stance = 'bisect'
                 THEN (xyP.y - e1.s_y) * (e1.e_x - e1.w_x) >= (xyP.x - e1.w_x) * (e1.n_y - e1.s_y)
                 ELSE FALSE END),
    ver_den = -(xyP.x - e1.w_x) * (xyP.y - e1.s_y) + (xyP.x - e1.w_x) * (:br3 - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - :br2) + e1.sw_SE_den,
    ver_num = power((xyP.x - e1.w_x), 2) * e1.hslope / 2 - (xyP.x - e1.w_x) * (xyP.y - e1.s_y) +
              power((xyP.y - e1.s_y), 2) / (2 * e1.vslope) + (xyP.x - e1.w_x) * (e1.hy - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - e1.vx) + e1.sw_SE_num,
    sol_den = -(xyP.x - e1.w_x) * (xyP.y - e1.s_y) + (xyP.x - e1.w_x) * (:br4 - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - :br1) + e1.sw_NW_den,
    sol_num = power((xyP.x - e1.w_x), 2) * e1.hslope / 2 - (xyP.x - e1.w_x) * (xyP.y - e1.s_y) +
              power((xyP.y - e1.s_y), 2) / (2 * e1.vslope) + (xyP.x - e1.w_x) * (e1.hy - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - e1.vx) + e1.sw_NW_num
    FROM e1
    WHERE e1.x = xyP.x AND e1.y = xyP.y;"

UPDATE_XYP_3 <- "
    UPDATE xyP SET
    ver_devi = (CASE WHEN xyP.x = :br2 THEN (xyP.y - e1.hy) / (xyP.y - :br3)
                     WHEN xyP.y = :br3 THEN (e1.vx - xyP.x) / (:br2 - xyP.x)
                     WHEN (xyP.x != :br2) AND (xyP.y != :br3)
                     THEN (CASE WHEN (e1.nesw_sqr = 'ne_sqr') AND (ver_den = 0)
                                THEN (xyP.y - e1.w_x) / (xyP.y - :br3)
                                WHEN (e1.nesw_sqr = 'sw_sqr') AND (ver_den = 0)
                                THEN (e1.e_x - xyP.x) / (:br2 - xyP.x)
                                ELSE ver_num / ver_den END)
                     ELSE 0 END),

    sol_devi = (CASE WHEN xyP.x = :br1 THEN(xyP.y - e1.hy) / (xyP.y - :br4)
                     WHEN xyP.y = :br4 THEN(xyP.x - e1.vx) / (:br1 - xyP.x)
                     WHEN (xyP.x != :br1) AND (xyP.y != :br4)
                     THEN (CASE WHEN (e1.nesw_sqr = 'ne_sqr') AND (sol_den = 0)
                                THEN (xyP.x - e1.w_x) / (xyP.x - :br1)
                                WHEN (e1.nesw_sqr = 'sw_sqr') AND (sol_den = 0)
                                THEN (e1.n_y - xyP.y) / (:br4 - xyP.y)
                                ELSE sol_num / sol_den END)
                     ELSE 0 END)

    FROM e1
    WHERE e1.x = xyP.x AND e1.y = xyP.y;"

UPDATE_XYP_4 <- "
    UPDATE xyP SET
    inver = COALESCE(
      (SELECT rp.inver FROM rp WHERE xyP.x = rp.rx AND xyP.y = rp.ry),
      (CASE WHEN feas THEN 0 ELSE ver_devi END)
    ),
    insol = COALESCE(
      (SELECT rp.insol FROM rp WHERE xyP.x = rp.rx AND xyP.y = rp.ry),
      (CASE WHEN ezon THEN 0 ELSE sol_devi END)
    );"

CREATE_MAZE <- paste(
  "CREATE TABLE maze",
  "AS SELECT",

  "'maze_' || ROW_NUMBER() OVER (ORDER by '') AS mazeID,",

  "rp.col AS e_col,",
  "rp.se_x AS e_x,",
  "rp.cardDirEW AS 'cardDirEW.x',",

  "rp.row AS s_row,",
  "rp.se_y AS s_y,",
  "rp.cardDirNS AS 'cardDirNS.x',",
  "rp.se_pos,",

  "r1.'col.y' AS w_col,",
  "r1.nw_x AS w_x,",
  "r1.cardDirEW AS 'cardDirEW.y',",

  "r1.'row.y' AS 'n_row',",
  "r1.nw_y AS n_y,",
  "r1.cardDirNS AS 'cardDirNS.y',",
  "r1.nw_pos,",

  # TODO Not needed?
  "(CASE WHEN 'enlarged' IN (rp.strainx, rp.strainy, r1.strainx, r1.strainy)
    THEN TRUE ELSE FALSE END) AS onEdge,",
  "(CASE WHEN (nw_pos = 'nw' AND se_pos = 'se') THEN 'bisect'
         WHEN (nw_pos = 'on' OR nw_pos = 'se') THEN 'se'
         ELSE 'nw' END) AS stance,",
  "(CASE WHEN (nw_pos = 'nw' AND se_pos = 'se')
         THEN (rp.se_x - r1.nw_x) * (r1.nw_y - rp.se_y) / 2
         ELSE (rp.se_x - r1.nw_x) * (r1.nw_y - rp.se_y)
         END) AS area,",

  "(rp.se_x - :br1) * (:br4 - rp.se_y) AS se_NW_den,",
  "(r1.nw_x - :br1) * (:br4 - r1.nw_y) AS nw_NW_den,",
  "(r1.nw_x - :br1) * (:br4 - rp.se_y) AS sw_NW_den,",
  "(rp.se_x - :br1) * (:br4 - r1.nw_y) AS ne_NW_den,",

  "(:br2 - rp.se_x) * (rp.se_y - :br3) AS se_SE_den,",
  "(:br2 - r1.nw_x) * (r1.nw_y - :br3) AS nw_SE_den,",
  "(:br2 - r1.nw_x) * (rp.se_y - :br3) AS sw_SE_den,",
  "(:br2 - rp.se_x) * (r1.nw_y - :br3) AS ne_SE_den,",

  "NULL AS se_NW_num,",
  "NULL AS nw_NW_num,",
  "NULL AS sw_NW_num,",
  "NULL AS ne_NW_num,",

  "NULL AS se_SE_num,",
  "NULL AS nw_SE_num,",
  "NULL AS sw_SE_num,",
  "NULL AS ne_SE_num,",

  "(CASE WHEN (r1.nw_x = :br1 AND r1.nw_y = :br4) THEN 'north-west'
         WHEN r1.nw_x = :br1 THEN 'west'
         WHEN r1.nw_y = :br4 THEN 'north'
         ELSE 'other' END) AS nw_loc,",
  "(CASE WHEN (rp.se_x = :br2 AND rp.se_y = :br3) THEN 'south-east'
         WHEN rp.se_x = :br2 THEN 'east'
         WHEN rp.se_y = :br3 THEN 'south'
         ELSE 'other' END) AS se_loc,",

  "colrangebyrow.'x.high' AS vx,",
  "colrangebyrow.'col.high' AS vc,",
  "rowrangebycol.'y.high' AS hy,",
  "rowrangebycol.'row.high' AS hr,",

  "NULL AS vslope,",
  "NULL AS hslope,",

  "NULL AS sw_insol,",
  "NULL AS ne_insol,",
  "NULL AS nw_insol,",
  "NULL AS se_insol,",

  "NULL AS sw_inver,",
  "NULL AS ne_inver,",
  "NULL AS nw_inver,",
  "NULL AS se_inver,",

  # TODO Not needed?
  "NULL AS se_inacc,",
  "NULL AS sw_inacc,",
  "NULL AS nw_inacc,",
  "NULL AS ne_inacc,",

  "NULL AS nesw_sqr",

  "\n",
  "FROM rp",
  "JOIN r1 ON rp.col = r1.col AND rp.row = r1.row",
  "JOIN colrangebyrow ON rp.row = colrangebyrow.row",
  "JOIN rowrangebycol ON r1.'col.y' = rowrangebycol.col",

  "WHERE area > 0",
  "ORDER BY e_col, e_x, s_y;")

# TODO Merge updates for efficiency?
UPDATE_MAZE_1 <- "UPDATE maze SET
    se_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.e_col AND m2.s_row >= maze.s_row
                       AND m2.stance IN ('bisect', 'se')),
    nw_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.w_col AND m2.s_row >= maze.n_row
                       AND m2.stance IN ('bisect', 'se')),
    sw_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.w_col AND m2.s_row >= maze.s_row
                       AND m2.stance IN ('bisect', 'se')),
    ne_NW_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.e_col <= maze.e_col AND m2.s_row >= maze.n_row
                       AND m2.stance IN ('bisect', 'se'));"

UPDATE_MAZE_2 <- "UPDATE maze SET
    se_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.e_col AND m2.n_row <= maze.s_row
                       AND m2.stance IN ('bisect', 'nw')),
    nw_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.w_col AND m2.n_row <= maze.n_row
                       AND m2.stance IN ('bisect', 'nw')),
    sw_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.w_col AND m2.n_row <= maze.s_row
                       AND m2.stance IN ('bisect', 'nw')),
    ne_SE_num = (SELECT COALESCE(SUM(m2.area), 0)
                 FROM maze m2
                 WHERE m2.w_col >= maze.e_col AND m2.n_row <= maze.n_row
                       AND m2.stance IN ('bisect', 'nw'));"

UPDATE_MAZE_3 <- "UPDATE maze SET
    vslope = (SELECT (CASE WHEN lm.stance = 'bisect'
                           THEN (lm.n_y - lm.s_y) / (lm.e_x - lm.w_x)
                           ELSE 1e999 END)
              FROM maze lm
              WHERE maze.vc = lm.w_col AND maze.s_row = lm.s_row),
    hslope = (SELECT (CASE WHEN lm.stance = 'bisect'
                           THEN (lm.n_y - lm.s_y) / (lm.e_x - lm.w_x)
                           ELSE 0 END)
              FROM maze lm
              WHERE maze.hr = lm.s_row AND maze.w_col = lm.w_col);"

UPDATE_MAZE_4 <- "UPDATE maze SET
    sw_insol = (CASE WHEN stance IN ('bisect', 'nw') THEN 0
                     WHEN nw_loc = 'west' THEN (hy - s_y) / (:br4 - s_y)
                     ELSE sw_NW_num / sw_NW_den END),
    ne_insol = (CASE WHEN stance IN ('bisect', 'nw') THEN 0
                     WHEN nw_loc = 'north' THEN (e_x - w_x) / (e_x - :br1)
                     ELSE ne_NW_num / ne_NW_den END),
    nw_insol = (CASE WHEN (stance IN ('bisect', 'nw') OR nw_NW_den = 0) THEN 0
                     ELSE nw_NW_num / nw_NW_den END),
    se_insol = se_NW_num / se_NW_den,

    sw_inver = (CASE WHEN stance IN ('bisect', 'se') THEN 0
                     WHEN se_loc = 'south' THEN (e_x - w_x) / (:br2 - w_x)
                     ELSE sw_SE_num / sw_SE_den END),
    ne_inver = (CASE WHEN stance IN ('bisect', 'se') THEN 0
                     WHEN se_loc = 'east' THEN (n_y - s_y) / (n_y - :br3)
                     ELSE ne_SE_num / ne_SE_den END),
    nw_inver = nw_SE_num / nw_SE_den,
    se_inver = (CASE WHEN (stance IN ('bisect', 'se') OR se_SE_den = 0) THEN 0
                     ELSE se_SE_num / se_SE_den END),

    nesw_sqr = (CASE WHEN (w_x = :br1 AND s_y = :br3 AND stance != 'bisect') THEN 'sw_sqr'
                     WHEN (e_x = :br2 AND n_y = :br4 AND stance != 'bisect') THEN 'ne_sqr'
                     ELSE 'other' END);"

UPDATE_MAZE_5 <- "UPDATE maze SET
    se_inacc  = MAX(se_insol, se_inver),
    sw_inacc  = MAX(sw_insol, sw_inver),
    nw_inacc  = MAX(nw_insol, nw_inver),
    ne_inacc  = MAX(ne_insol, ne_inver);"

UPDATE_RP <- "UPDATE rp SET
      insol = (CASE WHEN (cardDirEW != 'east' AND cardDirNS != 'north')
                    THEN (SELECT sw_insol FROM maze WHERE maze.w_x = se_x AND maze.s_y = se_y)
                    WHEN (cardDirEW != 'east' AND cardDirNS = 'north')
                    THEN (SELECT nw_insol FROM maze WHERE maze.w_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS != 'south')
                    THEN (SELECT ne_insol FROM maze WHERE maze.e_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS = 'south')
                    THEN (SELECT se_insol FROM maze WHERE maze.e_x = se_x AND maze.s_y = se_y)
                    END),
      inver = (CASE WHEN (cardDirEW != 'east' AND cardDirNS != 'north')
                    THEN (SELECT sw_inver FROM maze WHERE maze.w_x = se_x AND maze.s_y = se_y)
                    WHEN (cardDirEW != 'east' AND cardDirNS = 'north')
                    THEN (SELECT nw_inver FROM maze WHERE maze.w_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS != 'south')
                    THEN (SELECT ne_inver FROM maze WHERE maze.e_x = se_x AND maze.n_y = se_y)
                    WHEN (cardDirEW = 'east' AND cardDirNS = 'south')
                    THEN (SELECT se_inver FROM maze WHERE maze.e_x = se_x AND maze.s_y = se_y)
                    END);"

##############################################################################
# SV 4 POINTS
##############################################################################
CREATE_STANCE <- "
    CREATE TABLE stance AS
      SELECT xyP.x, xyP.y, GROUP_CONCAT(mz.stance) as stance
			FROM mz, xyP
			WHERE xyP.x >= mz.w_x AND xyP.x <= mz.e_x AND xyP.y >= mz.s_y AND xyP.y <= mz.n_y
			GROUP BY xyP.x, xyP.y;"

CREATE_E1 <- "
    CREATE TABLE e1 AS
      SELECT *
      FROM (
        SELECT xyP.x, xyP.y, mz.*, ROW_NUMBER() OVER (PARTITION BY xyP.x, xyP.y ORDER BY mz.e_x, mz.s_y) rn
        FROM mz, xyP
        WHERE xyP.x >= mz.w_x AND xyP.x <= mz.e_x AND xyP.y >= mz.s_y AND xyP.y <= mz.n_y
      )
      WHERE rn == 1;"

UPDATE_XYP_1 <- "
    UPDATE xyP SET
    se = (SELECT concat(',', S.stance, ',') FROM stance as S WHERE S.x = xyP.x AND S.y = xyP.y) LIKE '%,se,%',
    nw = (SELECT concat(',', S.stance, ',') FROM stance as S WHERE S.x = xyP.x AND S.y = xyP.y) LIKE '%,nw,%';"

UPDATE_XYP_2 <- "
    UPDATE xyP SET
    feas = (CASE WHEN se = 1 THEN TRUE
                 WHEN e1.stance = 'bisect'
                 THEN (xyP.y - e1.s_y) * (e1.e_x - e1.w_x) <= (xyP.x - e1.w_x) * (e1.n_y - e1.s_y)
                 ELSE FALSE END),
    ezon = (CASE WHEN nw = 1 THEN TRUE
                 WHEN e1.stance = 'bisect'
                 THEN (xyP.y - e1.s_y) * (e1.e_x - e1.w_x) >= (xyP.x - e1.w_x) * (e1.n_y - e1.s_y)
                 ELSE FALSE END),
    ver_den = -(xyP.x - e1.w_x) * (xyP.y - e1.s_y) + (xyP.x - e1.w_x) * (:br3 - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - :br2) + e1.sw_SE_den,
    ver_num = power((xyP.x - e1.w_x), 2) * e1.hslope / 2 - (xyP.x - e1.w_x) * (xyP.y - e1.s_y) +
              power((xyP.y - e1.s_y), 2) / (2 * e1.vslope) + (xyP.x - e1.w_x) * (e1.hy - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - e1.vx) + e1.sw_SE_num,
    sol_den = -(xyP.x - e1.w_x) * (xyP.y - e1.s_y) + (xyP.x - e1.w_x) * (:br4 - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - :br1) + e1.sw_NW_den,
    sol_num = power((xyP.x - e1.w_x), 2) * e1.hslope / 2 - (xyP.x - e1.w_x) * (xyP.y - e1.s_y) +
              power((xyP.y - e1.s_y), 2) / (2 * e1.vslope) + (xyP.x - e1.w_x) * (e1.hy - e1.s_y)
              - (xyP.y - e1.s_y) * (e1.w_x - e1.vx) + e1.sw_NW_num
    FROM e1
    WHERE e1.x = xyP.x AND e1.y = xyP.y;"

UPDATE_XYP_3 <- "
    UPDATE xyP SET
    ver_devi = (CASE WHEN xyP.x = :br2 THEN (xyP.y - e1.hy) / (xyP.y - :br3)
                     WHEN xyP.y = :br3 THEN (e1.vx - xyP.x) / (:br2 - xyP.x)
                     WHEN (xyP.x != :br2) AND (xyP.y != :br3)
                     THEN (CASE WHEN (e1.nesw_sqr = 'ne_sqr') AND (ver_den = 0)
                                THEN (xyP.y - e1.w_x) / (xyP.y - :br3)
                                WHEN (e1.nesw_sqr = 'sw_sqr') AND (ver_den = 0)
                                THEN (e1.e_x - xyP.x) / (:br2 - xyP.x)
                                ELSE ver_num / ver_den END)
                     ELSE 0 END),

    sol_devi = (CASE WHEN xyP.x = :br1 THEN(xyP.y - e1.hy) / (xyP.y - :br4)
                     WHEN xyP.y = :br4 THEN(xyP.x - e1.vx) / (:br1 - xyP.x)
                     WHEN (xyP.x != :br1) AND (xyP.y != :br4)
                     THEN (CASE WHEN (e1.nesw_sqr = 'ne_sqr') AND (sol_den = 0)
                                THEN (xyP.x - e1.w_x) / (xyP.x - :br1)
                                WHEN (e1.nesw_sqr = 'sw_sqr') AND (sol_den = 0)
                                THEN (e1.n_y - xyP.y) / (:br4 - xyP.y)
                                ELSE sol_num / sol_den END)
                     ELSE 0 END)

    FROM e1
    WHERE e1.x = xyP.x AND e1.y = xyP.y;"

UPDATE_XYP_4 <- "
    UPDATE xyP SET
    inver = COALESCE(
      (SELECT rp.inver FROM rp WHERE xyP.x = rp.rx AND xyP.y = rp.ry),
      (CASE WHEN feas THEN 0 ELSE ver_devi END)
    ),
    insol = COALESCE(
      (SELECT rp.insol FROM rp WHERE xyP.x = rp.rx AND xyP.y = rp.ry),
      (CASE WHEN ezon THEN 0 ELSE sol_devi END)
    );"

p_nca_cr_fdh <-
  function (loop.data, bn.data) {
    x <- loop.data$x
    y <- loop.data$y

    # Find the points on the ceiling ("PEERS")
    peers <- p_peers(loop.data)

    if (nrow(unique(peers)) > 1) {
      # Perform OLS through the peers
      x <- peers[, 1]
      y <- peers[, 2]
      line <- lm(y ~ x)

      intercept <- unname(coef(line)["(Intercept)"])
      slope <- unname(coef(line)["x"])
      line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
      ceiling <- p_ceiling(loop.data, slope, intercept)
      above <- p_above(loop.data, slope, intercept)
    } else {
      line <- NULL
      line_matrix <- NULL
      ceiling <- 0
      intercept <- NA
      slope <- NA
      above <- 0
      peers <- matrix(, nrow = 0, ncol = 2)
    }

    effect <- ceiling / loop.data$scope.area
    accuracy <- p_accuracy(loop.data, above)
    accuracy_nof <- nrow(loop.data$x) - above
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1

    metrics <- p_purity_empty
    # Extra metrics for certain lines
    skip <- !is.null(loop.data[[p_skip_purity]]) || Sys.getenv(p_skip_purity) == TRUE
    skip <- skip || loop.data$flip.x || loop.data$flip.y
    if (!skip) {
      df <- data.frame(loop.data$x, loop.data$y)
      condition <- colnames(loop.data$x)
      outcome <- colnames(loop.data$y)
      colnames(df) <- c(condition, outcome)
      metrics <- p_purity_metrics(df, condition, outcome, "cr_fdh")
    }

    return(list(line = line, line_matrix = line_matrix, peers = peers,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, accuracy_nof = accuracy_nof,
                fit = fit, ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_generate_title <-
  function (x.name, y.name) {
    return(paste(x.name, y.name, sep = " - "))
  }

p_pretty_name <-
  function (uglyName) {
    return(gsub("_", "-", toupper(uglyName)))
  }

p_is_number <-
  function (number) {
    if (typeof(number) == "list") {
      un_list = unlist(number, use.names = FALSE)
      if (is.null(un_list) || is.infinite(un_list)) {
        return(FALSE)
      }
    }

    if (is.null(number) ||
      is.na(number) ||
      number == "NA" ||
      number == "NN") {
      return(FALSE)
    } else {
      return(TRUE)
    }
  }

p_pretty_number <-
  function (uglyNumber, default = "", prec = 3, useSpaces = FALSE) {
    if (!p_is_number(uglyNumber)) {
      return(default)
    }

    if (is.integer(uglyNumber) && !useSpaces) {
      return(sprintf("%d", uglyNumber))
    }

    if (prec == "auto") {
      if (uglyNumber == 0) {
        prec <- 3
      } else {
        prec <- max(0, 3 - floor(log10(abs(uglyNumber))))
      }
    }
    fmt <- sprintf("%%.%df%%s", prec)

    nSpaces <- 0
    if (useSpaces) {
      nSpaces <- ifelse(prec == 0, 4, max(0, 3 - prec))
    }

    # We hate to see -0.0
    uglyNumber <- unlist(uglyNumber, use.names = FALSE)[1]
    uglyNumber[abs(uglyNumber) < 0.1^max(1, prec)] <- 0

    return(sprintf(fmt, uglyNumber, paste(rep(" ", nSpaces), collapse = '')))
  }

p_nof <-
  function (number) {
    if (!p_is_number(number)) {
      return("")
    }
    return(sprintf("(%d)", number))
  }

p_get_precision <-
  function (x) {
    if (!is.finite(x)) {
      return(NA)
    }
    x <- as.character(x)
    if (!grepl(".", x, fixed = TRUE)) {
      return(0)
    }
    frac <- strsplit(x, ".", fixed = TRUE)[[1]][2]
    if (is.na(frac) || nchar(frac) == 0) {
      return(0)
    }
    digits <- strsplit(frac, "", fixed = TRUE)[[1]]
    max(which(digits != "0"))
  }

p_warn_percentage_max <-
  function (loop.data, bn.data) {
    if (p_bottleneck_id(bn.data$bn.y) == 2 && loop.data$scope.theo[3] < 0) {
      warning(paste0("Using bottleneck.y with Y values < 0",
                     ", results might be counterintuitive!"), call. = F)
    }
  }

p_if_min_else_max <-
  function (use.min, ..., na.rm = FALSE) {
    dots <- c(...)
    return(ifelse(use.min, min(dots, na.rm = na.rm), max(dots, na.rm = na.rm)))
  }

p_is_equal <-
  function (value_1, value_2) {
    max.diff <- min(abs(value_1), abs(value_2)) / 1E6
    diff <- abs(value_1 - value_2)
    return(diff <= max.diff)
  }

p_weights <-
  function (loop.data, peers) {
    x <- loop.data$x
    flip.x <- loop.data$flip.x

    weights <- c()
    for (i in 1:(nrow(peers) - 1)) {
      if (!flip.x) {
        count <- x < peers[i + 1, 1]
      } else {
        count <- x > peers[i + 1, 1]
      }
      weights <- c(weights, length(x[count]) - sum(weights))
    }

    # Add the last column
    weights <- c(weights, length(x) - sum(weights))

    return(weights)
  }

p_get_digits <-
  function (tmp) {
    get_max_nchar <- function (n) { nchar(sub("0+$", "", sprintf("%f", n %% 1))) }
    return(min(3, max(sapply(tmp, get_max_nchar) - 2)))
  }

p_start_cluster <-
  function (condition) {
    # Always start cluster when forced
    if (Sys.getenv(p_parallel_force) == TRUE) {
      # Unless already started
      if (Sys.getenv(p_parallel_start) != TRUE) {
        registerDoParallel(p_max_cores())
        p_setenv(p_parallel_start)
      }
    }
    else if (condition) {
      paste("Starting the analysis on", p_max_cores(), "cores")
      if (grepl("windows", tolower(.Platform$OS.type))) {
        cat("Setting up parallelization, this might take a few seconds...")
      }

      # Create cluster for parallisation
      registerDoParallel(p_max_cores())

      if (grepl("windows", tolower(.Platform$OS.type))) {
        cat("\r", strrep(" ", 61), "\r")
      }
    }
    else {
      # Do sequential, this prohibits warnings on %dopar%
      registerDoSEQ()
    }
  }

p_stop_cluster <-
  function () {
    # When forced, don't stop cluster
    if (Sys.getenv(p_parallel_force) != TRUE) {
      stopImplicitCluster()
    }
  }

p_cluster_cleanup <-
  function () {
    # When forced, don't cleanup
    if (Sys.getenv(p_parallel_force) != TRUE) {
      env <- utils::getFromNamespace(".foreachGlobals", "foreach")
      if (!identical(ls(name = env), character(0))) {
        rm(list = ls(name = env), pos = env)
      }
    }
  }

p_max_cores <-
  function () {
    env.max.cores <- as.numeric(Sys.getenv("max.cores", 1000))
    max.cores <- min(env.max.cores, detectCores())
    return(max.cores)
  }

p_dopar_condition <-
  function (condition, min.cores = 2) {
    return(condition && p_max_cores() > min.cores)
  }

p_is_single_string <- function (input) {
  return(is.character(input) & length(input) == 1)
}

p_setenv <-
  function (name, value = TRUE) {
    args <- list(value)
    names(args) <- name
    do.call(Sys.setenv, args)
  }

p_generate_title <-
  function (x.name, y.name) {
    return(paste(x.name, y.name, sep = " - "))
  }

p_pretty_name <-
  function (uglyName) {
    return(gsub("_", "-", toupper(uglyName)))
  }

p_is_number <-
  function (number) {
    if (typeof(number) == "list") {
      un_list <- unlist(number, use.names = FALSE)
      if (any(is.null(un_list)) ||
        any(is.infinite(un_list)) ||
        any(is.na(un_list))) {
        return(FALSE)
      }
      if ("NA" %in% un_list || "NN" %in% un_list) {
        return(FALSE)
      }
    }

    if (is.null(number) ||
      is.na(number) ||
      number == "NA" ||
      number == "NN") {
      return(FALSE)
    } else {
      return(TRUE)
    }
  }

p_pretty_number <-
  function (uglyNumber, default = "", prec = 3, useSpaces = FALSE) {
    if (!p_is_number(uglyNumber)) {
      return(default)
    }

    if (is.integer(uglyNumber) && !useSpaces) {
      return(sprintf("%d", uglyNumber))
    }

    if (prec == "auto") {
      if (uglyNumber == 0) {
        prec <- 3
      } else {
        prec <- max(0, 3 - floor(log10(abs(uglyNumber))))
      }
    }
    fmt <- sprintf("%%.%df%%s", prec)

    nSpaces <- 0
    if (useSpaces) {
      nSpaces <- ifelse(prec == 0, 4, max(0, 3 - prec))
    }

    # We hate to see -0.0
    uglyNumber <- unlist(uglyNumber, use.names = FALSE)[1]
    uglyNumber[abs(uglyNumber) < 0.1^max(1, prec)] <- 0

    return(sprintf(fmt, uglyNumber, paste(rep(" ", nSpaces), collapse = '')))
  }

p_nof <-
  function (number) {
    if (!p_is_number(number)) {
      return("")
    }
    return(sprintf("(%d)", number))
  }

p_get_precision <-
  function (x) {
    if (!is.finite(x)) {
      return(NA)
    }
    x <- as.character(x)
    if (!grepl(".", x, fixed = TRUE)) {
      return(0)
    }
    frac <- strsplit(x, ".", fixed = TRUE)[[1]][2]
    if (is.na(frac) || nchar(frac) == 0) {
      return(0)
    }
    digits <- strsplit(frac, "", fixed = TRUE)[[1]]
    max(which(digits != "0"))
  }

p_warn_percentage_max <-
  function (loop.data, bn.data) {
    if (p_bottleneck_id(bn.data$bn.y) == 2 && loop.data$scope.theo[3] < 0) {
      warning(paste0("Using bottleneck.y with Y values < 0",
                     ", results might be counterintuitive!"), call. = F)
    }
  }

p_if_min_else_max <-
  function (use.min, ..., na.rm = FALSE) {
    dots <- c(...)
    return(ifelse(use.min, min(dots, na.rm = na.rm), max(dots, na.rm = na.rm)))
  }

p_is_equal <-
  function (value_1, value_2) {
    max.diff <- min(abs(value_1), abs(value_2)) / 1E6
    diff <- abs(value_1 - value_2)
    return(diff <= max.diff)
  }

p_weights <-
  function (loop.data, peers) {
    x <- loop.data$x
    flip.x <- loop.data$flip.x

    weights <- c()
    for (i in 1:(nrow(peers) - 1)) {
      if (!flip.x) {
        count <- x < peers[i + 1, 1]
      } else {
        count <- x > peers[i + 1, 1]
      }
      weights <- c(weights, length(x[count]) - sum(weights))
    }

    # Add the last column
    weights <- c(weights, length(x) - sum(weights))

    return(weights)
  }

p_get_digits <-
  function (tmp) {
    get_max_nchar <- function (n) { nchar(sub("0+$", "", sprintf("%f", n %% 1))) }
    return(min(3, max(sapply(tmp, get_max_nchar) - 2)))
  }

p_start_cluster <-
  function (condition) {
    # Always start cluster when forced
    if (Sys.getenv(p_parallel_force) == TRUE) {
      # Unless already started
      if (Sys.getenv(p_parallel_start) != TRUE) {
        registerDoParallel(p_max_cores())
        p_setenv(p_parallel_start)
      }
    }
    else if (condition) {
      paste("Starting the analysis on", p_max_cores(), "cores")
      if (p_is_windows()) {
        cat("Setting up parallelization, this might take a few seconds...")
      }

      # Create cluster for parallisation
      registerDoParallel(p_max_cores())

      if (p_is_windows()) {
        cat("\r", strrep(" ", 61), "\r")
      }
    }
    else {
      # Do sequential, this prohibits warnings on %dopar%
      registerDoSEQ()
    }
  }

p_stop_cluster <-
  function () {
    # When forced, don't stop cluster
    if (Sys.getenv(p_parallel_force) != TRUE) {
      stopImplicitCluster()
    }
  }

p_cluster_cleanup <-
  function () {
    # When forced, don't cleanup
    if (Sys.getenv(p_parallel_force) != TRUE) {
      env <- utils::getFromNamespace(".foreachGlobals", "foreach")
      if (!identical(ls(name = env), character(0))) {
        rm(list = ls(name = env), pos = env)
      }
    }
  }

p_max_cores <-
  function () {
    env.max.cores <- as.numeric(Sys.getenv("max.cores", 1000))
    max.cores <- min(env.max.cores, detectCores())
    return(max.cores)
  }

p_dopar_condition <-
  function (condition, min.cores = 2) {
    return(condition && p_max_cores() > min.cores)
  }

p_is_single_string <- function (input) {
  return(is.character(input) & length(input) == 1)
}

p_setenv <-
  function (name, value = TRUE) {
    args <- list(value)
    names(args) <- name
    do.call(Sys.setenv, args)
  }

p_is_windows <-
  function () {
    return(grepl("windows", tolower(.Platform$OS.type)))
  }

p_nca_cr_vrs <-
  function (loop.data, bn.data) {
    x <- loop.data$x
    y <- loop.data$y

    # Find the points on the ceiling ("PEERS")
    peers <- p_peers(loop.data, vrs = TRUE)

    if (nrow(unique(peers)) > 1) {
      # Perform OLS through the peers
      x <- peers[, 1]
      y <- peers[, 2]
      line <- lm(y ~ x)

      intercept <- unname(coef(line)["(Intercept)"])
      slope <- unname(coef(line)["x"])
      line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
      ceiling <- p_ceiling(loop.data, slope, intercept)
      above <- p_above(loop.data, slope, intercept)
    } else {
      ceiling <- 0
      intercept <- NA
      slope <- NA
      line <- NULL
      line_matrix <- NULL
      above <- 0
      peers <- matrix(, nrow = 0, ncol = 2)
    }

    effect <- ceiling / loop.data$scope.area
    accuracy <- p_accuracy(loop.data, above)
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1
    metrics <- p_purity_empty

    return(list(line = line, line_matrix = line_matrix, peers = peers,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_purity_metrics <-
  function (df, condition, outcome, ceiling) {
    # TODO Get from caller?
    scope.emp <- c(min(df[condition]), max(df[condition]),
                   min(df[outcome]), max(df[outcome]))
    fdh_peers <- p_get_peers(df, condition, outcome)
    vrs_peers <- p_get_peers(df, condition, outcome, vrs = TRUE)

    # Add case number, remove rownames, TODO Needed?
    caseNo <- seq(1, nrow(df))
    caseDescriptor <- rownames(df)
    df <- cbind(caseNo = caseNo, caseDescriptor = caseDescriptor, df)
    rownames(df) <- caseNo

    df <- p_corners(df, condition, outcome)
    xyO <- df[, c("caseNo", condition, outcome)]
    colnames(xyO) <- c('caseNo', 'x', 'y')

    if (ceiling == "ce_fdh") {
      stpceil <- p_makeStp(fdh_peers, scope.emp)
      stpweb <- p_make2SidedWeb(stpceil, scope.emp)

      meas <- p_sv4Points(xyO, stpweb)
      g_res <- p_svGauge(meas$compacts)
      gauge <- p_ensure_pct_cols(g_res$gauge, n_points = nrow(xyO))
      spread <- p_compute_spread(g_res$svP, stpweb)

      return(list(
        noise_pct = round(gauge$medPpct, 1),
        noise_nof = gauge$medP,
        exceptions_pct = round(100 * gauge$lowP / gauge$numC, 1),
        exceptions_nof = gauge$lowP,
        support_pct = round(gauge$medSsatDpct, 1),
        support_nof = gauge$medSsatD,
        spread = round(spread, 2),
        sharpness = round(gauge$sharp1, 2)
      ))
    }

    if (ceiling == "ce_vrs") {
      vrsceil <- p_makeVrs(vrs_peers, scope.emp)
      vrsweb <- p_make2SidedWeb(vrsceil, scope.emp)

      meas <- p_sv4Points(xyO, vrsweb)
      g_res <- p_svGauge(meas$compacts)
      gauge <- p_ensure_pct_cols(g_res$gauge, n_points = nrow(xyO))
      spread <- p_compute_spread(g_res$svP, vrsweb)

      return(list(
        noise_pct = round(gauge$medPpct, 1),
        noise_nof = gauge$medP,
        exceptions_pct = round(100 * gauge$lowP / gauge$numC, 1),
        exceptions_nof = gauge$lowP,
        support_pct = round(gauge$medSsatDpct, 1),
        support_nof = gauge$medSsatD,
        spread = round(spread, 2),
        sharpness = round(gauge$sharp1, 2)
      ))
    }

    if (ceiling == "cr_fdh") {
      linceil <- p_doLinCeil(fdh_peers, scope.emp, ilk = "RG")
      linweb <- p_make2SidedWeb(linceil, scope.emp)

      meas <- p_sv4Points(xyO, linweb)
      g_res <- p_svGauge(meas$compacts)
      gauge <- p_ensure_pct_cols(g_res$gauge, n_points = nrow(xyO))
      spread <- p_compute_spread(g_res$svP, linweb)

      return(list(
        noise_pct = round(gauge$medPpct, 1),
        noise_nof = gauge$medP,
        exceptions_pct = round(100 * gauge$lowP / gauge$numC, 1),
        exceptions_nof = gauge$lowP,
        support_pct = round(gauge$medSsatDpct, 1),
        support_nof = gauge$medSsatD,
        spread = round(spread, 2),
        sharpness = round(gauge$sharp1, 2)
      ))
    }

    if (ceiling == "c_lp") {
      elpceil <- p_doLinCeil(fdh_peers, scope.emp, ilk = "LP")
      elpweb <- p_make2SidedWeb(elpceil, scope.emp)

      meas <- p_sv4Points(xyO, elpweb)
      g_res <- p_svGauge(meas$compacts)
      gauge <- p_ensure_pct_cols(g_res$gauge, n_points = nrow(xyO))
      spread <- p_compute_spread(g_res$svP, elpweb)

      return(list(
        noise_pct = round(gauge$medPpct, 1),
        noise_nof = gauge$medP,
        exceptions_pct = round(100 * gauge$lowP / gauge$numC, 1),
        exceptions_nof = gauge$lowP,
        support_pct = round(gauge$medSsatDpct, 1),
        support_nof = gauge$medSsatD,
        spread = round(spread, 2),
        sharpness = round(gauge$sharp1, 2)
      ))
    }

    return(p_purity_empty)
  }

p_purity_empty <- list(
  noise_pct = NA,
  noise_nof = NA,
  exceptions_pct = NA,
  exceptions_nof = NA,
  support_pct = NA,
  support_nof = NA,
  spread = NA,
  sharpness = NA
)

p_display_plotly <-
  function (plot, peers, labels, name = 'peer', bottleneck.list = NULL) {
    # Get the params for plotting
    params <- get_plot_params()
    line_colors <- params[[1]]
    line_types <- params[[2]]
    lineWidth <- params[[3]] / 2
    point_type <- params[[4]]
    point_color <- params[[5]]

    # Missing (or too much) labels
    if (is.null(labels) ||
      length(labels) != length(plot$x) ||
      length(unique(labels)) > 5) {
      labels <- replicate(length(plot$x), 'obs')
      color.list <- 'blue'
    } else {
      color.list <- c("blue", "green3", "cyan", "magenta", "gray")
    }

    fig <- plot_ly(colors = color.list)

    template <- "<b>%s</b><br>%s, %s"
    agg_fun <- function (x) paste(unique(x), collapse = "<br>")
    text_fun <- function (r) { sprintf(template, r[3], r[1], r[2]) }

    # Add peers as separate trace first
    if (!is.null(peers)) {
      # https://github.com/plotly/plotly.R/issues/1859
      df <- data.frame(x = peers[, 1], y = peers[, 2])
      df$text <- rownames(peers)
      df <- aggregate(. ~ x + y, df, agg_fun)
      df$text <- apply(df, 1, text_fun)

      fig <- add_trace(fig, data = df, x = ~x, y = ~y,
                       type = 'scatter', mode = 'markers',
                       marker = list(color = 'red', size = 10),
                       hovertemplate = ~text,
                       showlegend = TRUE, name = name)
    }

    # Add the scatter plot for the remaining points, use colors if wanted
    include <- !(rownames(plot$x) %in% rownames(peers))

    df <- data.frame(x = plot$x[include], y = plot$y[include])
    df$label <- labels[include]
    df$text <- names(plot$x[include,])
    tmp <- aggregate(label ~ x + y, df, unique)
    df <- aggregate(text ~ x + y, df, agg_fun)
    df$text <- apply(df, 1, text_fun)
    df <- merge(df, tmp)

    fig <- add_trace(fig, data = df, x = ~x, y = ~y,
                     type = 'scatter', mode = 'markers',
                     marker = list(symbol = point_type),
                     hovertemplate = ~text,
                     showlegend = TRUE, color = df$label)

    # Plot the lines
    for (method in plot$methods) {
      if (method == "ols") {
        next
      }

      line <- plot$line_matrices[[method]]
      line_color <- line_colors[[method]]
      line_type <- line_types[[method]]
      line_list <- list(color = line_color, width = lineWidth,
                        dash = paste0(line_type, 'px'))

      if (is_infinite_null(line)) {
        next
      }

      fig <- add_trace(fig, type = 'scatter', mode = 'lines',
                       x = c(line[, 1]), y = c(line[, 2]),
                       line = line_list, showlegend = TRUE, name = method)
    }

    # Add the bottleneck lines
    # https://plotly.com/r/reference/layout/annotations/
    fig <- p_plotly_bottlenecks(fig, bottleneck.list)

    # Add title and axis labels
    title <- list(text = paste0("NCA Plot : ", plot$title), yanchor = "top")
    xaxis <- list(title = colnames(plot$x))
    if (plot$flip.x) {
      xaxis["autorange"] <- "reversed"
    }
    yaxis <- list(title = colnames(plot$y))
    if (plot$flip.y) {
      yaxis["autorange"] <- "reversed"
    }
    fig <- layout(fig, title = title, xaxis = xaxis, yaxis = yaxis)

    p_suppress_warnings({ print(fig) })
  }

p_allColor <-
  function (x) {
    result <- sapply(x, function (X) {
      tryCatch(is.matrix(col2rgb(X)),
               error = function (e) FALSE)
    })

    return(all(result))
  }

p_suppress_warnings <-
  function (.expr) {
    txt <- "Can't display both discrete & non-discrete data on same axis"
    eval.parent(substitute(
      withCallingHandlers(.expr, warning = function (w) {
        cond <- startsWith(conditionMessage(w), txt)
        if (cond) {
          invokeRestart("muffleWarning")
        }
      })
    ))
  }

p_plotly_bottlenecks <-
  function (fig, bottleneck.list) {
    for (name in names(bottleneck.list[['hor']])) {
      coord <- unlist(bottleneck.list[['hor']][[name]])
      line <- list(width = 1, dash = "dot", color = "lightgrey")
      fig <- add_lines(fig,
                       x = c(coord[1], coord[2]),
                       y = c(coord[3], coord[4]),
                       line = line, showlegend = F)
      a <- list(x = coord[1], y = coord[3],
                xref = "x", yref = "y", ax = -20, ay = -20,
                text = name, font = list(size = 10),
                arrowsize = 0.5, arrowwidth = 2, arrowcolor = "lightgrey")
      fig <- layout(fig, annotations = a)
    }
    for (name in names(bottleneck.list[['ver']])) {
      coord <- unlist(bottleneck.list[['ver']][[name]])
      line <- list(width = 1, dash = "dot", color = "lightgrey")
      fig <- add_lines(fig,
                       x = c(coord[1], coord[2]),
                       y = c(coord[3], coord[4]),
                       line = line, showlegend = F)
      a <- list(x = coord[1], y = coord[3],
                xref = "x", yref = "y", ax = -20, ay = 20,
                text = name, font = list(size = 10),
                arrowsize = 0.5, arrowwidth = 2, arrowcolor = "lightgrey")
      fig <- layout(fig, annotations = a)
    }
    return(fig)
  }

p_nca_ols <-
  function (loop.data, ...) {
    x <- loop.data$x
    y <- loop.data$y
    ols <- lm(y ~ x)

    intercept <- unname(coef(ols)["(Intercept)"])
    slope <- unname(coef(ols)["x"])
    line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
    ceiling <- p_ceiling(loop.data, slope, intercept)
    effect <- ceiling / loop.data$scope.area
    above <- p_above(loop.data, slope, intercept)
    accuracy <- p_accuracy(loop.data, above)
    fit <- 100 * ceiling / loop.data$ce_fdh_ceiling
    ineffs <- p_ineffs(loop.data, slope, intercept)
    complexity <- 1
    metrics <- p_purity_empty

    return(list(line = ols, line_matrix = line_matrix, peers = NULL,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = NULL,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_nca_ce_fdh <-
  function (loop.data, bn.data) {
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y

    # Find the points on the ceiling ("PEERS")
    peers <- p_peers(loop.data)

    line <- p_fdh_line(loop.data$scope.theo, peers, flip.x, flip.y)
    line_matrix <- p_fdh_line(loop.data$scope.theo, peers, flip.x, flip.y)
    ceiling <- p_ce_ceiling(loop.data, peers, "fdh")
    effect <- ceiling / loop.data$scope.area
    ineffs <- p_ineffs_ce(loop.data, peers)
    bottleneck <- p_bottleneck_ce(loop.data, bn.data, peers, "fdh")
    complexity <- nrow(unique(peers))

    metrics <- p_purity_empty
    # Extra metrics for certain lines
    skip <- !is.null(loop.data[[p_skip_purity]]) || Sys.getenv(p_skip_purity) == TRUE
    skip <- skip || loop.data$flip.x || loop.data$flip.y
    if (!skip) {
      df <- data.frame(loop.data$x, loop.data$y)
      condition <- colnames(loop.data$x)
      outcome <- colnames(loop.data$y)
      colnames(df) <- c(condition, outcome)
      metrics <- p_purity_metrics(df, condition, outcome, "ce_fdh")
    }

    return(list(line = line, line_matrix = line_matrix, peers = peers,
                slope = NA, intercept = NA,
                ceiling = ceiling, effect = effect,
                above = 0, accuracy = 100, fit = 100,
                ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_fdh_line <-
  function (scope, peers, flip.x, flip.y) {
    x.points <- NULL
    y.points <- NULL

    y.old <- scope[3 + flip.y]
    for (i in seq_len(nrow(peers))) {
      x.new <- peers[i, 1]
      y.new <- peers[i, 2]
      x.points <- c(x.points, x.new, x.new)
      y.points <- c(y.points, y.old, y.new)
      y.old <- y.new
    }

    x.points <- c(x.points, scope[2 - flip.x])
    y.points <- c(y.points, y.new)
    return(array(c(x.points, y.points), dim = c(length(x.points), 2)))
  }

p_nca_cols <-
  function (loop.data, bn.data) {
    x <- loop.data$x
    y <- loop.data$y
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y

    if (!flip.x && !flip.y) {
      # Upper left corner
      cols <- lm(y + max(residuals(lm(y ~ x))) ~ x)
    } else if (flip.x && !flip.y) {
      # Upper right corner
      cols <- lm(y + max(residuals(lm(y ~ x))) ~ x)
    } else if (flip.x && flip.y) {
      # Lower right corner
      cols <- lm(y + min(residuals(lm(y ~ x))) ~ x)
    } else if (!flip.x && flip.y) {
      # Lower left corner
      cols <- lm(y + min(residuals(lm(y ~ x))) ~ x)
    }

    intercept <- unname(coef(cols)["(Intercept)"])
    slope <- unname(coef(cols)["x"])
    line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
    ceiling <- p_ceiling(loop.data, slope, intercept)
    effect <- ceiling / loop.data$scope.area
    above <- p_above(loop.data, slope, intercept)
    accuracy <- p_accuracy(loop.data, above)
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1
    metrics <- p_purity_empty

    return(list(line = cols, line_matrix = line_matrix, peers = NULL,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = bottleneck,
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

p_new_pdf <-
function (name1, name2, path=NULL, paper="a4") {
  if (!is.null(path)) {
    name1 <- paste(path, name1, sep="/")
    gsub("//", "/", name1)
  }
  file.name <- paste(name1, gsub(" ", "_", name2), format(Sys.time(), "%Y-%m-%d_%H%M%S"), "pdf", sep=".")
  file.name <- gsub("_-_", "-", file.name)

  if (paper == "A4r") {
    pdf(file.name, paper=paper, width = 0, height = 0)
  } else {
    pdf(file.name, paper=paper)
  }
}

p_new_window <-
function (title="", width=7, height=7) {
  # RStudio can't handle multiple windows (yet?)
  cmds <- commandArgs(trailingOnly=FALSE)
  cmd <- cmds[1]
  if (substr(tolower(cmd), nchar(cmd)-6, nchar(cmd)) == "rstudio" ||
      substr(tolower(cmd), nchar(cmd)-10, nchar(cmd)) == "rstudio.exe") {
    return ()
  }

  # Don't open windows when on Jupyter Notebooks
  if (any(grepl("kernel", ignore.case=TRUE, cmds))) {
    return()
  }

  # Don't make windows smaller than 7x7
  width <- max(7, width)
  height <- max(7, height)
  dev.new(title=title, width=width, height=height, rescale='fixed')
}

p_columns <-
function (loop.data, is.confidence) {
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y
  x.sorted <- loop.data$x[order(loop.data$x, decreasing=flip.x)]
  y.sorted <- loop.data$y[order(loop.data$x, decreasing=flip.x)]

  columns <- p_initial_columns (x.sorted, y.sorted, loop.data, flip.x, flip.y)
  columns <- p_merge_columns(columns, flip.x, flip.y)

  if (is.confidence) {
    columns <- p_bootstrap(y.sorted, columns, loop.data)
    columns <- p_con_ce(columns, loop.data)
  }

  return ( columns )
}

p_initial_columns <-
function (x, y, loop.data, flip.x, flip.y) {
  # Initial nbr and widths of columns
  ux <- unique(x)
  k  <- length(ux)
  colwidth <- ux[-1] - ux[-k]

  # Get the boundaries from the unique x values and the column widths
  boundries <- ux[-k] + colwidth/2
  if (!flip.x) {
    first <- min(loop.data$scope.theo[1], ux[1] - colwidth[1]/2)
    second <- max(loop.data$scope.theo[2], ux[k] + colwidth[k-1]/2)
  } else {
    first <- max(loop.data$scope.theo[2], ux[1] - colwidth[1]/2)
    second <- min(loop.data$scope.theo[1], ux[k] + colwidth[k-1]/2)
  }
  boundries <- c(first, boundries, second)

  # Row 1 : nof datapoints in the column
  # Row 2 + 3 : x boundaries of the columns
  # Row 4 : x value of the max(y) datapoint
  # Row 5 : first max(y) of the datapoints, changes into bootstrap value

  columns <- matrix(0, nrow=5, ncol=k)
  for (i in 1:k) {
    # Get the boudries for the columns
    cumm1 <- sum(columns[1,]) + 1
    if (!flip.x) {
      count <- length(x[x >= boundries[i] & x < boundries[i+1] ])
    } else {
      count <- length(x[x <= boundries[i] & x > boundries[i+1] ])
    }
    columns[1:3,i] <- c(count, boundries[i], boundries[i+1])

    cumm2 <- sum(columns[1,])
    # No points in this column
    if (cumm1 > cumm2) {
      y.max <- loop.data$scope.theo[3 + flip.y]
      x.max <- loop.data$scope.theo[1 + flip.x]
    } else {
      y.max <- ifelse(flip.y, min(y[cumm1:cumm2]), max(y[cumm1:cumm2]))
      id <- intersect(cumm1:cumm2, which(y==y.max))
      id <- ifelse(!flip.x, head(id, n=1), tail(id, n=1))
      x.max <- x[id]
    }

    columns[4:5, i] <- c(x.max, y.max)
  }

  return ( columns )
}

p_merge_columns <-
function (columns, flip.x, flip.y) {
  # Min nof datapoint per column
  min.count <- round( sqrt(sum(columns[1,]) / 2) )

  while (TRUE) {
    # No more columns to merged
    if (ncol(columns) <= 1) {
      break
    }

    # If the smallest column has enough points we're done
    if (min(columns[1, ]) >= min.count) {
      break
    }

    # Search for (leftmost) largest below min.count
    max.below.min <- max(columns[1,][which(columns[1,] < min.count)])
    min.col <- min(which(columns[1,]==max.below.min))

    # For outer columns there's no choice
    if (min.col == 1 || min.col == ncol(columns)) {
      min.neighbor <- min.col + ifelse(min.col == 1, 1, -1)
    } else {
      # Find (left most) smallest above min.count OR largest below min.count
      neighbor.cols <- c(min.col - 1, min.col + 1)
      values <- columns[1, neighbor.cols]

      if ( min(values) >= min.count ) {
        # Find smallest (left most) above min.count for both above
        min.neighbor <- min.col + ifelse(values[1] > values[2], 1, -1)
      } else {
        # Find largest (left most) for both below or mixed
        min.neighbor <- min.col + ifelse(values[2] > values[1], 1, -1)
      }
    }

    columns <- p_merge_2_columns(columns, min.col, min.neighbor, flip.x, flip.y)
  }

  return (columns)
}

p_merge_2_columns <-
function (columns, col1, col2, flip.x, flip.y) {
  # Add the counts
  columns[1, col1] <- columns[1, col1] + columns[1, col2]

  # Then move the left or right 'border'
  if (col1 > col2) {
    columns[2, col1] <- columns[2, col2]
  } else {
    columns[3, col1] <- columns[3, col2]
  }

  # Use highest (or lowest on flip.y) points
  if (!flip.y && columns[5, col1] < columns[5, col2]) {
    columns[4, col1] <- columns[4, col2]
    columns[5, col1] <- columns[5, col2]
  } else if (flip.y && columns[5, col1] > columns[5, col2]) {
    columns[4, col1] <- columns[4, col2]
    columns[5, col1] <- columns[5, col2]
  } else if (columns[5, col1] == columns[5, col2]) {
    columns[5, col1] <- p_if_min_else_max(!flip.x, columns[5, col1], columns[5, col2])
  }

  # Delete the column
  columns <- columns[,-col2]

  # Make sure we have a matrix
  if (is.null(dim(columns))) {
    columns <- matrix(columns , ncol = 1)
  }

  return ( columns )
}

p_bootstrap <-
function (y, columns, loop.data) {
  conf <- loop.data$conf
  conf.rep <- loop.data$conf.rep
  flip.y <- loop.data$flip.y

  # Normalize the y values
  y.min   <- ifelse(!flip.y, min(y), max(y))
  y.range <- ifelse(!flip.y, max(y) - min(y), min(y) - max(y))
  y <- (y - y.min) / y.range

  start <- 1
  for (col in seq_len(ncol(columns))) {
    end <- start + columns[1, col] - 1
    ci <- p_bootstrap_column(y[start:end], columns[1, col], conf, conf.rep)
    columns[5, col] <- (ci * y.range) + y.min
    start <- end + 1
  }

  return ( columns )
}

# Smooth Bootstrap (adapted from 'Estimation and Inference in Nonparametric
# Frontier Models: Recent Developments and Perspectives'
# By Leopold Simar and Paul W. Wilson, page 237)
p_bootstrap_column <-
function (z, n, conf, nrep) {
  n2        <- n * 2
  zeta.hat  <- max(z)
  zeta.star <- vector(length=nrep)

  for (b in 1:nrep) {
    # reflection method (Silverman, 1986) to restore consistency
    zr <- c(z, 2 * zeta.hat - z)
    # add a mirror-image of the data to the data to create set Z^R_n
    # draw n random integers on [1,..,n*2]
    ind <- floor( runif(n) * n2 + 1 )
    # Naive bootstrap: random sample of size n from Z^R_n
    # (independently, uniformly, and with replacement)
    zs <- zr[ind]
    # determine  bandwidth h to estimate
    # the density of the 2n elements in Z^R_n via plug-in method
    hr <- p_dpikSafe(zr)
    # (Sheather and Jones, 1991) with the reflected data
    # rescale bandwidth by multiplying by 2^(1/5)
    # since the optimal bandwidth is of order O(n^(???1/5))
    h <- (2 ^ 0.2) * hr
    # instead of current O((2n)^(???1/5));
    # we only have n real obeservations instead of 2n.
    t1 <- zs + h * rnorm(n)
    zss <- ifelse(t1 <= zeta.hat, t1, 2 * zeta.hat - t1)
    # sample mean of naive bootstrap sample
    t2 <- mean(zs)
    # It can be shown that the Zsss behave as draws from \hat{f}_h(t) \
    # with E(zsss_i | Z_n) = the real sample mean.
    zsss <- t2 + (zss - t2) / sqrt( 1+(h^2) / var(zss) )
    # maxima of Zsss sample
    zeta.star[b] <- max(zsss)
  }

  # sample variance is zeta.hat/n
  g.star <- (n/zeta.hat) * (zeta.hat-zeta.star)
  qq <- quantile(g.star, probs=conf)
  ci <- zeta.hat / (1 - qq/n)

  return ( ci )
}

# Function to handle cases where dpik() is unable to estimate
# a bandwidth, usually because a data vector has an interquartile range of 0.
p_dpikSafe <- function(x)
{
  result <- try(dpik(x), silent = TRUE)

  if (typeof(result) == "double") {
    return(result)
  }

  msg <- geterrmessage()
  if (grepl("scale estimate is zero for input data", msg)) {
    return(dpik(x, scalest = "stdev"))
  } else {
    stop(msg)
  }
}

p_con_ce <-
function (columns, loop.data) {
  flip.x <- loop.data$flip.x
  flip.y <- loop.data$flip.y
  scope.theo <- loop.data$scope.theo
  peers <- loop.data$ce_fdh_peers

  # Limit the confidence by the theoretical scope
  columns[5,] <-ifelse(columns[5,] > scope.theo[4], scope.theo[4], columns[5,])
  columns[5,] <-ifelse(columns[5,] < scope.theo[3], scope.theo[3], columns[5,])

  # Limit the confidence by the CE-FDH
  for (col in seq_len(ncol(columns))) {
    if (flip.x) {
      peer.max = max(which(peers[,1] >= columns[3, col]))
    } else {
      peer.max = max(which(peers[,1] <= columns[3, col]))
    }

    if (flip.y) {
      columns[5, col] <- min(columns[5, col], peers[peer.max, 2])
    } else {
      columns[5, col] <- max(columns[5, col], peers[peer.max, 2])
    }
  }

  return ( columns )
}

p_con_ce_org <-
function (columns, flip.x, flip.y) {
  i <- 1
  while (i < ncol(columns)) {
    # Merge if left hand column is bigger than right hand column# Or
    if (columns[5, i + flip.y] >= columns[5, i + !flip.y]) {
      columns <- p_merge_2_columns(columns, i, i + 1, flip.x, flip.y)
    } else {
      i <- i + 1
    }

    if (is.vector(columns)) {
      columns <- matrix(columns, ncol=1)
      break
    }
  }

  return ( columns )
}

p_conf_line <-
function (columns) {
  x.points <- c()
  y.points <- c()
  for (col in seq_len(ncol(columns))) {
    x.points <- c(x.points, columns[2, col], columns[3, col])
    y.points <- c(y.points, columns[5, col], columns[5, col])
  }
  return (list(x.points, y.points))
}

p_peers <-
  function (loop.data, vrs = FALSE) {
    x <- loop.data$x
    y <- loop.data$y
    flip.x <- loop.data$flip.x
    flip.y <- loop.data$flip.y

    if (length(x) < 2) {
      return(NULL)
    }

    the.order <- order(if (flip.x) -x else x, if (flip.y) -y else y)
    x.sorted <- x[the.order]
    y.sorted <- y[the.order]
    rownames.org <- rownames(x)[the.order]
    peers <- matrix(c(x.sorted[1], y.sorted[1]), ncol = 2)
    rownames <- rownames.org[1]

    for (i in 2:length(x.sorted)) {
      x.curr <- x.sorted[i]
      y.curr <- y.sorted[i]
      x.prev <- tail(peers, n = 1)[1]
      y.prev <- tail(peers, n = 1)[2]
      x.equal <- p_is_equal(x.prev, x.curr)
      y.equal <- p_is_equal(y.prev, y.curr)

      # Choose this peer if the Y value is larger than previous (dep. on flipping)
      next.peer <- ifelse(flip.y, y.curr < y.prev, y.curr > y.prev)
      # Also choose this peer if it is a duplicate
      if (next.peer || (x.equal && y.equal)) {
        # If X of this peer is equal to X of last peer(s), replace last peer(s)
        while (x.equal && !y.equal) {
          peers <- head(peers, -1)
          rownames <- head(rownames, -1)
          if (!length(peers)) {
            break
          }
          x.prev <- tail(peers, n = 1)[1]
          y.prev <- tail(peers, n = 1)[2]
          x.equal <- p_is_equal(x.prev, x.curr)
          y.equal <- p_is_equal(y.prev, y.curr)
        }

        # Remove 'invalid' if VRS
        if (vrs) {
          while (p_invalid_peers(peers, x.curr, y.curr, flip.x, flip.y)) {
            peers <- head(peers, -1)
            rownames <- head(rownames, -1)
          }
        }

        peers <- rbind(peers, c(x.curr, y.curr))
        rownames <- c(rownames, rownames.org[i])
      }
    }

    rownames(peers) <- rownames
    colnames(peers) <- c("X", "Y")

    return(peers)
  }

p_invalid_peers <-
  function (peers, x3, y3, flip.x, flip.y) {
    if (nrow(peers) < 2) {
      return(FALSE)
    }

    # Take x and y from last 2 rows
    tmp <- tail(peers, 2)
    x1 <- tmp[1, 1]; y1 <- tmp[1, 2]
    x2 <- tmp[2, 1]; y2 <- tmp[2, 2]

    # Last 2 rows equal or last row and new row equal
    if ((p_is_equal(x1, x2) && p_is_equal(y1, y2)) ||
      (p_is_equal(x2, x3) && p_is_equal(y2, y3))) {
      return(FALSE)
    }

    slope0 <- (y2 - y1) / (x2 - x1)
    slope1 <- (y3 - y2) / (x3 - x2)

    if (flip.x == flip.y) {
      return(slope0 <= slope1)
    } else {
      return(slope0 >= slope1)
    }
  }

p_aggregate_peers <-
  function (model.peers, x) {
    # Aggregate all ceiling peers for independent variable x
    peers <- NULL
    for (ceiling in names(model.peers)) {
      peers <- rbind(peers, model.peers[[ceiling]][[x]])
    }
    return(peers)
  }

p_get_line_peers <-
  function (loop.data, intercept, slope) {
    peers <- p_peers(loop.data, vrs = TRUE)
    if (is.null(intercept) || is.null(slope) || nrow(peers) <= 1) {
      return(NULL)
    }
    return(p_best_peers(peers, intercept, slope))
  }

p_best_peers <-
  function (peers, intercept, slope) {
    df <- cbind(peers, abs(intercept + slope * peers[, 1] - peers[, 2]))
    df <- df[order(df[, 3]),]
    delta <- abs(df[1, 2] / 1e6)
    peers <- subset(df[, c(1, 2)], abs(df[, 3]) < delta)
    return(peers)
  }

nca_power <-
  function (n = c(20, 50, 100), effect = 0.10, slope = 1, ceiling = "ce_fdh",
            corner = 1, p = 0.05,
            distribution.x = "uniform", distribution.y = "uniform",
            rep = 100, test.rep = 200) {

    if (any(effect <= 0) || any(effect >= 1)) {
      cat("The effect size needs to be larger than 0 and smaller than 1\n")
      return()
    }
    if (!is.null(corner)) {
      p_validate_corner(1, corner)
      if (slope > 0 && (corner %in% c(2, 3))) {
        cat(p_errors[['corner_23']], "\n")
        return(NULL)
      }
      if (slope < 0 && (corner %in% c(1, 4))) {
        cat(p_errors[['corner_14']], "\n")
        return(NULL)
      }
      if (corner != 1) {
        cat("Using corner ", corner, "\n")
      }
    }

    # Make sure we're not doing extra work
    distribution.x <- unique(distribution.x)
    distribution.y <- unique(distribution.y)
    n <- unique(n)
    effect <- unique(effect)
    slope <- unique(slope)
    ceiling <- unique(ceiling)

    # Calculate the total number of iterations
    n_batch <- (
      length(distribution.x) *
        length(distribution.y) *
        length(n) *
        length(effect) *
        length(slope) *
        length(ceiling))
    n_iterations <- rep * n_batch

    # Define the variables that will store results
    results <- data.frame(n = numeric(), ES = numeric(), slope = numeric(),
                          ceiling = character(),
                          distribution.x = character(), distribution.y = character(),
                          power = numeric())

    # Never do purity
    p_setenv(p_skip_purity)

    # Start a cluster if needed
    p_start_cluster(p_dopar_condition(n_iterations > 250))

    batch <- 0
    offset <- sample.int(2^24, 1)
    for (distr.x in distribution.x) {
      for (distr.y in distribution.y) {
        for (ceil in ceiling) {
          for (sample_size in n) {
            for (effect.loop in effect) {
              for (slope.loop in slope) {
                intercept <- p_intercept(slope.loop, corner, effect.loop)

                cat("\r", strrep(" ", 100),
                    "\rDoing batch", batch + 1, "of", paste(n_batch, ""))

                r <- NULL
                ids <- seq(1, rep, max(1, round(rep / 30)))
                sig_results <- foreach(r = 1:rep, .combine = c) %dopar% {
                  set.seed(batch * rep + r + offset)
                  if (r %in% ids) {
                    cat(".")
                  }

                  df <- nca_random(sample_size, intercept, slope.loop, corner = corner,
                                   distribution.x = distr.x, distribution.y = distr.y)
                  capture.output(
                    model <- nca_analysis(df, 1, 2, ceilings = ceil, corner = corner,
                                          test.rep = test.rep, scope = c(0, 1, 0, 1))
                  )

                  # Estimated p-value
                  pval <- nca_extract(model, 'X', ceil, 'p-value')
                  # Power
                  return(as.numeric(pval <= p))
                }

                # Store the results for this iteration
                df <- data.frame(n = sample_size, ES = effect.loop,
                                 slope = slope.loop, ceiling = ceil,
                                 distribution.x = distr.x, distribution.y = distr.y,
                                 power = mean(sig_results))
                results <- rbind(results, df)

                batch <- batch + 1
              }
            }
          }
        }
      }
    }
    cat("\r", strrep(" ", 120), "\r\n")

    p_stop_cluster()
    Sys.unsetenv(p_skip_purity)

    return(results)
  }

p_intercept <-
  function (slope, corner, effect) {
    if (corner == 1) {
      intercept <- p_intercept_1_4(slope, effect)
    }
    else if (corner == 2) {
      intercept <- p_intercept_2_3(slope, effect)
    }
    else if (corner == 3) {
      intercept <- p_intercept_2_3(slope, 1 - effect)
    }
    else if (corner == 4) {
      intercept <- p_intercept_1_4(slope, 1 - effect)
    }
    return(intercept)
  }

p_intercept_1_4 <-
  function (slope, effect) {
    # Assume intercept >= 0, line through right, y on x == 1 should be < 1
    intercept <- 1 - effect - slope / 2
    if (intercept >= 0 && (intercept + slope) <= 1) {
      return(intercept)
    }

    # Assume intercept >= 0, line through roof, y on x == 1 should be >= 1
    intercept <- 1 - sqrt(2 * effect * slope)
    if (intercept >= 0 && (intercept + slope) >= 1) {
      return(intercept)
    }

    # Assume intercept < 0, line through roof, y on x == 1 should be >= 1
    intercept <- 0.5 - effect * slope
    if (intercept < 0 && (intercept + slope) >= 1) {
      return(intercept)
    }

    # Assume intercept < 0, line through side, y on x == 1 should be <> 1
    y <- sqrt(2 * (1 - effect) * slope)
    intercept <- y - slope
    return(intercept)
  }

p_intercept_2_3 <-
  function (slope, effect) {
    # Assume intercept <= 1, line through right, y on x == 1 should be >= 0
    intercept <- 1 - effect - (slope / 2)
    if (intercept <= 1 && (intercept + slope) >= 0) {
      return(intercept)
    }

    # Assume intercept <= 1, line through floor, y on x == 0 should be < 0
    intercept <- sqrt(2 * (effect - 1) * slope)
    if (intercept <= 1 && (intercept + slope) <= 0) {
      return(intercept)
    }

    # Assume intercept > 1, line through side, y on x == 0 should be >= 1
    intercept <- 1 - slope - sqrt(-2 * slope * effect)
    if (intercept > 1 && (intercept + slope) >= 0) {
      return(intercept)
    }

    # # Assume intercept > 1, line through floor, y on x == 0 should be <> 1
    intercept <- 1 / 2 - slope * (1 - effect)
    return(intercept)
  }

nca_powerplot <-
  function (df_power,
            x.variable = 'n', x.name = "Sample size",
            x.min = 0, x.max = NULL,
            line.variable = 'ES', line.name = "Effect size") {
    # Reduce vectorized variabes to single values (always first)
    has.warning <- FALSE
    plot.variables <- c('power', x.variable, line.variable)
    for (name in setdiff(colnames(df_power), plot.variables)) {
      values <- unique(df_power[[name]])
      if (length(values) != 1) {
        cat("\nVariable '", name, "' is vectorized: ", toString(values))
        cat("\nUsing only '", name, "' = ", values[1])
        df_power <- df_power[df_power[[name]] == values[1],]
        has.warning <- TRUE
      }
    }
    if (has.warning) {
      cat("\n")
    }

    if (is.null(x.min)) {
      x.min <- min(df_power[[x.variable]])
    }
    if (is.null(x.max)) {
      x.max <- max(df_power[[x.variable]])
    }

    p_new_window(title = "Power plot")
    df_power$var2 <- as.character(df_power[[line.variable]])
    b <- ggplot(df_power, aes(x = .data[[x.variable]], y = .data$power, col = .data$var2))
    b <- b + xlim(x.min, x.max) + ylim(0, 1) + geom_line()
    b <- b + labs(x = x.name, y = "Power", colour = line.name)
    b <- b + theme(legend.key = element_rect(fill = "white"))
    print(b)
    return(b)
  }

test_that("test p_intercept function", {
  # Only test manually, takes too long for building
  if (exists("p_intercept")) {
    return(NULL)
  }

  source("../../R/nca_power.R")

  df <- data.frame(x = c(0, 1, 0, 1), y = c(0, 0, 1, 1))
  corner_seq <- 1:4
  effect_seq <- seq(from = 0.01, to = 1, by = 0.01)
  slope_seq <- seq(from = -2, to = 2, by = 0.01)

  for (corner in corner_seq) {
    for (effect in effect_seq) {
      for (slope in slope_seq) {
        if (corner %in% c(1, 4) && slope < 0) {
          next
        }
        else if (corner %in% c(2, 3) && slope > 0) {
          next
        }
        if (slope == 0) {
          next
        }

        intercept <- p_intercept(slope, corner, effect)
        custom <- c(intercept, slope)
        model <- nca_analysis(df, 1, 2, ceilings = NULL,
                              custom = custom, corner = corner)
        calc <- model$`Effect size`[[1]][[1]]

        expect_equal(effect, calc, tolerance = 1e-6)
      }
    }
  }
})

nca_power <-
  function (n = c(20, 50, 100), effect = 0.10, slope = 1, ceiling = "ce_fdh",
            corner = 1, p = 0.05,
            distribution.x = "uniform", distribution.y = "uniform",
            rep = 100, test.rep = 200) {

    if (any(effect <= 0) || any(effect >= 1)) {
      cat("The effect size needs to be larger than 0 and smaller than 1\n")
      return()
    }
    if (!is.null(corner)) {
      p_validate_corner(1, corner)
      if (slope > 0 && (corner %in% c(2, 3))) {
        cat(p_errors[['corner_23']], "\n")
        return(NULL)
      }
      if (slope < 0 && (corner %in% c(1, 4))) {
        cat(p_errors[['corner_14']], "\n")
        return(NULL)
      }
      if (corner != 1) {
        cat("Using corner ", corner, "\n")
      }
    }

    # Never do purity
    p_setenv(p_skip_purity)

    # Make sure we're not doing extra work
    distribution.x <- unique(distribution.x)
    distribution.y <- unique(distribution.y)
    n <- unique(n)
    effect <- unique(effect)
    slope <- unique(slope)
    ceiling <- unique(ceiling)

    # Calculate the total number of iterations
    n_batch <- (
      length(distribution.x) *
        length(distribution.y) *
        length(n) *
        length(effect) *
        length(slope) *
        length(ceiling))
    n_iterations <- rep * n_batch

    # Define the variables that will store results
    results <- data.frame(n = numeric(), ES = numeric(), slope = numeric(),
                          ceiling = character(),
                          distribution.x = character(), distribution.y = character(),
                          power = numeric())

    # Start a cluster if needed
    p_start_cluster(p_dopar_condition(n_iterations > 250))

    batch <- 0
    offset <- sample.int(2^24, 1)
    for (distr.x in distribution.x) {
      for (distr.y in distribution.y) {
        for (ceil in ceiling) {
          for (sample_size in n) {
            for (effect.loop in effect) {
              for (slope.loop in slope) {
                intercept <- p_intercept(slope.loop, corner, effect.loop)

                cat("\r", strrep(" ", 100),
                    "\rDoing batch", batch + 1, "of", paste(n_batch, ""))

                r <- NULL
                ids <- seq(1, rep, max(1, round(rep / 30)))
                sig_results <- foreach(r = 1:rep, .combine = c) %dopar% {
                  set.seed(batch * rep + r + offset)
                  if (r %in% ids) {
                    cat(".")
                  }

                  df <- nca_random(sample_size, intercept, slope.loop, corner = corner,
                                   distribution.x = distr.x, distribution.y = distr.y)
                  capture.output(
                    model <- nca_analysis(df, 1, 2, ceilings = ceil, corner = corner,
                                          test.rep = test.rep, scope = c(0, 1, 0, 1))
                  )

                  # Estimated p-value
                  pval <- nca_extract(model, 'X', ceil, 'p-value')
                  # Power
                  return(as.numeric(pval <= p))
                }

                # Store the results for this iteration
                df <- data.frame(n = sample_size, ES = effect.loop,
                                 slope = slope.loop, ceiling = ceil,
                                 distribution.x = distr.x, distribution.y = distr.y,
                                 power = mean(sig_results))
                results <- rbind(results, df)

                batch <- batch + 1
              }
            }
          }
        }
      }
    }
    cat("\r", strrep(" ", 120), "\r\n")

    p_stop_cluster()
    Sys.unsetenv(p_skip_purity)

    return(results)
  }

p_intercept <-
  function (slope, corner, effect) {
    if (corner == 1) {
      intercept <- p_intercept_1_4(slope, effect)
    }
    else if (corner == 2) {
      intercept <- p_intercept_2_3(slope, effect)
    }
    else if (corner == 3) {
      intercept <- p_intercept_2_3(slope, 1 - effect)
    }
    else if (corner == 4) {
      intercept <- p_intercept_1_4(slope, 1 - effect)
    }
    return(intercept)
  }

p_intercept_1_4 <-
  function (slope, effect) {
    # Assume intercept >= 0, line through right, y on x == 1 should be < 1
    intercept <- 1 - effect - slope / 2
    if (intercept >= 0 && (intercept + slope) <= 1) {
      return(intercept)
    }

    # Assume intercept >= 0, line through roof, y on x == 1 should be >= 1
    intercept <- 1 - sqrt(2 * effect * slope)
    if (intercept >= 0 && (intercept + slope) >= 1) {
      return(intercept)
    }

    # Assume intercept < 0, line through roof, y on x == 1 should be >= 1
    intercept <- 0.5 - effect * slope
    if (intercept < 0 && (intercept + slope) >= 1) {
      return(intercept)
    }

    # Assume intercept < 0, line through side, y on x == 1 should be <> 1
    y <- sqrt(2 * (1 - effect) * slope)
    intercept <- y - slope
    return(intercept)
  }

p_intercept_2_3 <-
  function (slope, effect) {
    # Assume intercept <= 1, line through right, y on x == 1 should be >= 0
    intercept <- 1 - effect - (slope / 2)
    if (intercept <= 1 && (intercept + slope) >= 0) {
      return(intercept)
    }

    # Assume intercept <= 1, line through floor, y on x == 0 should be < 0
    intercept <- sqrt(2 * (effect - 1) * slope)
    if (intercept <= 1 && (intercept + slope) <= 0) {
      return(intercept)
    }

    # Assume intercept > 1, line through side, y on x == 0 should be >= 1
    intercept <- 1 - slope - sqrt(-2 * slope * effect)
    if (intercept > 1 && (intercept + slope) >= 0) {
      return(intercept)
    }

    # # Assume intercept > 1, line through floor, y on x == 0 should be <> 1
    intercept <- 1 / 2 - slope * (1 - effect)
    return(intercept)
  }

nca_powerplot <-
  function (df_power,
            x.variable = 'n', x.name = "Sample size",
            x.min = 0, x.max = NULL,
            line.variable = 'ES', line.name = "Effect size") {
    # Reduce vectorized variabes to single values (always first)
    has.warning <- FALSE
    plot.variables <- c('power', x.variable, line.variable)
    for (name in setdiff(colnames(df_power), plot.variables)) {
      values <- unique(df_power[[name]])
      if (length(values) != 1) {
        cat("\nVariable '", name, "' is vectorized: ", toString(values))
        cat("\nUsing only '", name, "' = ", values[1])
        df_power <- df_power[df_power[[name]] == values[1],]
        has.warning <- TRUE
      }
    }
    if (has.warning) {
      cat("\n")
    }

    if (is.null(x.min)) {
      x.min <- min(df_power[[x.variable]])
    }
    if (is.null(x.max)) {
      x.max <- max(df_power[[x.variable]])
    }

    p_new_window(title = "Power plot")
    df_power$var2 <- as.character(df_power[[line.variable]])
    b <- ggplot(df_power, aes(x = .data[[x.variable]], y = .data$power, col = .data$var2))
    b <- b + xlim(x.min, x.max) + ylim(0, 1) + geom_line()
    b <- b + labs(x = x.name, y = "Power", colour = line.name)
    b <- b + theme(legend.key = element_rect(fill = "white"))
    print(b)
    return(b)
  }

p_bottleneck_data <-
function (x, y, scope, flip.y, ceilings, bottleneck.x, bottleneck.y, steps, step.size, cutoff) {
  bn.x <- p_validate_bottleneck(bottleneck.x, "x")
  bn.y <- p_validate_bottleneck(bottleneck.y, "y")
  bn.x.id <- p_bottleneck_id(bn.x)
  bn.y.id <- p_bottleneck_id(bn.y)

  bottleneck.xy <- p_mp_mpy(y, scope, steps, step.size, bn.y.id, flip.y)
  mp <- as.data.frame(bottleneck.xy[[1]])
  mpy <- bottleneck.xy[[2]]

  attr(mp, "bn.x") <- bn.x
  attr(mp, "bn.y") <- bn.y
  attr(mp, "bn.y.id") <- bn.y.id
  attr(mp, "size") <- nrow(x)
  attr(mp, "cutoff") <- cutoff

  bottlenecks <- list()
  for (ceil in setdiff(ceilings, p_no_bottleneck)) {
    bottlenecks[[ceil]] <- mp
  }

  return( list(
    bottlenecks=bottlenecks,
    bn.x=bn.x,
    bn.y=bn.y,
    bn.x.id=bn.x.id,
    bn.y.id=bn.y.id,
    mpy=mpy,
    cutoff=cutoff,
    steps=steps
  ) )
}

p_mp_mpy <-
function (y, scope, steps, step.size, bn.y.id, flip.y) {
  if (length(steps) == 1 && steps < 1) {
    steps <- 10
  }

  if (bn.y.id == 3) {
    mp_mpy <- p_mp_mpy_actual(y, scope, steps, step.size, flip.y)
  }
  else {
    mp_mpy <- p_mp_mpy_perc(y, scope, steps, step.size, bn.y.id, flip.y)
  }

  colnames(mp_mpy[[1]]) <- colnames(y)
  colnames(mp_mpy[[2]]) <- colnames(y)

  return( mp_mpy )
}

p_low_high <-
function (y, scope, bn.y.id) {
  # Try to get the low/high values from the scope
  py.low <- scope[[1]][3]
  py.high <- scope[[1]][4]

  # Scope might be NULL
  if (is.null(py.low) || is.null(py.high)) {
    py.low <- min(y, na.rm=TRUE)
    py.high <- max(y, na.rm=TRUE)
  } else {
    py.low <- min(y, py.low, na.rm=TRUE)
    py.high <- max(y, py.high, na.rm=TRUE)
  }

  # User want from zero
  if (bn.y.id == 2) {
    py.low <- 0
  }

  return ( c(py.low, py.high) )
}

p_sanitize_steps <-
function (steps, low, high) {
  steps <- sort(steps)

  if (low > steps[1]) {
    warning("Bottleneck: Some steps below scope, excluded", call.=FALSE)
    steps <- steps[steps >= low]
  }

  if (high < steps[length(steps)]) {
    warning("Bottleneck: Some steps above scope, excluded", call.=FALSE)
    steps <- steps[steps <= high]
  }

  return (steps)
}

p_mp_mpy_actual <-
function (y, scope, steps, step.size, flip.y) {
  py.low.high <- p_low_high(y, scope, 3)
  py.low <- py.low.high[1]
  py.high <- py.low.high[2]

  if (is.null(step.size) || step.size <= 0) {
    if (length(steps) > 1) {
      # Interpret the list of steps as the values
      values <- p_sanitize_steps(steps, py.low, py.high)
    }
    else {
      # Single step
      step <- (py.high - py.low) / steps
      values <- seq(py.low, py.high, by=step)
    }
  } else {
    values <- NULL
    value <- py.low
    while (value <= py.high) {
      values <- c(values, value)
      value <- value + step.size
    }
    if (abs(values[length(values)] - py.high) > 1E-6) {
      values <- c(values, py.high)
    }
  }

  if (flip.y) {
    mpy <- matrix(rev(values), ncol=1)
  } else {
    mpy <- matrix(values, ncol=1)
  }

  return (list(mpy, mpy))
}

p_mp_mpy_perc <-
function (y, scope, steps, step.size, bn.y.id, flip.y) {
  py.low.high <- p_low_high(y, scope, bn.y.id)
  py.low <- py.low.high[1]
  py.high <- py.low.high[2]

  if (is.null(step.size) || step.size <= 0) {
    if (length(steps) > 1) {
      # Interpret the list of steps as the probabilities (0 - 100)
      probs <- p_sanitize_steps(steps, 0, 100) / 100
    }
    else {
      # Single step
      probs <- seq(0, 1, length.out=steps + 1)
    }
  } else {
    probs <- seq(0, 1, by=min(1000, step.size / 100))
  }

  if (bn.y.id == 4) {
    values <- quantile(y[[1]], probs, na.rm=TRUE)
  }
  else {
    values <- py.low + probs * (py.high - py.low)
  }

  if (flip.y) {
    mpy <- matrix(rev(values), ncol=1)
  } else {
    mpy <- matrix(values, ncol=1)
  }
  mp <- matrix(100 * probs, ncol=1)

  return (list(mp, mpy))
}

p_bottleneck_options <- list(
  "percentage.range",
  "percentage.max",
  "actual",
  "percentile"
)

p_validate_bottleneck <-
function (option, x_or_y) {
  if (is.na(match(option, p_bottleneck_options))) {
    warning(paste0("Bottleneck option '", option, "' for ", x_or_y,
      " is not valid, using '", p_bottleneck_options[[1]], "'"), call.=FALSE)
    return(p_bottleneck_options[[1]])
  }
  return(option)
}

p_bottleneck_id <-
function (option) {
  if (is.na(match(option, p_bottleneck_options))) {
    return(1)
  }
  return(match(option, p_bottleneck_options))
}

p_validate_clean <-
  function (data, x, y, outliers = FALSE) {
    data <- as.data.frame(data)

    if (length(y) != 1) {
      message()
      stop("Dependent variable can only be a single column!\n\n", call. = F)
    }

    # Replace all non-numeric values with NA (will be discarded later on)
    for (i in seq_len(ncol(data))) {
      suppressWarnings(data[[i]] <- as.numeric(as.character(data[[i]])))
    }

    # Needed to collapse multiple scope warnings
    found <- FALSE
    for (i in 1:sys.nframe()) {
      haystack <- deparse(sys.calls()[[sys.nframe() - i + 1]])[1]
      found <- found || grepl("nca_outliers(", haystack, fixed = TRUE)
    }
    if (outliers == found) {
      .GlobalEnv$scope_warnings_nca <- NULL
    }

    return(list(x = data[x], y = data[y]))
  }

p_validate_ceilings <-
  function (methods) {
    ceilings <- c(ceilings, p_ceiling_custom)
    methods <- unique(methods)

    valid_methods <- intersect(tolower(methods), ceilings)
    if (length(valid_methods) == 0) {
      warning("Invalid ceilings, using ols, ce_fdh and cr_fdh", call. = FALSE)
      return(c("ols", "ce_fdh", "cr_fdh"))
    }

    if (length(valid_methods) != length(methods)) {
      for (diff in setdiff(methods, valid_methods)) {
        warning(sprintf("Ignoring invalid ceiling(s) '%s'", diff), call. = FALSE)
      }
    }
    return(valid_methods)
  }

p_validate_custom <-
  function (custom, data.x) {
    if (is.null(custom)) {
      return(NULL)
    }

    if (!(length(custom) %in% c(2, 2 * ncol(data.x)))) {
      extra <- ifelse(ncol(data.x) > 1, paste("or", 2 * ncol(data.x)), "")
      warning("Invalid custom, it should be of length 2 ", extra, call. = FALSE)
      return(NULL)
    }

    # We can not warn for 'outside scope': scope not yet defined

    if (length(custom) == 2) {
      custom <- rep(custom, ncol(data.x))
    }
    return(matrix(custom, ncol = 2, byrow = TRUE))
  }

p_validate_flipx <-
  function (x, flip.x) {
    if (length(flip.x) != length(x)) {
      # Use single value for all Xs
      if (length(flip.x) == 1) {
        # Use single value for all Xs
        flip.x <- replicate(length(x), flip.x)
      } else {
        message()
        stop("The length of 'flip.x' needs to be equal to the length of x ",
             "or a single Boolean!\n", call. = F)
      }
    }

    return(flip.x)
  }

p_validate_corner <-
  function (x, corner) {
    # Check length of (or create) the vector
    if (length(corner) != length(x)) {
      # Use single value for all Xs
      if (length(corner) == 1) {
        # Use single value for all Xs
        corner <- replicate(length(x), corner)
      } else {
        message()
        stop("The length of 'corner' needs to be equal to the length of x ",
             "or a integer 1:4!\n", call. = F)
      }
    }

    # Check if all corners are 1=4
    if (!all(corner %in% 1:4)) {
      message()
      stop("All corners must be an integer between 1 and 4 !\n", call. = F)
    }

    # Check if corners are all upper or all lower
    if (!all(corner %in% c(1, 2)) && !all(corner %in% c(3, 4))) {
      message()
      stop("All corners need to be in the upper half (1 and 2) \n       ",
           "or the lower half (3 and 4).\n       ",
           "You can not mix upper and lower !\n", call. = F)
    }

    return(corner)
  }

p_validate_effect_aggregation <-
  function (effect_aggregation) {
    effect_aggregation <- intersect(c(2, 3, 4), effect_aggregation)
    if (length(effect_aggregation) > 0) {
      total <- paste(c(1, effect_aggregation), collapse = ', ')
      warning(paste("Using corners", total, "for effect_aggregation"), call. = FALSE)
      warning(paste("The effect aggregation argument will be deprecated:",
                    "For each condition only one corner can be analysed at a time.",
                    "For the analysis of multiple corners for a condition,",
                    "conduct multiple analyses."), call. = FALSE)
    }
    return(effect_aggregation)
  }

p_nca_c_lp <-
  function (loop.data, bn.data) {
    peers <- p_peers(loop.data)

    if (nrow(unique(peers)) > 1) {
      K <- nrow(peers)
      factor <- ifelse(sum(loop.data$flip.x, loop.data$flip.y) == 1, -1, 1)
      f.obj <- c(K, -K, factor * sum(peers[, 1]))
      f.con <- cbind(rep(1, K), rep(-1, K), factor * peers[, 1])
      f.dir <- rep(ifelse(loop.data$flip.y, "<=", ">="), K)
      f.rhs <- peers[, 2]
      sol <- lp(ifelse(loop.data$flip.y, "max", "min"),
                f.obj, f.con, f.dir, f.rhs)$solution

      # Intercept can always be negative
      intercept <- sol[[1]] - sol[[2]]
      # Slope needs to be positive in corner 1 and 4
      slope <- factor * sol[[3]]
      line <- c(intercept, slope)
      line_matrix <- p_get_line_matrix(intercept, slope, loop.data)
      ceiling <- p_ceiling(loop.data, slope, intercept)
      above <- 0
      # Find the peers on the line
      peers <- p_best_peers(peers, intercept, slope)
    } else {
      line <- NULL
      line_matrix <- NULL
      slope <- NA
      intercept <- NA
      ceiling <- 0
      above <- NA
      peers <- matrix(, nrow = 0, ncol = 2)
    }

    effect <- ceiling / loop.data$scope.area
    accuracy <- p_accuracy(loop.data, above)
    fit <- p_fit(ceiling, loop.data$ce_fdh_ceiling)
    ineffs <- p_ineffs(loop.data, slope, intercept)
    bottleneck <- p_bottleneck(loop.data, bn.data, slope, intercept)
    complexity <- 1

    metrics <- p_purity_empty
    # Extra metrics for certain lines
    skip <- !is.null(loop.data[[p_skip_purity]]) || Sys.getenv(p_skip_purity) == TRUE
    skip <- skip || loop.data$flip.x || loop.data$flip.y
    if (!skip) {
      df <- data.frame(loop.data$x, loop.data$y)
      condition <- colnames(loop.data$x)
      outcome <- colnames(loop.data$y)
      colnames(df) <- c(condition, outcome)
      metrics <- p_purity_metrics(df, condition, outcome, "c_lp")
    }

    return(list(line = line, line_matrix = line_matrix, peers = peers,
                slope = slope, intercept = intercept,
                ceiling = ceiling, effect = effect,
                above = above, accuracy = accuracy, fit = fit,
                ineffs = ineffs, bottleneck = bottleneck, 
                complexity = complexity,
                noise_pct = metrics$noise_pct,
                noise_nof = metrics$noise_nof,
                exceptions_pct = metrics$exceptions_pct,
                exceptions_nof = metrics$exceptions_nof,
                support_pct = metrics$support_pct,
                support_nof = metrics$support_nof,
                spread = metrics$spread,
                sharpness = metrics$sharpness
    ))
  }

nca <-
function (data, x, y, ceilings=c("ols", "ce_fdh", "cr_fdh")) {
  model <- nca_analysis(data, x, y, ceilings=ceilings)
  attr(model, "show.plots") <- TRUE

  model["bottlenecks"] <- NULL
  model["tests"] <- NULL
  model["test.time"] <- NULL

  return( model )
}

nca_analysis <-
function (data, x, y, ceilings=c("ols", "ce_fdh", "cr_fdh"), custom=NULL,
          corner=NULL, flip.x=FALSE, flip.y=FALSE, scope=NULL,
          bottleneck.x='percentage.range', bottleneck.y='percentage.range',
          steps=10, step.size=NULL, cutoff=0, qr.tau=0.95, effect_aggregation=1,
          test.rep=0, test.p_confidence=0.95, test.p_threshold=0.05) {

  # Cleans up any cluster registration
  p_cluster_cleanup()

  # Validate and clean data
  cleaned <- p_validate_clean(data, x, y)
  data.x <- cleaned$x
  data.y <- cleaned$y

  # Validate ceiling types and custom if applicable
  custom <- p_validate_custom(custom, data.x)
  if (!is.null(custom)) {
    ceilings <- c(ceilings, p_ceiling_custom)
  }
  ceilings <- p_validate_ceilings(ceilings)

  # Overrule flip.x and flip.y if corners is defined
  if (!is.null(corner)) {
    corner <- p_validate_corner(x, corner)
    if (flip.x || flip.y) {
      warning("Ignoring 'flip.x' and 'flip.y': 'corner' is defined", call.=FALSE)
    }
    flip.y <- all(corner %in% c(3, 4))
    flip.x <- corner %in% c(2, 4)
  }
  # Always validate flip.x and flip.y
  flip.x <- p_validate_flipx(x, flip.x)
  flip.y <- isTRUE(flip.y)

  # Validate scope
  scope <- p_scope(x, scope)

  # Validate effect size aggregation
  effect_aggregation <- p_validate_effect_aggregation(effect_aggregation)

  # Data object for bottlenecks
  bn.data <- p_bottleneck_data(data.x, data.y, scope, flip.y, ceilings,
                               bottleneck.x, bottleneck.y, steps, step.size, cutoff)

  # Data for tests
  # test.rep can never be larger than n!, only test for small n's
  # fact(15) is 1e12, if user chooses that for test.rep he's got other problems
  if (nrow(data.y) < 16 && test.rep > factorial(nrow(data.x))) {
    test.rep <- factorial(nrow(data.y))
    fmt <- "\nLowered test.rep to %s as it can not be larger than N!\n\n"
    cat(sprintf(fmt, test.rep))
  }
  test.params <- list(rep=test.rep,
                      p_confidence=test.p_confidence,
                      p_threshold=test.p_threshold)

  # Create cluster for parallisation if needed
  p_start_cluster(p_dopar_condition(
    length(ceilings) * length(data.x) * test.rep > 6000, min.cores=1))

  # Create output lists
  plots <- list()
  summaries <- list()
  tests <- list()
  peers <- list()
  test.time <- 0

  # Loop the independent variables
  for (id.x in seq_along(data.x)) {
    loop.data <- p_create_loop_data(data.x, data.y, scope, flip.x, flip.y, id.x, qr.tau, custom)
    p_warn_percentage_max(loop.data, bn.data)
    x.name <- loop.data$names[id.x]

    # We need this for the 'FIT' number, regardless of user preference
    analysis_ce_fdh <- p_nca_wrapper("ce_fdh", loop.data, bn.data, effect_aggregation)
    loop.data$ce_fdh_ceiling <- analysis_ce_fdh$ceiling
    loop.data$ce_fdh_peers <- analysis_ce_fdh$peers

    analyses <- list()
    for (ceiling in ceilings) {
      if (ceiling == "ce_fdh") {
        analysis <- analysis_ce_fdh
      } else {
        analysis <- p_nca_wrapper(ceiling, loop.data, bn.data, effect_aggregation)
      }
      if (!is.null(analysis$bottleneck) && !(ceiling %in% p_no_bottleneck)) {
        bn.data$bottlenecks[[ceiling]][x.name] <- analysis$bottleneck
        attr(bn.data$bottlenecks[[ceiling]][x.name], 'cases') <- attr(analysis$bottleneck, 'cases')
      }
      analysis$bottleneck <- NULL
      analyses[[ceiling]] <- analysis
      peers[[ceiling]][[x.name]] <- analysis$peers
    }

    test_tuple <- p_test(analyses, loop.data, test.params, effect_aggregation)
    if (!is.null(test_tuple)) {
      tests[[x.name]] <- test_tuple$test
      test.time <- test.time + test_tuple$test.time
    }

    # Add P-value/accuracy for displaying in summary
    for (ceiling in ceilings) {
      if (ceiling %in% names(tests[[x.name]])) {
        analyses[[ceiling]]$p <- tests[[x.name]][[ceiling]]$p
        analyses[[ceiling]]$p_accuracy <- tests[[x.name]][[ceiling]]$test.params$p_accuracy
      }
      else {
        analyses[[ceiling]]$p <- NA
        analyses[[ceiling]]$p_accuracy <- NA
      }
    }
    plots[[x.name]] <- p_plot(analyses, loop.data, corner)
    summaries[[x.name]] <- p_summary(analyses, loop.data)
  }

  # Shut down cluster for parallisation
  p_stop_cluster()

  # Add the bottlenecks with mpy attribute
  bottlenecks <- bn.data$bottlenecks
  attr(bottlenecks, "mpy") <- bn.data$mpy
  attr(bottlenecks, "flip.y") <- flip.y

  model <- list(plots=plots,
                summaries=summaries,
                bottlenecks=bottlenecks,
                peers=peers,
                tests=tests,
                test.time=p_test_time(test.time))

  model <- p_add_extracts(model)

  class(model) <- "nca_result"
  attr(model, "show.plots") <- FALSE

  return ( model )
}

nca_util_normalize <-
  function (data, scale = NULL, min_max = NULL) {
    scales <- p_scales(scale, ncol(data))
    min_max <- p_min_max(data, min_max)

    for (col_idx in seq_along(data)) {
      if (NA %in% scales[col_idx,]) {
        next
      }
      column <- data[[col_idx]]
      min_val <- min_max[col_idx, 1]
      max_val <- min_max[col_idx, 2]
      new_min <- scales[col_idx, 1]
      new_max <- scales[col_idx, 2]

      # Normalize according to min_max
      numerator <- column - min_val
      denominator <- (max_val - min_val)
      org_norm <- numerator / denominator

      # Scale according to scale
      scale <- (new_max - new_min)
      column <- org_norm * scale + new_min

      # avoid floating-point error
      tol <- sqrt(.Machine$double.eps) * abs(new_max - new_min)
      column[abs(column - new_min) < tol] <- new_min
      column[abs(column - new_max) < tol] <- new_max
      column <- pmax(new_min, pmin(new_max, column))
      data[, col_idx] <- column
    }
    return(data)
  }

p_scales <-
  function (scales, no.cols) {
    if (is.null(scales)) {
      scales <- c(0, 1)
    }
    if (length(scales) == 2) {
      return(t(replicate(no.cols, scales)))
    }

    req_length <- 2 * no.cols
    if (length(scales) == req_length) {
      return(t(matrix(scales, nrow = 2)))
    }

    stop("The length of scale need to be 2 or ", req_length)
  }

p_min_max <-
  function (data, min_max) {
    data_min_max <- matrix(nrow = ncol(data), ncol = 2)
    for (col_idx in seq_along(data)) {
      data_min_max[col_idx,] <- c(min(data[[col_idx]], na.rm = TRUE),
                                  max(data[[col_idx]], na.rm = TRUE))
    }

    if (is.null(min_max)) {
      return(data_min_max)
    }

    no.cols <- ncol(data)
    if (length(min_max) == 2) {
      # One mix-max for all columns (not ideal)
      min_max <- rep(min_max, no.cols)
    }
    else if (length(min_max) != (2 * no.cols)) {
      stop("The length of min_max need to be 2 or ", 2 * no.cols)
    }
    min_max <- t(matrix(min_max, nrow = 2))

    for (idx in seq_along(data)) {
      extended <- NULL
      if (min_max[idx, 1] > data_min_max[idx, 1] || min_max[idx, 2] < data_min_max[idx, 2]) {
        extended <- min_max[idx,]
      }
      min_max[idx, 1] <- min(data_min_max[idx, 1], min_max[idx, 1])
      min_max[idx, 2] <- max(data_min_max[idx, 2], min_max[idx, 2])
      if (!is.null(extended)) {
        msg <- sprintf("Extending theoretical scope for '%s' from c(%s) to c(%s)",
                       colnames(data)[idx],
                       paste(extended, collapse = ', '),
                       paste(min_max[idx,], collapse = ', '))
        warning(msg, call. = FALSE)
      }
    }

    return(min_max)
  }

nca_util_normalize <-
  function (data, scale = NULL, min_max = NULL) {
    scales <- p_scales(scale, ncol(data))
    min_max <- p_min_max(data, min_max)

    for (col_idx in seq_along(data)) {
      if (NA %in% scales[col_idx,]) {
        next
      }
      column <- data[[col_idx]]
      min_val <- min_max[col_idx, 1]
      max_val <- min_max[col_idx, 2]
      new_min <- scales[col_idx, 1]
      new_max <- scales[col_idx, 2]

      # Normalize according to min_max
      numerator <- column - min_val
      denominator <- (max_val - min_val)
      org_norm <- numerator / denominator

      # Scale according to scale
      scale <- (new_max - new_min)
      column <- org_norm * scale + new_min

      # avoid floating-point error
      tol <- sqrt(.Machine$double.eps) * abs(new_max - new_min)
      column[abs(column - new_min) < tol] <- new_min
      column[abs(column - new_max) < tol] <- new_max
      column <- pmax(new_min, pmin(new_max, column))
      data[, col_idx] <- column
    }
    return(data)
  }

p_scales <-
  function (scales, no.cols) {
    if (is.null(scales)) {
      scales <- c(0, 1)
    }
    if (length(scales) == 2) {
      return(t(replicate(no.cols, scales)))
    }

    req_length <- 2 * no.cols
    if (length(scales) == req_length) {
      return(t(matrix(scales, nrow = 2)))
    }

    stop("The length of scale need to be 2 or ", req_length)
  }

p_min_max <-
  function (data, min_max) {
    data_min_max <- matrix(nrow = ncol(data), ncol = 2)
    for (col_idx in seq_along(data)) {
      data_min_max[col_idx,] <- c(min(data[[col_idx]], na.rm = TRUE),
                                  max(data[[col_idx]], na.rm = TRUE))
    }

    if (is.null(min_max)) {
      return(data_min_max)
    }

    no.cols <- ncol(data)
    if (length(min_max) == 2) {
      # One mix-max for all columns (not ideal)
      min_max <- rep(min_max, no.cols)
    }
    else if (length(min_max) != (2 * no.cols)) {
      stop("The length of min_max need to be 2 or ", 2 * no.cols)
    }
    min_max <- t(matrix(min_max, nrow = 2))

    for (idx in seq_along(data)) {
      if (min_max[idx, 1] > data_min_max[idx, 1] || min_max[idx, 2] < data_min_max[idx, 2]) {
        msg <- sprintf("Extending theoretical scope for '%s' from c(%s) to c(%s)",
                       colnames(data)[idx],
                       paste(min_max[idx,], collapse = ', '),
                       paste(data_min_max[idx,], collapse = ', '))
        warning(msg, call. = FALSE)
      }
      min_max[idx, 1] <- min(data_min_max[idx, 1], min_max[idx, 1])
      min_max[idx, 2] <- max(data_min_max[idx, 2], min_max[idx, 2])
    }

    return(min_max)
  }

\name{nca_analysis}
\alias{nca_analysis}
\title{Run NCA analyses on a data set}
\description{Run multiple types of NCA analyses on a dataset}
\usage{nca_analysis(data, x, y, ceilings=c('ols', 'ce_fdh', 'cr_fdh'), custom=NULL,
             corner=NULL, flip.x=FALSE, flip.y=FALSE, scope=NULL,
             bottleneck.x='percentage.range', bottleneck.y='percentage.range',
             steps=10, step.size=NULL, cutoff=0, qr.tau=0.95,
             effect_aggregation = 1, test.rep=0,
             test.p_confidence=0.95, test.p_threshold=0.05)}
\arguments{
\item{data}{dataframe with columns of the variables}
\item{x}{index or name (or a vector of those) with condition(s) x}
\item{y}{index or name of the column with the outcome y}
\item{ceilings}{vector with the ceiling techniques to include in this analysis}
\item{custom}{custom ceiling line defined by a intercept-slope pair per condition. If only 1 pair given it will be used for all conditions.}
\item{corner}{either an integer or a vector of integers, indicating the corner to analyze, see Details}
\item{flip.x}{reverse the direction of the conditions\cr
 Use either a boolean for all conditions, or a vector with the same length as x}
\item{flip.y}{reverse the direction of the outcomes, boolean}
\item{scope}{a theoretical scope in list format : (x.low, x.high, y.low, y.high), see Details}
\item{bottleneck.x}{options for displaying the conditions in the bottleneck table\cr
 'percentage.range' to display the percentage of range between min(x) and max(x)\cr
 'percentage.max' to display the percentage of max(x)\cr
 'actual' to display the actual values\cr
 'percentile' to display the percentiles\cr
 Using percentage.max with negative values might yield counterintuitive results.}
\item{bottleneck.y}{options for displaying the outcomes in the bottleneck table. \cr See bottleneck.x}
\item{steps}{this argument accepts 2 types : \cr - an integer (number of steps in the bottleneck table) \cr - a list of values (used as Y values in the bottleneck table). \cr Only useful if step.size is not defined.}
\item{step.size}{define the step size in the bottleneck table.\cr
  The user will be warned if the stepsize does not fit the Y range.\cr
  Defaults to null for using steps.}
\item{cutoff}{display calculated x,y values that are lower/higher than lowest/highest observed x,y values in the bottleneck table as:\cr
0 : NN (not necessary) and NA (not available)\cr 1 : NN (not necessary) and highest observed values\cr 2 : calculated values}
\item{qr.tau}{define the qr tau (between 0 and 1) for the quantile regression ceiling technique, default 0.95}
\item{effect_aggregation}{define the corners to aggregate into the effect size. 1 is upper-left and is always selected, 2 is upper-right, 3 is lower-left and 4 is lower-right}
\item{test.rep}{number of resamples in the statistical approximate permutation test. For test.rep = 0 no statistical test is performed}
\item{test.p_confidence}{confidence level of the estimated p-value. \cr
Is used to calculate the p-accuracy for a given number of resamples (test.rep), default 0.95.}
\item{test.p_threshold}{define the threshold significance level in the returned plot of the statistical test, default 0.05}\cr
}

\value{Returns a list of 6 items (see examples for further explanation):
\item{plots}{A list of plot-data for each x-y combination}
\item{summaries}{A list of dataframes with the summaries for each x-y combination}
\item{bottlenecks}{A list of dataframes with a bottleneck table for each ceiling technique}
\item{peers}{A list of ceilings, with a list of peers for each condition.
Peers are corner points of the CE-FDH ceiling line (e.g., the northwest-corners points for corner = 1)}
\item{tests}{The results of the test for each condition (not human friendly, use nca_output)}
\item{test.time}{The total time needed to run the tests for all conditions}
}
\details{
\bold{Corners}\cr
Corner 1 is the upper-left corner and corner 2 is the upper-right corner.
These two corners are used for an analysis of the necessity of the presence/high level
 if x (corner = 1 ) or the absence/low level if x (corner = 2) for the presence/high level
 of y, respectively.\cr
Corner 3 is the lower-left corner and corner 4 is the lower-right corner.
These two corners are used for an analysis of the necessity of the presence/high level
of x (corner = 3 ) or the absence/low level if x (corner = 4) for the absence/low level
of y, respectively.\cr
By default the upper left corner is analysed for all conditions and corner
is not defined. If corner is defined, flip.x and flip.y are ignored.
\cr
\cr
\bold{Scope}\cr
By default, the theoretical scope is not defined and the empirical scope is used based on the minimum and maximum observed values of x and y.
}
\examples{
# Load the data
data(nca.example)
data <- nca.example

# Basic NCA analysis, with conditions in the first 2 columns
# and the outcome in the third column
model <- nca_analysis(data, c(1, 2), 3)

# Use nca_output to show the summaries (see nca_output documentation for more options)
nca_output(model)
\dontshow{Sys.setenv(NCA_SKIP_PURITY=TRUE)}
# Columns can be selected by name as well
model <- nca_analysis(data, c('Individualism', 'Risk taking'), 'Innovation performance')

# Define the ceiling techniques via the ceilings parameter, see 'ceilings' for all types
model <- nca_analysis(data, c(1, 2), 3, ceilings=c('ce_fdh', 'ce_vrs'))

# These are the available ceiling techniques
print(ceilings)

# Define a custom ceiling line by providing the custom parameter, pairs of intercept-slope
model <- nca_analysis(data, c(1, 2), 3, custom=c(0, 1, 0, 2))

# If a single pair is provided, it will be used for all conditions
model <- nca_analysis(data, c(1, 2), 3, custom=c(0, 1))

# By default, the upper-left corner is analysed. With the corner argument for each
# condition a different corner can be selected. Select corner 1 or 2
# for an analysis of necessary conditions for the presence/high level of the
# outcome, and corner 3 or 4 for an analysis of necessary conditions for
# the absence/low level of the outcome. It is not possible to combine
# corner 1 or 2 with corner 3 or 4 in the same analysis as different outcomes are analysed.
# This analyses the upper right corner for the first condition
# and the upper left corner for the second condition:
model <- nca_analysis(data, c(1, 2), 3, corner=c(2, 1))

# Alternatively, for using the upper right corner(s), 'flip' the x variables
model <- nca_analysis(data, c(1, 2), 3, flip.x=TRUE)

# It is also possible to flip a single x variable
model <- nca_analysis(data, c(1, 2), 3, flip.x=c(TRUE, FALSE))

# Flip the y variable if the lower corners need analysing
model <- nca_analysis(data, c(1, 2), 3, flip.x=c(TRUE, FALSE), flip.y=TRUE)

# Use a theoretical scope instead of the (calculated) empirical scope
model <- nca_analysis(data, c(1, 2), 3, scope=c(0, 120, 0, 240))

# Display the peers for a ceiling and a condition
print(model$peers$ce_fdh$Individualism)

# By default, the bottleneck tables use percentages of the range for the x and y values.
# Using the percentage of the max value is also possible
model <- nca_analysis(data, c(1, 2), 3, bottleneck.y='percentage.max')

# Use the actual values, in this case the x-value
model <- nca_analysis(data, c(1, 2), 3, bottleneck.x='actual')

# Use percentile, in this case for the y-values
model <- nca_analysis(data, c(1, 2), 3, bottleneck.y='percentile')

# Any combination is possible
model <- nca_analysis(data, c(1, 2), 3, bottleneck.x='actual', bottleneck.y='percentile')

# The number of steps is adjustible via the steps parameter
model <- nca_analysis(data, c(1, 2), 3, steps=20)

# The steps parameter also accepts a list of values
# These are interpreted as actual or percentage / percentile depending on bottleneck.y
model <- nca_analysis(data, c(1, 2), 3, steps=seq(50, 120, 10))

# Or via the step.size parameter, this ignores the steps parameter
model <- nca_analysis(data, c(1, 2), 3, step.size=5)

# If the ceiling line crosses the X = Xmax line at a point C below Y = Ymax,
# for Y < Yc < Ymax, the corresponding X in the bottleneck table is displayed as 'NA'
# It is also possible to display them as Xmax
model <- nca_analysis(data, c(1, 2), 3, cutoff=1)

# or as the calculated value on the ceiling line
model <- nca_analysis(data, c(1, 2), 3, cutoff=2)

\dontshow{data <- head(data, n=3)}
# To run tests, the test.rep needs to be larger than 0
# Optionally the p_confidence (default 0.95) and the p_threshold (default 0) can be set
model <- nca_analysis(data, c(1), 3, test.rep=1000,
                      test.p_confidence=0.9, test.p_threshold=0.05)
# The output of the tests can be shown via nca_output with test=TRUE
nca_output(model, test=TRUE)

# If reproducible outputs are needed, set the seed to a fixed number
set.seed(42)
model <- nca_analysis(data, c(1), 3, test.rep=1000)
}
\keyword{functions}

\name{nca_output}
\alias{nca_output}
\title{display the result of the NCA analysis}
\description{Show the plots, NCA summaries and bottleneck tables of a NCA analysis.}
\usage{nca_output(model, plots=TRUE, plotly=FALSE, bottlenecks=FALSE,
           summaries=TRUE, test=FALSE, pdf=FALSE, path=NULL, selection = NULL, 
           plot_bottlenecks=0)}
\arguments{
\item{model}{Displays output of the nca or nca_analysis command}
\item{plots}{If true (default) show the scatter plot(s), one  for each condition.\cr
In Rstudio the plots are shown the Plots window.}
\item{plotly}{If true shows the interactive scatter plot(s), one for each condition.\cr
In RStudio the plots are shown in the Viewer window. See Details.}
\item{bottlenecks}{If true displays the bottleneck table(s) in the Console window, one table for each ceiling line}
\item{summaries}{If true shows the summaries for each condition in the Console window, see Details}
\item{test}{If true shows the result of the statistical sigificance test (if present), see Details}
\item{pdf}{If true exports the output to a pdf file, except for the plotly plot}
\item{path}{Optional path for the output file(s)}
\item{selection}{Optionally selects the conditions for inclusion in the output.\cr
For example, only the plots for the selected condition are shown.\cr
In the bottleneck table the required levels of only the selected conditions for given levels of Y are shown.}
\item{plot_bottlenecks}{Optionally show the bottleneck lines}
}
\details{
\strong{Plotly}\cr
The plot highlights the points that construct the ceiling line ('peers').
The names and XY coordinates of all points pop up when moving the pointer over the plot.
The toolbar allows several actions such as zoom and selection of parts of the plot. Optionally subgroups of points can be labeled.
\cr
\cr
\bold{Summaries}\cr
The output starts with 6 lines of basic information ("global") about the dataset ("Number of observations", "Scope", "Xmin", "Xmax", "Ymin", and "Ymax").
"Scope" refers to the empirical area of possible x-y combinations, given the minimum and maximum observed x and y values.\cr
The next 11 lines present the NCA parameters ("param", see below) for each of the selected ceiling techniques (the defaults techniques are ce_fdh and cr_fdh).\cr
\cr
The 11 printed NCA parameters are:\cr
- \emph{Ceiling zone}, which is the size of the "empty" area in the upper-left corner\cr
- \emph{Effect size}, which is the ceiling zone divided by the scope\cr
- \emph{# above}, which is the number of observations that are above the ceiling line and hence in the "empty" ceiling zone\cr
- \emph{Ceiling accuracy}, which is the number of observations on or below the ceiling line divided by the total number of observations and multiplied by 100 percent\cr
- \emph{Fit}, which relates to the "closeness" of the selected ceiling line to the ce_fdh ceiling line\cr
- \emph{Slope} and \emph{Intercept}, which are the slope and the intercept of the straight ceiling line (no values are printed if the ceiling line is not a straight line, but a step function)\cr
- \emph{Abs. ineff.}, which is the total xy-space where x does not constrain y, and y is not constrained by x\cr
- \emph{Rel. ineff.}, which is the total xy-space where x does not constrain y, and y is not constrained by x as percentage of the scope\cr
- \emph{Condition ineff.}, which is the condition inefficiency that indicates for which range of x (as a percentage of the total range) x does not constrain y (i.e., there is no ceiling line in that x-range)\cr
- \emph{Outcome ineff.}, which is the outcome efficiency that indicates for which range of y (as a percentage of the total range of y) y is not constrained by x (i.e., there is no ceiling line in that y-range)
\cr
\cr
\bold{Test}\cr
NCA's statistical test is a randomness test to evaluate if the observed effect size may be a random result of unrelated X and Y variables
}
\examples{\dontshow{Sys.setenv(NCA_SKIP_PURITY=TRUE)}
# Use the result of the nca command:
data(nca.example)
data <- nca.example
model <- nca_analysis(data, c(1, 2), 3)

# Show the full summaries in the Console window
nca_output(model)

# Suppress the summaries and display the plots
nca_output(model, plots=TRUE, summaries=FALSE)

# Display the plots via Plotly
\donttest{nca_output(model, plotly=TRUE, summaries=FALSE)}

# Label the observation of the Plotly plot by using a vector of names (no more than 5).
# For example label the observations in nca.example
labels <- c('Australia', 'Europe', 'Europe', 'North America', 'Europe', 'Europe',
'Europe', 'Europe', 'Europe', 'Europe', 'Europe', 'Europe', 'Europe', 'Asia',
'North America', 'Europe', 'Australia', 'Europe', 'Europe', 'Europe', 'Europe',
'Asia', 'Europe', 'Europe', 'Europe', 'Europe', 'Europe', 'North America')
nca_output(model, plotly=labels, summaries=FALSE)

# Suppress the summaries and display the bottlenecks
nca_output(model, bottlenecks=TRUE, summaries=FALSE)

# Show the results of the statistical significance test (p-value)
# Make sure to set test.rep in nca_analysis
nca_output(model, test=TRUE)

# Show all five
\donttest{nca_output(model, plots=TRUE, plotly=TRUE, bottlenecks=TRUE, test=TRUE)}

# Per condition, export plots and summaries to PDF files,
# and export all the bottleneck tables to a single PDF file
\donttest{nca_output(model, plots=TRUE, bottlenecks=TRUE, pdf=TRUE)}
\dontshow{file.remove(dir(path=".", pattern = "*.pdf"))}
# Use the path option to export to an existing directory
outdir <- '/tmp' \dontshow{outdir <- tempdir()}
nca_output(model, plots=TRUE, pdf=TRUE, path=outdir)

# Limit the output to a selection of conditions by name
nca_output(model, plots=TRUE, selection=c("Individualism"))

# Or by column index, in both cases the order matters
nca_output(model, plots=TRUE, selection=c(2, 1))

# Show the bottleneck lines in the plots or plotly
# 1 shows the outcome lines, 2 also shows the lines for the condition
nca_output(model, plots = TRUE, bottlenecks = TRUE, summaries = FALSE, plot_bottlenecks = 1)
}
\keyword{functions}

\name{nca_power}
\alias{nca_power}
\title{Function to evaluate power}
\description{Function to evaluate power,
test if a sample size is large enough to detect necessity.}
\usage{nca_power(n = c(20, 50, 100), effect = 0.10, slope = 1, ceiling = "ce_fdh",
corner = 1, p = 0.05, distribution.x = "uniform", distribution.y = "uniform",
rep = 100, test.rep = 200)}
\arguments{
\item{n}{Number of datapoints to generate, either an integer or a vector of integers.}
\item{effect}{Effect size of the generated datasets (single value or vector of values).}
\item{slope}{Slope of the line (single value or vector of values).}
\item{ceiling}{Ceiling technique to use for this analysis (single value or vector of values).}
\item{corner}{Integer, indicating the corner to analyze, see nca_analysis.}
\item{p}{Significance level.}
\item{distribution.x}{Distribution type(s) for X, "uniform" (default) or "normal" (single value or vector of values).}
\item{distribution.y}{Distribution type(s) for Y, "uniform" (default) or "normal" (single value or vector of values).}
\item{rep}{Number of analyses done per iteration.}
\item{test.rep}{Number of resamples in the statistical approximate permutation test.
For test.rep = 0 no statistical test is performed. See \code{\link{nca_analysis}}
for reproducible outputs.}
}
\examples{
# Simple example
\dontrun{results <- nca_power()}
\dontshow{results <- nca_power(rep=1, test.rep = 1)}
print(results)
}
\keyword{functions}

\name{nca.example}
\docType{data}
\alias{nca.example}
\title{NCA example data with 2 conditions and 1 output}
\description{
This data set has Individualism and Risk taking	as conditions,
and Innovation performance as the output for 28 countries.
}
\usage{data(nca.example)}
\format{A matrix containing 28 observations, incl. headers and row names.}
\keyword{datasets}

\name{nca_difference}
\alias{nca_difference}
\title{Function for permutation tests}
\description{Function for permutation tests}
\usage{nca_difference(data1, data2 = NULL, x, y,
ceilings = c("ce_fdh", "cr_fdh"), scope = NULL, test.rep = 1000,
test.type = c("contrast", "independent", "paired"))}
\arguments{
\item{data1}{Dataframe 1.}
\item{data2}{Dataframe 2, defaults to NULL.}
\item{x}{Index or name (or a vector of those) with condition(s) x}
\item{y}{Index or name of the column with the outcome y}
\item{ceilings}{Vector with the ceiling techniques to test}
\item{scope}{A theoretical scope in list format : (x.low, x.high, y.low, y.high), see nca_analysis}
\item{test.rep}{Number of permutations to use, defaults to 1000.}
\item{test.type}{Name of the test type: "contrast", "independent" or "paired".}
}
\examples{
ceilings <- "cr_fdh"
y <- "Y"
test.rep <- 1000
\dontshow{test.rep = 1}
# Contrast test: effect size difference between two conditions
test.type <- "contrast"
data1 <- nca_random(100, c(0.2,0.4), c(1,1))
x <- c("X1", "X2")

difference <- nca_difference(data1 = data1, x = x, y = y,
                             ceilings = ceilings, test.rep = test.rep,
                             test.type = test.type)
print(difference)

# Independent test: effect size difference between two datasets
test.type <- "independent"
data1 <- nca_random(100, 0.2, 1)
data2 <- nca_random(100, 0.4, 1)
x <- "X"

difference <- nca_difference(data1 = data1, data2 = data2, x = x, y = y,
                             ceilings = ceilings, test.rep = test.rep,
                             test.type = test.type)
print(difference)

# paired test: effect size differences between 2 measurements
test.type <- "paired"
data1 <- nca_random(100, 0.2, 1)
data2 <- nca_random(100, 0.4, 1)
x <- "X"

difference <- nca_difference(data1 = data1, data2 = data2, x = x, y = y,
                             ceilings = ceilings, test.rep = test.rep,
                             test.type = test.type)
print(difference)
}
\keyword{functions}

\name{nca_difference}
\alias{nca_difference}
\title{Function for permutation tests}
\description{Function for permutation tests}
\usage{nca_difference(data1, data2 = NULL, x, y,
ceilings = c("ce_fdh", "cr_fdh"), common_scope, test.rep = 1000,
test.type = c("contrast", "independent", "paired"))}
\arguments{
\item{data1}{Dataframe 1.}
\item{data2}{Dataframe 2, defaults to NULL.}
\item{x}{Index or name (or a vector of those) with condition(s) x}
\item{y}{Index or name of the column with the outcome y}
\item{ceilings}{Vector with the ceiling techniques to test}
\item{common_scope}{The common scope for both datasets : (x.low, x.high, y.low, y.high), see nca_analysis}
\item{test.rep}{Number of permutations to use, defaults to 1000.}
\item{test.type}{Name of the test type: "contrast", "independent" or "paired".}
}
\examples{
ceilings <- "cr_fdh"
y <- "Y"
test.rep <- 1000
\dontshow{test.rep = 1}
# Contrast test: effect size difference between two conditions
test.type <- "contrast"
data1 <- nca_random(100, c(0.2, 0.4), c(1,1))
x <- c("X1", "X2")

# For the contrast test, the dataset should be normalized. Use c(0, 1, 0, 1) as common scope
data1 <- nca_util_normalize(data1)
common_scope <- c(0, 1, 0, 1)

difference <- nca_difference(data1 = data1, x = x, y = y,
                             ceilings = ceilings, common_scope = common_scope,
                             test.rep = test.rep, test.type = test.type)
print(difference)

# Independent test: effect size difference between two datasets
test.type <- "independent"
data1 <- nca_random(100, 0.2, 1)
data2 <- nca_random(100, 0.4, 1)
x <- "X"

# For the independent test, use the (empirical) pooled scope as commn scope
# Or use a theoretical scope, larger then the emperical pooled scope 
common_scope <- c(min(data1[, x], data2[, x]), max(data1[, x], data2[, x]),
                  min(data1[, y], data2[, y]), max(data1[, y], data2[, y]))

difference <- nca_difference(data1 = data1, data2 = data2, x = x, y = y,
                             ceilings = ceilings, common_scope = common_scope,
                             test.rep = test.rep, test.type = test.type)
print(difference)

# paired test: effect size differences between 2 measurements
test.type <- "paired"
data1 <- nca_random(100, 0.2, 1)
data2 <- nca_random(100, 0.4, 1)
x <- "X"

# For the paired test, use the (empirical) pooled scope as commn scope
# Or use a theoretical scope, larger then the emperical pooled scope 
common_scope <- c(min(data1[, x], data2[, x]), max(data1[, x], data2[, x]),
                  min(data1[, y], data2[, y]), max(data1[, y], data2[, y]))

difference <- nca_difference(data1 = data1, data2 = data2, x = x, y = y,
                             ceilings = ceilings, common_scope = common_scope,
                             test.rep = test.rep, test.type = test.type)
print(difference)
}
\keyword{functions}

\name{NCA-package}
\alias{NCA-package}
\docType{package}
\title{Necessary Condition Analysis}
\description{The NCA package implements Necessary Condition Analysis (NCA) as developed by Dul (2016). For running the NCA package a data file (e.g., mydata.csv, which contains the input data) must be available. An example data file (presented in above article) is included in the package. The user must load the data and call the nca function.}
\details{\tabular{ll}{
Package:  \tab NCA\cr
Type:     \tab Package\cr
Version:  \tab 5.0.1\cr
Date:     \tab 2026-04-29\cr
License:  \tab GPL (>= 3)\cr}
}
\author{
Author: Jan Dul \email{jdul@rsm.nl}\cr
Maintainer: Govert Buijs \email{buijs@rsm.nl}
}
\references{
Dul, J. 2016. Necessary Condition Analysis (NCA).Logic and methodology of 'necessary but not sufficient' causality.
Organizational Research Methods 19(1), 10-52. \doi{10.1177/1094428115584005}\cr
Dul, J. (2020). Conducting Necessary Condition Analysis. Sage publishers. ISBN: 9781526460141.
\href{https://uk.sagepub.com/en-gb/eur/conducting-necessary-condition-analysis-for-business-and-management-students/book262898}{https://uk.sagepub.com/en-gb/eur/conducting-necessary-condition-analysis-for-business-and-management-students/book262898}\cr
Dul, J., van der Laan, E., & Kuik, R. (2020). A statistical significance test for Necessary Condition Analysis. Organizational Research Methods, 23(2), 385-395.
\doi{10.1177/1094428118795272}

}
\seealso{
\code{\link{nca_analysis}}, \code{\link{nca_output}}
}
\keyword{package}

\examples{
# A more detailed guide can be found here : https://repub.eur.nl/pub/78323/
# or https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2624981

# Load data from a CSV file with header and row names:
\donttest{\dontshow{try(}data <- read.csv('mydata.csv', row.names=1)\dontshow{, silent=TRUE)}}
# Or load and rename the example dataset
data(nca.example)
data <- nca.example

# Run NCA with the dataset and name the analysis 'model'.
# Specify the conditions and outcomes by column index or name
# More than 1 condition can be specified with a vector
model <- nca_analysis(data, c(1, 2), 3)

# A quick summary of the analysis can be displayed by 'model'
model

# A full summary of the analysis is shown by nca_output (see documentation for more options)
nca_output(model)

# The results of the analysis is a list of 6 items :
# - plots (1 for each condition)
# - summaries (1 for each condition)
# - bottleneck tables (1 for each ceiling technique)
# - peers (1 dataframe for each condition)
# - tests (1 list for each condition)
# - test.time (total time to run all tests)
names(model)

# The first item contains the graphical outputs for each condition
# This is not really useful to humans
model$plots[[1]]

# The second item contains a list with the summaries for the conditions
# Extract individual elements with nca_extract
model$summaries[[1]]

# The third item contains a list with the bottleneck tables, one for each ceiling technique
model$bottlenecks$cr_fdh

# The fourth item shows the peers, for each condition
model$peers$Individualism

# For the fifth and sixth item, the test.rep needs to be larger than 0
# for performing the statistical test
# Optionally the p_confidence (default 0.95) and the p_threshold (default 0) can be set
model <- nca_analysis(data, c(1, 2), 3, test.rep=100)

# The fifth item shows the tests for each condition
# This is not really useful to humans
model$tests$Individualism

# The last item shows the total time needed to perform the analysis.
# For large values of test.rep the test may take long.
model$test.time
}

\name{NCA-package}
\alias{NCA-package}
\docType{package}
\title{Necessary Condition Analysis}
\description{The NCA package implements Necessary Condition Analysis (NCA) as developed by Dul (2016). For running the NCA package a data file (e.g., mydata.csv, which contains the input data) must be available. An example data file (presented in above article) is included in the package. The user must load the data and call the nca function.}
\details{\tabular{ll}{
Package:  \tab NCA\cr
Type:     \tab Package\cr
Version:  \tab 5.0.2\cr
Date:     \tab 2026-06-16\cr
License:  \tab GPL (>= 3)\cr}
}
\author{
Author: Jan Dul \email{jdul@rsm.nl}\cr
Maintainer: Govert Buijs \email{buijs@rsm.nl}
}
\references{
Dul, J. 2016. Necessary Condition Analysis (NCA).Logic and methodology of 'necessary but not sufficient' causality.
Organizational Research Methods 19(1), 10-52. \doi{10.1177/1094428115584005}\cr
Dul, J. (2020). Conducting Necessary Condition Analysis. Sage publishers. ISBN: 9781526460141.
\href{https://uk.sagepub.com/en-gb/eur/conducting-necessary-condition-analysis-for-business-and-management-students/book262898}{https://uk.sagepub.com/en-gb/eur/conducting-necessary-condition-analysis-for-business-and-management-students/book262898}\cr
Dul, J., van der Laan, E., & Kuik, R. (2020). A statistical significance test for Necessary Condition Analysis. Organizational Research Methods, 23(2), 385-395.
\doi{10.1177/1094428118795272}

}
\seealso{
\code{\link{nca_analysis}}, \code{\link{nca_output}}
}
\keyword{package}

\examples{
# A more detailed guide can be found here : https://repub.eur.nl/pub/78323/
# or https://papers.ssrn.com/sol3/papers.cfm?abstract_id=2624981

# Load data from a CSV file with header and row names:
\donttest{\dontshow{try(}data <- read.csv('mydata.csv', row.names=1)\dontshow{, silent=TRUE)}}
# Or load and rename the example dataset
data(nca.example)
data <- nca.example

# Run NCA with the dataset and name the analysis 'model'.
# Specify the conditions and outcomes by column index or name
# More than 1 condition can be specified with a vector
model <- nca_analysis(data, c(1, 2), 3)

# A quick summary of the analysis can be displayed by 'model'
model

# A full summary of the analysis is shown by nca_output (see documentation for more options)
nca_output(model)

# The results of the analysis is a list of 6 items :
# - plots (1 for each condition)
# - summaries (1 for each condition)
# - bottleneck tables (1 for each ceiling technique)
# - peers (1 dataframe for each condition)
# - tests (1 list for each condition)
# - test.time (total time to run all tests)
names(model)

# The first item contains the graphical outputs for each condition
# This is not really useful to humans
model$plots[[1]]

# The second item contains a list with the summaries for the conditions
# Extract individual elements with nca_extract
model$summaries[[1]]

# The third item contains a list with the bottleneck tables, one for each ceiling technique
model$bottlenecks$cr_fdh

# The fourth item shows the peers, for each condition
model$peers$Individualism

# For the fifth and sixth item, the test.rep needs to be larger than 0
# for performing the statistical test
# Optionally the p_confidence (default 0.95) and the p_threshold (default 0) can be set
model <- nca_analysis(data, c(1, 2), 3, test.rep=100)

# The fifth item shows the tests for each condition
# This is not really useful to humans
model$tests$Individualism

# The last item shows the total time needed to perform the analysis.
# For large values of test.rep the test may take long.
model$test.time
}

\name{nca.example2}
\docType{data}
\alias{nca.example2}
\title{NCA example data with 3 conditions and 1 outcome}
\description{
This data set has Contractual detail, Goodwill trust, and Competence trust
as conditions, and Innovation as outcome for 48 buyer-supplier relationships.
See Table 2 in: Van der Valk, W., Sumo, R., Dul, J., & Schroeder, R. G. (2016).
When are contracts and trust necessary for innovation in buyer-supplier relationships?
A necessary condition analysis.
Journal of Purchasing and Supply Management, 22(4), 266-277.
}
\usage{data(nca.example2)}
\format{A matrix containing 48 observations, incl. headers.}
\keyword{datasets}

\name{nca_powerplot}
\alias{nca_powerplot}
\title{Function to show the powerplot}
\description{Function to show the powerplot, optional over various variables.}
\usage{nca_powerplot(df_power, x.variable = 'n', x.name = 'Sample size',
            x.min = 0, x.max = NULL,
            line.variable = 'ES', line.name = 'Effect size')}
\arguments{
\item{df_power}{Dataframe as returned by the nca_power function.}
\item{x.variable}{The variable to use for the X scale, default is sample size.}
\item{x.name}{The name for the X scale.}
\item{x.min}{The minimum value for the X scale, defaults to 0.}
\item{x.max}{The maximum value for the X scale, defaults to the maximum X.}
\item{line.variable}{Variable that represented the line.}
\item{line.name}{The name of the variable that represented the line.}
}
\examples{
# Simple example, showing a normal plot with the power as a function of the sample size
# A line for each effect size will be shown
\dontrun{results <- nca_power(effect = c(.1, .2, .3))}
\dontshow{results <- nca_power(effect = c(.1, .2, .3), rep=2, test.rep = 1)}
\donttest{nca_powerplot(results)}

# This example shows the power for different ceiling lines
\dontrun{results <- nca_power(ceiling = c("ce_fdh", "cr_fdh"))}
\dontshow{results <- nca_power(ceiling = c("ce_fdh", "cr_fdh"), rep=2, test.rep = 1)}
\donttest{powerplot <- nca_powerplot(results, line.variable='ceiling', line.name='Ceiling')}
}
\keyword{functions}

\name{nca_util_normalize}
\alias{nca_util_normalize}
\title{Normalizes numeric data}
\description{(Min-max) Normalize the columns of the data}
\usage{nca_util_normalize(data, scale = NULL, min_max = NULL)}
\arguments{
\item{data}{Input dataframe or matrix.}
\item{scale}{Set(s) of output Min-Max values.}
\item{min_max}{Set(s) of (theoretical) Min-Max values.}
}
\details{
\strong{scale}\cr
If scale is NULL (default), all columns will be normalized between 0 and 1.\cr
Use c(0, 100) to normalize between 0 and 100.\cr
Replicate for each column for different scale, e.g. c(0, 1, 0, 1, 0, 100).\cr
Use NA in the min-max pair if a column does not need to be normalized.\cr
}
\examples{
# Normalize the data between 0 and 1
data(nca.example)
data.normalized <- nca_util_normalize(nca.example)
# Equivalent to this
data.normalized <- nca_util_normalize(nca.example, c(0, 1))

# Normalize the data between 0 and 100
data.normalized <- nca_util_normalize(nca.example, c(0, 100))

# Normalize each column with different values
data.normalized <- nca_util_normalize(nca.example, c(0, 1, 0, 10, 0, 100))

# Use NA if a column doesn't need to be normalized
data.normalized <- nca_util_normalize(nca.example, c(0, 1, NA, NA, 0, 100))

# It is also possible to use a theoretic range for the normalization
data.normalized <- nca_util_normalize(nca.example, min_max = c(0, 100, 0, 150, 0, 300))
}
\keyword{functions}

\name{line.width}
\alias{line.width}
\docType{data}
\title{parameter defining the width of the lines in the plots}
\description{This will be used for the \bold{lwd} parameter of the plot, default is 1.5.\cr
Set before calling nca_output\cr
> line.width <- 5
}
\keyword{parameter}

\name{line.types}
\alias{line.types}
\docType{data}
\title{a set defining the line types for the plots}
\description{Set before calling nca_output\cr
> line.types['ce_fdh'] <- 1\cr\cr
Reset one line type by setting it to NULL\cr
> line.types['ce_fdh'] <- NULL\cr\cr
Reset all line types by setting line.types to NULL\cr
> line.types <- NULL
}
\format{
This is a list with default line types for each ceiling technique
\tabular{lllll}{
\tab \bold{ols}         \tab 1\tab \bold{c_lp}        \tab 2\cr
\tab \bold{cols}        \tab 3\tab \bold{qr}          \tab 4\cr
\tab \bold{ce_vrs}      \tab 5\tab \bold{cr_vrs}      \tab 1\cr
\tab \bold{ce_fdh}      \tab 6\tab \bold{cr_fdh}      \tab 1\cr
}
}
\keyword{parameter}

\name{line.colors}
\alias{line.colors}
\docType{data}
\title{a set defining the line colors for the plots}
\description{Set before calling nca_output\cr
> line.colors['ce_fdh'] <- 'blue'\cr\cr
Reset one line color by setting it to NULL\cr
> line.colors['ce_fdh'] <- NULL\cr\cr
Reset all line colors by setting line.colors to NULL\cr
> line.colors <- NULL
}
\format{
This is a list with default line colors for each ceiling technique
\tabular{lllll}{
\tab \bold{ols}         \tab 'green'          \tab \bold{c_lp}        \tab 'blue'\cr
\tab \bold{cols}        \tab 'darkgreen'      \tab \bold{qr}          \tab 'lightpink'\cr
\tab \bold{ce_vrs}      \tab 'orchid4'        \tab \bold{cr_vrs}      \tab 'violet'\cr
\tab \bold{ce_fdh}      \tab 'red'            \tab \bold{cr_fdh}      \tab 'orange'\cr
}
}

\keyword{parameter}

\name{nca_random}
\alias{nca_random}
\title{generating random data that meets necessity}
\description{Generate N datapoints, with 'normal' or 'uniform' distributions for X and Y}
\usage{nca_random(n, intercepts, slopes, corner=1,
                  distribution.x = "uniform", distribution.y = "uniform",
                  mean.x = 0.5, mean.y = 0.5, sd.x = 0.2, sd.y = 0.2)}
\arguments{
\item{n}{Number of observations to generate, should be an integer > 1.}
\item{intercepts}{The intercept or a vector of intercepts of the line.}
\item{slopes}{The slope or a vector if slopes of the line.}
\item{corner}{Define which corner should be empty, default is 1 (upper left).}
\item{distribution.x}{Type of the distribution for X, "uniform" (default) or "normal".\cr
The latter is a truncated normal distribution.}
\item{distribution.y}{Type of the distribution for Y, "uniform" (default) or "normal".\cr
The latter is a truncated normal distribution.}
\item{mean.x}{Distribution Mean of X (default 0.5), ignored distribution.x == "uniform".}
\item{mean.y}{Distribution Mean of Y (default 0.5), ignored distribution.y == "uniform".}
\item{sd.x}{Distribution SD of X (default 0.2), ignored distribution.x == "uniform".}
\item{sd.y}{Distribution SD of Y (default 0.2), ignored distribution.y == "uniform".}
}
\examples{
# Generate a uniform dataset, default for X and Y
data <- nca_random(100, 0, 1)

# It is also possible to generate a dataset with multiple conditions,
# by supplying vectors for the intercepts and slopes
data <- nca_random(100, c(0, 0.25), c(1, 0.75))
# Single values will be repeated to complement a vector
data <- nca_random(100, c(0, 0.25), 1)

# The default is an empty space in the upper left corner.
# A different corner can be selected with the corner argument
data <- nca_random(100, 0, 1, corner=4)

# Generate a dataset with a normal distribution for X and a uniform distribution for Y
data <- nca_random(100, 0, 1, distribution.x = "normal", distribution.y = "uniform")

# Generate a dataset with a normal distribution for X and Y, with adjusted MEAN
data <- nca_random(100, 0, 1, distribution.x = "normal", distribution.y = "normal",
                   mean.x = 0.75, mean.y = 0.75)

# Generate a dataset with a normal distribution for X and Y, with adjusted SD
data <- nca_random(100, 0, 1, distribution.x = "normal",
                   distribution.y = "normal", sd.x = 0.1, sd.y = 0.1)
}
\keyword{functions}

\name{point.type}
\alias{point.type}
\docType{data}
\title{parameter defining the plotting symbol in the plots}
\description{This will be used for the \bold{pch} parameter of the plots, default is 21.\cr
Set before calling nca_output\cr
> point.type <- 22\cr\cr
See \href{https://www.sthda.com/english/wiki/r-plot-pch-symbols-the-different-point-shapes-available-in-r?title=r-plot-pch-symbols-the-different-point-shapes-available-in-r}{
http://www.sthda.com/english/wiki/r-plot-pch-symbols-the-different-point-shapes-available-in-r} for more symbols
}
\keyword{parameter}

\name{nca}
\alias{nca}
\title{Run a basic NCA analyses on a data set}
\description{Run a basic NCA analyses on a data set}
\usage{nca(data, x, y, ceilings=c('ols', 'ce_fdh', 'cr_fdh'))}
\arguments{
\item{data}{dataframe with columns of the variables}
\item{x}{collection of the columns with the conditions}
\item{y}{index or name of the column with the outcomes}
\item{ceilings}{vector with the ceiling techniques to include in this analysis}
}
\value{Returns a list with 3 items (see examples for further explanation):
\item{plots}{A list of plot-data for each x-y combination}
\item{summaries}{A list of dataframes with the summaries for each x-y combination}
\item{bottlenecks}{A list of dataframes with a bottleneck table for each ceiling technique}
}
\examples{
# Load the data
data(nca.example)
data <- nca.example

# Basic NCA analysis
# Conditions in the first 2 columns, outcome in the third column
# This shows scatter plot(s) with the ceiling lines and the effect size(s) on the console
nca(data, c(1, 2), 3)

# Columns can be selected by name as well
nca(data, c('Individualism', 'Risk taking'), 'Innovation performance')

# Define the ceiling techniques via the ceilings parameter
nca(data, c(1, 2), 3, ceilings=c('ols', 'ce_vrs'))
# These are the available ceiling techniques
print(ceilings)
}
\keyword{functions}

\name{ceilings}
\alias{ceilings}
\docType{data}
\title{a set of all available ceiling techniques}
\description{Ceilings to use for the \code{\link{nca}} or \code{\link{nca_analysis}} methods \cr
> nca(data, c(1, 2), 3, ceilings=c('ols', 'ce_fdh', 'cr_fdh'))

Note that the ols regression line is not a ceiling but is included as a reference. 

\tabular{ll}{
\bold{Ceiling Technique} \tab \bold{Name}\cr

\bold{cols}         \tab Corrected Ordinary Least Squares                 \cr
\bold{qr}           \tab Quantile Regression                              \cr
\bold{ce_vrs}       \tab Ceiling Envelopment with Varying Return to Scale \cr
\bold{cr_vrs}       \tab Ceiling Regression with Varying Return to Scale  \cr
\bold{ce_fdh}       \tab Ceiling Envelopment with Free Disposal Hull      \cr
\bold{cr_fdh}       \tab Ceiling Regression with Free Disposal Hull       \cr
\bold{c_lp}         \tab Ceiling Linear Programming                       \cr
}

Note: The SFA and LH ceiling lines are deprecated (discontinued) from version 3.2.0
}

\keyword{parameter}

\name{nca_extract}
\alias{nca_extract}
\title{get model parameters}
\description{Get the model parameters.}
\usage{nca_extract(model, x=NULL, ceiling=NULL, param='Effect size')}
\arguments{
\item{model}{The model produced by nca_analysis}
\item{x}{Name of the condition, defaults to the first}
\item{ceiling}{Name of the ceiling, defaults to the first}
\item{param}{Name of the parameter, default 'Effect size'}
}
\examples{
data(nca.example)
model <- nca_analysis(nca.example, c('Individualism', 'Risk taking'), 3)

# The names of the parameters can be found in the summaries
nca_output(model, plots=FALSE, summaries=TRUE)

# Get the Ceiling zone
extract <- nca_extract(model, 'Individualism', 'ce_fdh', 'Ceiling zone')
print(extract)

# Get the Scope
extract <- nca_extract(model, 'Individualism', 'ce_fdh', 'Scope')
print(extract)

# For multiple values: vectorize one argument and pin the others
names <- c('Individualism', 'Risk taking')
extracts <- sapply(names, nca_extract,
                   model=model, ceiling='ce_fdh', param='Ceiling zone')
print(extracts)

params <- c('Effect size', 'Ceiling zone')
extracts <- sapply(params, nca_extract,
                   model=model, x='Individualism', ceiling='ce_fdh')
print(extracts)

# It is also possible to get the extract directly from the model
# Get the Ceiling zone
print(model$`Ceiling zone`$Individualism$ce_fdh)

# Get the Scope, global parameters are independent of the ceiling  
print(model$Scope$Individualism)
}
\keyword{functions}

\name{nca_outliers}
\alias{nca_outliers}
\title{Outlier detection}
\description{Detect outliers on the dataset.}
\usage{nca_outliers(data, x, y, ceiling = NULL,
corner = NULL, flip.x = FALSE, flip.y = FALSE, scope=NULL,
k = 1, min.dif = 1e-2, max.results = 25, plotly=FALSE, condensed = FALSE)}
\arguments{
\item{data}{Dataframe with columns of the variables}
\item{x}{Index or name of the column with the condition}
\item{y}{Index or name of the column with the outcome}
\item{ceiling}{Name of the ceiling technique to be used.
If not provided, the default ceilings (CE_FDH) will be used}
\item{corner}{either an integer or a vector of integers, indicating the corner to analyze, see Details}
\item{flip.x}{reverse the direction of the conditions\cr
 Use either a boolean for all conditions, or a vector with the same length as x}
\item{flip.y}{reverse the direction of the outcomes, boolean}
\item{scope}{a theoretical scope in list format : (x.low, x.high, y.low, y.high), see nca_analysis}
\item{k}{use combinations of observations, default is 1 (single observation)}
\item{min.dif}{set the threshold for the minimum dif.rel to be considered as outlier, default is 1e-2}
\item{max.results}{only show the first 'max.results' outliers, default is 25}
\item{plotly}{If true shows the interactive scatter plot(s), one for each condition.\cr
In RStudio the plots are shown in the Viewer window. See Details.}
\item{condensed}{If true and k > 1, hide outlier combinations for which the effect size is 
not increased compared to the effect size of single elements of that combination. (default FALSE)}
}
\details{
\strong{Outliers}\cr
The potential outliers are displayed with the original effects size and the
effect size if this outlier is removed. The absolute and relative differences
between both effect sizes is also shown.
The table also displays if a point is a ceiling zone outlier or a scope outlier.

\strong{Plotly}\cr
The plot highlights the potential outliers.
The names, relative difference and XY coordinates of all points pop up when moving the pointer over the plot.
The toolbar allows several actions such as zoom and selection of parts of the plot.
\cr
\cr
}
\examples{
# A basic example of the nca_outliers command:
data(nca.example)
outliers <- nca_outliers(nca.example, 1, 3)

# This prints the outlier table
print(outliers)

# Plotly displays a scatterplot with the outliers
nca_outliers(nca.example, 1, 3, plotly = TRUE)

# Test for combinations of observations
# Useful to detect clusters of observations as possible outliers
nca_outliers(nca.example, 1, 3, k = 2)

# Just like the nca_analysis command, nca_outliers accept both flip and corner arguments
nca_outliers(nca.example, 1, 3, corner=3)

# It is possible to define the maximum number of results (default is 25)
nca_outliers(nca.example, 1, 3, max.results=5)

# Do no show possible outliers where the abs(dif.rel) is smaller than min.dif
nca_outliers(nca.example, 1, 3, min.dif=10)

# If k > 1, the effect size of a single observation might not change
# when paired with another observation, e.g. dif.rel of Obs1 == dif.rel of Obs1+Obs2.
# The example below hides combinations of Japan with Portugal, Greece, etc.
nca_outliers(nca.example, 1, 3, k = 2, condensed = TRUE)
}
\keyword{functions}

\name{point.color}
\alias{point.color}
\docType{data}
\title{parameter defining the point color in the plots}
\description{This will be used for the \bold{col} parameters of the plots, default is blue.\cr
Set before calling nca_output\cr
> point.color <- red
}
\keyword{parameter}
